""" =================================================
# Copyright (c) Facebook, Inc. and its affiliates
Authors  :: Cameron Berg (cameronberg@meta.com), Vikash Kumar (vikashplus@gmail.com), Vittorio Caggiano (caggiano@gmail.com)
================================================= """

import collections
import numpy as np
import gym

from myosuite.envs.myo.base_v0 import BaseV0
from myosuite.utils.quat_math import euler2quat, mulQuat, negQuat, mat2quat
from myosuite.utils.vector_math import calculate_cosine

class ProprioceptiveEnvV0(BaseV0):

    SAR_CREDIT = """\
    SAR: Generalization of Physiological Agility and Dexterity via Synergistic Action Representation
        Cameron Berg, Vittorio Caggiano*, Vikash Kumar*
        RSS-2023 | https://sites.google.com/view/sar-rl/
    """

    DEFAULT_OBS_KEYS =['hand_jnt','obj_pos','obj_vel','obj_rot','obj_des_rot','obj_err_pos','obj_err_rot','act','mlen','mvel','mforce']

    DEFAULT_RWD_KEYS_AND_WEIGHTS= {
        'pos_align':1.0,
        'rot_align':1.0,
        'act_reg':5.,
        'drop':5.0,
        'bonus':10.0
    }

    def __init__(self, model_path, obsd_model_path=None, seed=None, **kwargs):
        # EzPickle.__init__(**locals()) is capturing the input dictionary of the init method of this class.
        # In order to successfully capture all arguments we need to call gym.utils.EzPickle.__init__(**locals())
        # at the leaf level, when we do inheritance like we do here.
        # kwargs is needed at the top level to account for injection of __class__ keyword.
        # Also see: https://github.com/openai/gym/pull/1497
        gym.utils.EzPickle.__init__(self, model_path, obsd_model_path, seed, **kwargs)

        # This two step construction is required for pickling to work correctly. All arguments to all __init__
        # calls must be pickle friendly. Things like sim / sim_obsd are NOT pickle friendly. Therefore we
        # first construct the inheritance chain, which is just __init__ calls all the way down, with env_base
        # creating the sim / sim_obsd instances. Next we run through "setup"  which relies on sim / sim_obsd
        # created in __init__ to complete the setup.
        super().__init__(model_path=model_path, obsd_model_path=obsd_model_path, seed=seed, env_credits=self.SAR_CREDIT)

        self._setup(**kwargs)


    def _setup(self,
            obs_keys:list = DEFAULT_OBS_KEYS,
            weighted_reward_keys:list = DEFAULT_RWD_KEYS_AND_WEIGHTS,
            **kwargs,
        ):

        self.target_obj_bid = self.sim.model.body_name2id("target")
        self.S_grasp_sid = self.sim.model.site_name2id('S_grasp')
        self.obj_bid = self.sim.model.body_name2id('Object')
        self.eps_ball_sid = self.sim.model.site_name2id('eps_ball')

        self.success_indicator_sid = self.sim.model.site_name2id("success")

        self.obj_t_gid = self.sim.model.geom_name2id('top')
        self.obj_b_gid = self.sim.model.geom_name2id('bot')
        self.tar_t_gid = self.sim.model.geom_name2id('t_top')
        self.tar_b_gid = self.sim.model.geom_name2id('t_bot')

        self.pen_length = np.linalg.norm(self.sim.model.geom_pos[self.obj_t_gid] - self.sim.model.geom_pos[self.obj_b_gid])
        self.tar_length = np.linalg.norm(self.sim.model.geom_pos[self.tar_t_gid] - self.sim.model.geom_pos[self.tar_b_gid])

        self.sim.model.body_mass[self.obj_bid] *= 1.25
        super()._setup(obs_keys=['hand_jnt','obj_pos','obj_vel','obj_rot','obj_des_rot',
                                 'obj_err_pos','obj_err_rot','mlen','mvel','mforce'],
                        weighted_reward_keys=weighted_reward_keys,
                        **kwargs,
                    )
        self.init_qpos[:-6] *= 0 # Use fully open as init pos
        self.init_qpos[0] = -1.5 # place palm up


    def get_obs_dict(self, sim):
        obs_dict = {}
        obs_dict['time'] = np.array([sim.data.time])
        obs_dict['hand_jnt'] = sim.data.qpos[:-6].copy()
        obs_dict['obj_pos'] = sim.data.body_xpos[self.obj_bid].copy()
        obs_dict['obj_des_pos'] = sim.data.site_xpos[self.eps_ball_sid].ravel()
        obs_dict['obj_vel'] = sim.data.qvel[-6:].copy()*self.dt
        obs_dict['obj_rot'] = (sim.data.geom_xpos[self.obj_t_gid] - sim.data.geom_xpos[self.obj_b_gid])/self.pen_length
        obs_dict['obj_des_rot'] = (sim.data.geom_xpos[self.tar_t_gid] - sim.data.geom_xpos[self.tar_b_gid])/self.tar_length
        obs_dict['obj_err_pos'] = obs_dict['obj_pos']-obs_dict['obj_des_pos']
        obs_dict['obj_err_rot'] = obs_dict['obj_rot']-obs_dict['obj_des_rot']
        if sim.model.na>0:
            obs_dict['act'] = sim.data.act[:].copy()
            obs_dict['mlen'] = sim.data.actuator_length[:].copy()
            obs_dict['mvel'] = sim.data.actuator_velocity[:].copy()
            obs_dict['mforce'] = sim.data.actuator_force[:].copy()

        return obs_dict


    def get_reward_dict(self, obs_dict):
        pos_err = obs_dict['obj_err_pos']
        pos_align = np.linalg.norm(pos_err, axis=-1)
        rot_align = calculate_cosine(obs_dict['obj_rot'], obs_dict['obj_des_rot'])
        dropped = (pos_align > 0.075)
        act_mag = np.linalg.norm(self.obs_dict['act'], axis=-1)/self.sim.model.na if self.sim.model.na !=0 else 0
        rwd_dict = collections.OrderedDict((
            # Optional Keys
            ('pos_align',   -1.*pos_align),
            ('rot_align',   rot_align),
            ('act_reg',     -1.*act_mag),
            ('drop',        -1.*dropped),
            ('bonus',       1.*(rot_align > 0.9)*(pos_align<0.075) + 5.0*(rot_align > 0.95)*(pos_align<0.075) ),
            # Must keys
            ('sparse',      -1.0*pos_align+rot_align),
            ('solved',      (rot_align > 0.95)*(~dropped)),
            ('done',        dropped),
        ))
        rwd_dict['dense'] = np.sum([wt*rwd_dict[key] for key, wt in self.rwd_keys_wt.items()], axis=0)
        if list(self.sim.model.site_rgba[self.success_indicator_sid, :2]) != [0.0, 2.0]:
            self.sim.model.site_rgba[self.success_indicator_sid, :2] = np.array([0, 2]) if rwd_dict['solved'] else np.array([2, 0])
        return rwd_dict

class Geometries8EnvV0(ProprioceptiveEnvV0):
    def reset(self):
        ellips = {0: [[0.011, 0.025, 0.025], [0.74792, 0.35159, 0.80154, 1.0]], 1: [[0.019, 0.040, 0.040], [0.23366, 0.67864, 0.53721, 1.0]]}

        box = {0: [[0.017, 0.017, 0.017], [0.42829, 0.76091, 0.4914, 1.0]], 1: [[0.023, 0.023, 0.023], [0.21995, 0.60938, 0.18821, 1.0]]}

        caps = {0: [[0.013, 0.025, 0.025], [0.2574, 0.6124, 0.5413, 1.0]], 1: [[0.019, 0.040, 0.040], [0.6466, 0.5044, 0.4588, 1.0]]}

        cyl = {0: [[0.013, 0.025, 0.025], [0.1647, 0.3373, 0.3611, 1.0]], 1: [[0.019, 0.040, 0.040], [0.4554, 0.208, 0.8395, 1.0]]}

        # randomize target
        self.obj_gid = self.sim.model.geom_name2id('obj')
        self.tar_gid = self.sim.model.geom_name2id('target')

        self.obj_t_gid = self.sim.model.geom_name2id('top')
        self.obj_b_gid = self.sim.model.geom_name2id('bot')
        self.tar_t_gid = self.sim.model.geom_name2id('t_top')
        self.tar_b_gid = self.sim.model.geom_name2id('t_bot')

        geom_type = self.np_random.choice([3,4,5,6])

        if geom_type == 3:
            ind = self.np_random.choice(range(len(caps)))
            size = caps[ind][0]
            color = caps[ind][1]
            top_pos = np.array([0, 0, 1.3*size[1]])
            bot_pos = np.array([0, 0, 1.3*-size[1]])
        elif geom_type == 4:
            ind = self.np_random.choice(range(len(ellips)))
            size = ellips[ind][0]
            color = ellips[ind][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])
        elif geom_type == 5:
            ind = self.np_random.choice(range(len(cyl)))
            size = cyl[ind][0]
            color = cyl[ind][1]
            top_pos = np.array([0, 0, size[1]])
            bot_pos = np.array([0, 0, -size[1]])
        else:
            ind = self.np_random.choice(range(len(box)))
            size = box[ind][0]
            color = box[ind][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])

        color = [1, .9, 0, 1.0]

        self.sim.model.geom_size[self.obj_gid] = size
        self.sim.model.geom_type[self.obj_gid] = geom_type
        self.sim.model.geom_rgba[self.obj_gid] = color
        self.sim.model.geom_pos[self.obj_t_gid] = top_pos
        self.sim.model.geom_pos[self.obj_b_gid] = bot_pos

        self.sim.model.body_mass[self.obj_bid] = 1.2

        self.sim.model.geom_size[self.tar_gid] = size
        self.sim.model.geom_type[self.tar_gid] = geom_type
        self.sim.model.geom_rgba[self.tar_gid] = color
        self.sim.model.geom_pos[self.tar_t_gid] = top_pos
        self.sim.model.geom_pos[self.tar_b_gid] = bot_pos

        desired_orien = np.zeros(3)
        desired_orien[0] = self.np_random.uniform(low=-1, high=1)
        desired_orien[1] = self.np_random.uniform(low=-.8, high=1.2)

        self.sim.model.site_rgba[self.success_indicator_sid, :2] = np.array([2, 0])

        self.sim.model.body_quat[self.target_obj_bid] = euler2quat(desired_orien)
        self.robot.sync_sims(self.sim, self.sim_obsd)
        obs = super().reset()
        return obs

class Geometries100EnvV0(ProprioceptiveEnvV0):
    def reset(self):
        ellips = {0: [[0.02843, 0.0256, 0.02902], [0.74792, 0.35159, 0.80154, 1.0]], 1: [[0.01057, 0.02655, 0.0328], [0.23366, 0.67864, 0.53721, 1.0]], 2: [[0.01126, 0.0273, 0.04264], [0.65043, 0.2313, 0.50699, 1.0]], 3: [[0.02641, 0.03524, 0.02831], [0.62441, 0.5802, 0.43566, 1.0]], 4: [[0.02804, 0.03722, 0.04313], [0.13892, 0.45695, 0.11598, 1.0]], 5: [[0.02305, 0.04456, 0.03709], [0.13836, 0.67832, 0.31776, 1.0]], 6: [[0.02332, 0.02673, 0.02606], [0.15742, 0.40625, 0.11557, 1.0]], 7: [[0.01247, 0.03233, 0.03759], [0.83, 0.24511, 0.30415, 1.0]], 8: [[0.02199, 0.029, 0.04484], [0.23557, 0.72447, 0.75669, 1.0]], 9: [[0.02674, 0.0428, 0.03764], [0.8393, 0.75063, 0.18226, 1.0]], 10: [[0.02278, 0.04006, 0.03556], [0.32785, 0.49373, 0.5858, 1.0]], 11: [[0.02392, 0.04095, 0.03467], [0.8965, 0.22427, 0.41412, 1.0]], 12: [[0.01928, 0.0348, 0.03044], [0.20289, 0.70564, 0.55928, 1.0]], 13: [[0.02388, 0.03644, 0.02817], [0.67021, 0.70081, 0.36769, 1.0]], 14: [[0.02739, 0.04338, 0.03457], [0.28136, 0.77765, 0.28719, 1.0]], 15: [[0.00962, 0.04047, 0.02614], [0.49566, 0.72634, 0.52086, 1.0]], 16: [[0.0163, 0.04443, 0.04326], [0.1731, 0.8899, 0.10808, 1.0]], 17: [[0.02417, 0.03157, 0.04038], [0.21701, 0.29525, 0.62152, 1.0]], 18: [[0.01927, 0.02814, 0.03786], [0.76065, 0.49735, 0.27818, 1.0]], 19: [[0.02477, 0.04456, 0.04493], [0.80959, 0.8233, 0.5421, 1.0]], 20: [[0.01656, 0.0291, 0.03996], [0.33429, 0.36101, 0.28275, 1.0]], 21: [[0.01763, 0.03877, 0.03636], [0.21692, 0.27226, 0.65917, 1.0]], 22: [[0.01915, 0.0346, 0.04245], [0.41227, 0.43577, 0.31358, 1.0]], 23: [[0.02485, 0.03324, 0.02881], [0.65375, 0.75452, 0.47755, 1.0]], 24: [[0.00856, 0.04185, 0.03749], [0.44235, 0.15332, 0.76038, 1.0]]}

        box = {0: [[0.02295, 0.02306, 0.02221], [0.42829, 0.76091, 0.4914, 1.0]], 1: [[0.02447, 0.0185, 0.02192], [0.21995, 0.20938, 0.48821, 1.0]], 2: [[0.01853, 0.01837, 0.01546], [0.76916, 0.26635, 0.16801, 1.0]], 3: [[0.01586, 0.02079, 0.022], [0.1423, 0.75584, 0.15317, 1.0]], 4: [[0.02293, 0.02116, 0.02255], [0.2323, 0.3077, 0.42493, 1.0]], 5: [[0.01542, 0.01651, 0.02381], [0.26685, 0.38304, 0.60438, 1.0]], 6: [[0.0186, 0.02402, 0.02333], [0.1624, 0.45497, 0.14676, 1.0]], 7: [[0.01782, 0.01584, 0.02208], [0.17459, 0.54131, 0.68087, 1.0]], 8: [[0.01907, 0.0195, 0.02161], [0.54092, 0.40078, 0.68101, 1.0]], 9: [[0.01751, 0.0211, 0.01864], [0.65326, 0.59045, 0.30555, 1.0]], 10: [[0.02258, 0.02334, 0.01856], [0.80835, 0.76567, 0.63477, 1.0]], 11: [[0.02195, 0.01617, 0.02438], [0.86178, 0.17993, 0.61248, 1.0]], 12: [[0.01627, 0.02254, 0.02073], [0.39115, 0.68792, 0.78923, 1.0]], 13: [[0.02364, 0.01946, 0.01777], [0.57742, 0.55447, 0.48724, 1.0]], 14: [[0.01754, 0.02463, 0.01549], [0.73537, 0.1708, 0.49452, 1.0]], 15: [[0.02394, 0.02382, 0.02387], [0.13934, 0.18804, 0.67206, 1.0]], 16: [[0.01997, 0.02372, 0.02032], [0.55577, 0.62793, 0.42524, 1.0]], 17: [[0.01741, 0.02316, 0.02203], [0.8914, 0.54996, 0.31562, 1.0]], 18: [[0.02032, 0.0217, 0.02432], [0.38009, 0.75075, 0.64515, 1.0]], 19: [[0.01961, 0.0248, 0.0176], [0.5283, 0.86192, 0.77579, 1.0]], 20: [[0.01906, 0.01999, 0.02399], [0.49911, 0.89906, 0.44505, 1.0]], 21: [[0.02472, 0.01826, 0.0151], [0.79248, 0.49588, 0.41427, 1.0]], 22: [[0.01636, 0.0158, 0.01958], [0.57198, 0.58271, 0.78801, 1.0]], 23: [[0.01542, 0.02434, 0.02237], [0.22467, 0.88589, 0.83947, 1.0]], 24: [[0.01731, 0.02185, 0.02019], [0.77122, 0.73006, 0.14257, 1.0]]}

        caps = {0: [[0.0162, 0.0422, 0.0484], [0.2574, 0.6124, 0.5413, 1.0]], 1: [[0.016, 0.0457, 0.0496], [0.6466, 0.5044, 0.4588, 1.0]], 2: [[0.0187, 0.0259, 0.0248], [0.3565, 0.7914, 0.1228, 1.0]], 3: [[0.0192, 0.0483, 0.0216], [0.1584, 0.208, 0.8504, 1.0]], 4: [[0.0213, 0.0218, 0.0481], [0.649, 0.3251, 0.4248, 1.0]], 5: [[0.0169, 0.0331, 0.0388], [0.1653, 0.1763, 0.5914, 1.0]], 6: [[0.0138, 0.0299, 0.0471], [0.2764, 0.297, 0.3948, 1.0]], 7: [[0.0194, 0.0252, 0.0419], [0.7098, 0.8251, 0.6623, 1.0]], 8: [[0.014, 0.0362, 0.0201], [0.4234, 0.7236, 0.8917, 1.0]], 9: [[0.0125, 0.029, 0.0298], [0.5309, 0.397, 0.5795, 1.0]], 10: [[0.0162, 0.0396, 0.0323], [0.5987, 0.6183, 0.2126, 1.0]], 11: [[0.019, 0.0365, 0.0421], [0.1012, 0.1079, 0.4266, 1.0]], 12: [[0.0143, 0.0228, 0.0255], [0.6569, 0.5273, 0.6765, 1.0]], 13: [[0.0147, 0.0391, 0.0369], [0.2067, 0.8218, 0.5556, 1.0]], 14: [[0.0192, 0.0324, 0.043], [0.1487, 0.4226, 0.5643, 1.0]], 15: [[0.0145, 0.0491, 0.0234], [0.6435, 0.8165, 0.5704, 1.0]], 16: [[0.013, 0.0458, 0.0457], [0.3264, 0.1895, 0.3025, 1.0]], 17: [[0.0187, 0.0219, 0.0434], [0.7298, 0.3122, 0.5842, 1.0]], 18: [[0.0198, 0.0276, 0.0238], [0.6896, 0.247, 0.3097, 1.0]], 19: [[0.0175, 0.0375, 0.0339], [0.8754, 0.7424, 0.8578, 1.0]], 20: [[0.0191, 0.049, 0.0472], [0.7943, 0.1871, 0.4875, 1.0]], 21: [[0.0145, 0.0425, 0.0356], [0.2251, 0.2459, 0.8375, 1.0]], 22: [[0.0134, 0.0291, 0.0379], [0.4273, 0.7564, 0.3796, 1.0]], 23: [[0.0185, 0.0445, 0.0454], [0.7564, 0.293, 0.7573, 1.0]], 24: [[0.0164, 0.041, 0.0328], [0.1576, 0.8699, 0.1782, 1.0]]}

        cyl = {0: [[0.0118, 0.044, 0.0265], [0.1647, 0.3373, 0.3611, 1.0]], 1: [[0.0189, 0.0316, 0.0415], [0.4554, 0.208, 0.8395, 1.0]], 2: [[0.0123, 0.0364, 0.0238], [0.8701, 0.7737, 0.4562, 1.0]], 3: [[0.0145, 0.0362, 0.032], [0.3982, 0.7781, 0.8505, 1.0]], 4: [[0.0146, 0.0306, 0.0283], [0.8632, 0.8666, 0.1886, 1.0]], 5: [[0.0155, 0.0237, 0.0383], [0.488, 0.3793, 0.6675, 1.0]], 6: [[0.0198, 0.0323, 0.03], [0.7845, 0.2764, 0.8887, 1.0]], 7: [[0.011, 0.0368, 0.0343], [0.5305, 0.607, 0.651, 1.0]], 8: [[0.0197, 0.0305, 0.0206], [0.1644, 0.8826, 0.1977, 1.0]], 9: [[0.0162, 0.0277, 0.0329], [0.7852, 0.6644, 0.6417, 1.0]], 10: [[0.0202, 0.0339, 0.0223], [0.3046, 0.4812, 0.4465, 1.0]], 11: [[0.0181, 0.0204, 0.0314], [0.5126, 0.7779, 0.4888, 1.0]], 12: [[0.0118, 0.0374, 0.0228], [0.7434, 0.7433, 0.8634, 1.0]], 13: [[0.0197, 0.0319, 0.0375], [0.5661, 0.2352, 0.3933, 1.0]], 14: [[0.0133, 0.0392, 0.0448], [0.8689, 0.5574, 0.1234, 1.0]], 15: [[0.0157, 0.0236, 0.0301], [0.6846, 0.8593, 0.2673, 1.0]], 16: [[0.0178, 0.0351, 0.0414], [0.1313, 0.8603, 0.8791, 1.0]], 17: [[0.0123, 0.024, 0.0399], [0.3679, 0.6363, 0.4691, 1.0]], 18: [[0.0158, 0.0254, 0.022], [0.2256, 0.1914, 0.8973, 1.0]], 19: [[0.0192, 0.021, 0.0355], [0.1923, 0.8149, 0.21, 1.0]], 20: [[0.011, 0.0266, 0.0261], [0.7158, 0.2207, 0.4712, 1.0]], 21: [[0.0121, 0.0369, 0.0226], [0.4497, 0.7305, 0.5118, 1.0]], 22: [[0.012, 0.0221, 0.0254], [0.6128, 0.4154, 0.2382, 1.0]], 23: [[0.0213, 0.0298, 0.0297], [0.8346, 0.4971, 0.3623, 1.0]], 24: [[0.0164, 0.0302, 0.025], [0.2496, 0.1486, 0.8377, 1.0]]}

        # randomize target
        self.obj_gid = self.sim.model.geom_name2id('obj')
        self.tar_gid = self.sim.model.geom_name2id('target')

        self.obj_t_gid = self.sim.model.geom_name2id('top')
        self.obj_b_gid = self.sim.model.geom_name2id('bot')
        self.tar_t_gid = self.sim.model.geom_name2id('t_top')
        self.tar_b_gid = self.sim.model.geom_name2id('t_bot')

        geom_type = self.np_random.choice([3,4,5,6])

        if geom_type == 3:
            size = caps[self.np_random.choice(range(len(caps)))][0]
            color = caps[self.np_random.choice(range(len(caps)))][1]
            top_pos = np.array([0, 0, 1.3*size[1]])
            bot_pos = np.array([0, 0, 1.3*-size[1]])
        elif geom_type == 4:
            size = ellips[self.np_random.choice(range(len(ellips)))][0]
            color = ellips[self.np_random.choice(range(len(ellips)))][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])
        elif geom_type == 5:
            size = cyl[self.np_random.choice(range(len(cyl)))][0]
            color = cyl[self.np_random.choice(range(len(cyl)))][1]
            top_pos = np.array([0, 0, size[1]])
            bot_pos = np.array([0, 0, -size[1]])
        else:
            size = box[self.np_random.choice(range(len(box)))][0]
            color = box[self.np_random.choice(range(len(box)))][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])

        self.sim.model.geom_size[self.obj_gid] = size
        self.sim.model.geom_type[self.obj_gid] = geom_type
        self.sim.model.geom_rgba[self.obj_gid] = color
        self.sim.model.geom_pos[self.obj_t_gid] = top_pos
        self.sim.model.geom_pos[self.obj_b_gid] = bot_pos

        self.sim.model.body_mass[self.obj_bid] = 1.2

        self.sim.model.geom_size[self.tar_gid] = size
        self.sim.model.geom_type[self.tar_gid] = geom_type
        self.sim.model.geom_rgba[self.tar_gid] = color
        self.sim.model.geom_pos[self.tar_t_gid] = top_pos
        self.sim.model.geom_pos[self.tar_b_gid] = bot_pos

        self.sim.model.geom_condim[self.obj_gid] = 3
        desired_orien = np.zeros(3)
        desired_orien[0] = self.np_random.uniform(low=-1, high=1)
        desired_orien[1] = self.np_random.uniform(low=-.8, high=1.2)

        self.sim.model.site_rgba[self.success_indicator_sid, :2] = np.array([2, 0])


        self.sim.model.body_quat[self.target_obj_bid] = euler2quat(desired_orien)
        self.robot.sync_sims(self.sim, self.sim_obsd)
        obs = super().reset()
        return obs

class InDistribution(ProprioceptiveEnvV0):
    def reset(self):
        ellips = {0: [[0.0179, 0.0446, 0.0356], [0.34, 0.2627, 0.3952, 1.0]], 1: [[0.0218, 0.0327, 0.0237], [0.3313, 0.8439, 0.3636, 1.0]], 2: [[0.0144, 0.0335, 0.0322], [0.6041, 0.4344, 0.1575, 1.0]], 3: [[0.0171, 0.0285, 0.0388], [0.8304, 0.58, 0.2844, 1.0]], 4: [[0.0091, 0.0373, 0.0214], [0.204, 0.5209, 0.2448, 1.0]], 5: [[0.0216, 0.0294, 0.0408], [0.8079, 0.4409, 0.8124, 1.0]], 6: [[0.012, 0.0342, 0.0324], [0.7284, 0.2844, 0.874, 1.0]], 7: [[0.0213, 0.0369, 0.0343], [0.3488, 0.1383, 0.132, 1.0]], 8: [[0.0119, 0.0309, 0.0338], [0.3082, 0.8545, 0.8087, 1.0]], 9: [[0.0092, 0.0368, 0.0249], [0.312, 0.4161, 0.3201, 1.0]], 10: [[0.0111, 0.0324, 0.0333], [0.5433, 0.8573, 0.5046, 1.0]], 11: [[0.0084, 0.0233, 0.0431], [0.2799, 0.1479, 0.8274, 1.0]], 12: [[0.0217, 0.0343, 0.029], [0.7758, 0.6188, 0.765, 1.0]], 13: [[0.0157, 0.0311, 0.0283], [0.7821, 0.1809, 0.8041, 1.0]], 14: [[0.012, 0.0408, 0.0395], [0.1677, 0.4763, 0.1768, 1.0]], 15: [[0.0193, 0.0333, 0.0411], [0.2521, 0.1056, 0.6368, 1.0]], 16: [[0.0124, 0.02, 0.024], [0.164, 0.1243, 0.7993, 1.0]], 17: [[0.0188, 0.0432, 0.0255], [0.843, 0.1857, 0.489, 1.0]], 18: [[0.0184, 0.024, 0.0298], [0.1337, 0.5341, 0.6517, 1.0]], 19: [[0.0135, 0.038, 0.0371], [0.6872, 0.7654, 0.373, 1.0]], 20: [[0.0216, 0.0225, 0.0261], [0.4133, 0.5901, 0.7897, 1.0]], 21: [[0.0082, 0.0204, 0.0241], [0.7814, 0.2021, 0.4204, 1.0]], 22: [[0.0096, 0.0232, 0.0246], [0.8161, 0.5409, 0.2123, 1.0]], 23: [[0.0095, 0.025, 0.0274], [0.1395, 0.7302, 0.4261, 1.0]], 24: [[0.0167, 0.0359, 0.0212], [0.4984, 0.6301, 0.5516, 1.0]], 25: [[0.0143, 0.0313, 0.0443], [0.5803, 0.2667, 0.4889, 1.0]], 26: [[0.0122, 0.0261, 0.0388], [0.3106, 0.1305, 0.5166, 1.0]], 27: [[0.0084, 0.0441, 0.0342], [0.2434, 0.1669, 0.6743, 1.0]], 28: [[0.0174, 0.0218, 0.0244], [0.4817, 0.4323, 0.7243, 1.0]], 29: [[0.008, 0.0232, 0.0404], [0.1758, 0.4021, 0.512, 1.0]], 30: [[0.0095, 0.029, 0.0313], [0.7042, 0.1984, 0.366, 1.0]], 31: [[0.0173, 0.0277, 0.0276], [0.4663, 0.3589, 0.2309, 1.0]], 32: [[0.0127, 0.0344, 0.0359], [0.1129, 0.8764, 0.2991, 1.0]], 33: [[0.0178, 0.0392, 0.033], [0.6397, 0.6061, 0.3422, 1.0]], 34: [[0.0193, 0.0272, 0.0208], [0.3213, 0.4007, 0.1996, 1.0]], 35: [[0.0096, 0.0202, 0.0359], [0.8733, 0.5662, 0.7954, 1.0]], 36: [[0.0138, 0.0413, 0.0359], [0.4423, 0.1624, 0.4841, 1.0]], 37: [[0.0158, 0.0291, 0.0403], [0.3637, 0.7927, 0.7459, 1.0]], 38: [[0.0205, 0.0251, 0.0208], [0.5048, 0.5567, 0.3631, 1.0]], 39: [[0.0086, 0.028, 0.0206], [0.2369, 0.1793, 0.5048, 1.0]], 40: [[0.0156, 0.0332, 0.0322], [0.4873, 0.8613, 0.6544, 1.0]], 41: [[0.0197, 0.0437, 0.0261], [0.3981, 0.1823, 0.2919, 1.0]], 42: [[0.0093, 0.0348, 0.0416], [0.624, 0.1488, 0.4907, 1.0]], 43: [[0.0135, 0.0423, 0.0398], [0.1624, 0.107, 0.6028, 1.0]], 44: [[0.0195, 0.0285, 0.023], [0.4965, 0.1902, 0.3155, 1.0]], 45: [[0.0163, 0.0268, 0.0375], [0.566, 0.8469, 0.8336, 1.0]], 46: [[0.0197, 0.0245, 0.0399], [0.1049, 0.7715, 0.563, 1.0]], 47: [[0.0199, 0.0317, 0.0309], [0.6912, 0.7888, 0.2288, 1.0]], 48: [[0.0132, 0.0374, 0.0393], [0.1806, 0.6692, 0.4167, 1.0]], 49: [[0.0134, 0.0421, 0.0387], [0.5341, 0.6567, 0.8963, 1.0]], 50: [[0.0126, 0.0389, 0.0264], [0.5698, 0.302, 0.5201, 1.0]], 51: [[0.0161, 0.0281, 0.0355], [0.8565, 0.4137, 0.6611, 1.0]], 52: [[0.0164, 0.0381, 0.0341], [0.8276, 0.2435, 0.7939, 1.0]], 53: [[0.0208, 0.0343, 0.0435], [0.2087, 0.7211, 0.6089, 1.0]], 54: [[0.011, 0.0355, 0.0249], [0.4889, 0.2257, 0.6651, 1.0]], 55: [[0.0163, 0.0355, 0.0245], [0.5397, 0.842, 0.8906, 1.0]], 56: [[0.0198, 0.022, 0.0284], [0.2942, 0.1247, 0.4229, 1.0]], 57: [[0.0091, 0.0304, 0.0211], [0.4287, 0.1466, 0.395, 1.0]], 58: [[0.0161, 0.039, 0.0276], [0.6075, 0.8244, 0.3582, 1.0]], 59: [[0.0202, 0.0442, 0.0352], [0.391, 0.3324, 0.3166, 1.0]], 60: [[0.0089, 0.0271, 0.0323], [0.5451, 0.225, 0.7743, 1.0]], 61: [[0.0136, 0.0223, 0.039], [0.2951, 0.8605, 0.7955, 1.0]], 62: [[0.0113, 0.0215, 0.0378], [0.2177, 0.6716, 0.1103, 1.0]], 63: [[0.0188, 0.0303, 0.0332], [0.8879, 0.1028, 0.5333, 1.0]], 64: [[0.0194, 0.0234, 0.0225], [0.8584, 0.1146, 0.5777, 1.0]], 65: [[0.0104, 0.0234, 0.0416], [0.1862, 0.5018, 0.3055, 1.0]], 66: [[0.0209, 0.0286, 0.0217], [0.5629, 0.4863, 0.8173, 1.0]], 67: [[0.01, 0.043, 0.021], [0.4847, 0.8588, 0.1886, 1.0]], 68: [[0.0147, 0.0349, 0.0404], [0.3467, 0.7475, 0.7833, 1.0]], 69: [[0.0217, 0.0402, 0.0293], [0.8445, 0.6944, 0.1076, 1.0]], 70: [[0.0108, 0.024, 0.0401], [0.2511, 0.5804, 0.2071, 1.0]], 71: [[0.0177, 0.0214, 0.0294], [0.814, 0.2083, 0.1068, 1.0]], 72: [[0.008, 0.0422, 0.0436], [0.5886, 0.1532, 0.3145, 1.0]], 73: [[0.0118, 0.0405, 0.0418], [0.1081, 0.5158, 0.695, 1.0]], 74: [[0.0086, 0.0321, 0.0365], [0.5083, 0.8788, 0.6599, 1.0]], 75: [[0.0202, 0.0231, 0.0276], [0.2305, 0.617, 0.5928, 1.0]], 76: [[0.0172, 0.0412, 0.0292], [0.7684, 0.5089, 0.6511, 1.0]], 77: [[0.013, 0.0284, 0.0422], [0.3124, 0.6509, 0.4347, 1.0]], 78: [[0.0181, 0.03, 0.0305], [0.7501, 0.4577, 0.5444, 1.0]], 79: [[0.0187, 0.0223, 0.0329], [0.1123, 0.3249, 0.4961, 1.0]], 80: [[0.0085, 0.0378, 0.0427], [0.6627, 0.4468, 0.8872, 1.0]], 81: [[0.014, 0.0311, 0.0343], [0.7632, 0.1089, 0.4045, 1.0]], 82: [[0.0177, 0.0324, 0.0359], [0.8607, 0.153, 0.3992, 1.0]], 83: [[0.0198, 0.0429, 0.0274], [0.4244, 0.6196, 0.4516, 1.0]], 84: [[0.0217, 0.0341, 0.045], [0.6909, 0.1468, 0.7247, 1.0]], 85: [[0.0146, 0.0398, 0.0203], [0.3826, 0.3718, 0.3496, 1.0]], 86: [[0.0123, 0.0394, 0.0258], [0.6635, 0.705, 0.2369, 1.0]], 87: [[0.0178, 0.0222, 0.0214], [0.3096, 0.5513, 0.5841, 1.0]], 88: [[0.0214, 0.0302, 0.0364], [0.7606, 0.1013, 0.8378, 1.0]], 89: [[0.0176, 0.027, 0.032], [0.7617, 0.7469, 0.2241, 1.0]], 90: [[0.01, 0.0383, 0.03], [0.89, 0.3931, 0.1526, 1.0]], 91: [[0.0183, 0.0221, 0.0229], [0.7542, 0.3996, 0.5302, 1.0]], 92: [[0.0196, 0.0205, 0.0345], [0.2052, 0.4023, 0.3618, 1.0]], 93: [[0.0219, 0.0429, 0.026], [0.7436, 0.3471, 0.2587, 1.0]], 94: [[0.0176, 0.0381, 0.0371], [0.7742, 0.1907, 0.3718, 1.0]], 95: [[0.0128, 0.0343, 0.0428], [0.2232, 0.2526, 0.2038, 1.0]], 96: [[0.0157, 0.0206, 0.0299], [0.7561, 0.6327, 0.4422, 1.0]], 97: [[0.0218, 0.0431, 0.0449], [0.1949, 0.842, 0.5496, 1.0]], 98: [[0.0104, 0.0233, 0.0417], [0.5364, 0.5095, 0.7928, 1.0]], 99: [[0.0218, 0.0357, 0.0252], [0.1999, 0.4042, 0.3603, 1.0]], 100: [[0.0175, 0.0318, 0.0299], [0.2983, 0.2628, 0.5035, 1.0]], 101: [[0.0138, 0.0251, 0.0415], [0.5903, 0.784, 0.8442, 1.0]], 102: [[0.0144, 0.037, 0.029], [0.8542, 0.4365, 0.7302, 1.0]], 103: [[0.016, 0.0318, 0.0409], [0.6409, 0.1171, 0.2057, 1.0]], 104: [[0.0198, 0.0298, 0.0394], [0.1016, 0.4901, 0.7547, 1.0]], 105: [[0.0197, 0.0433, 0.0311], [0.1091, 0.4195, 0.6097, 1.0]], 106: [[0.0081, 0.0204, 0.0431], [0.8465, 0.7843, 0.5533, 1.0]], 107: [[0.0167, 0.0406, 0.0447], [0.6874, 0.8019, 0.3514, 1.0]], 108: [[0.0158, 0.0441, 0.025], [0.3182, 0.8791, 0.5031, 1.0]], 109: [[0.0171, 0.0426, 0.0444], [0.5549, 0.6867, 0.6405, 1.0]], 110: [[0.0128, 0.0214, 0.025], [0.4276, 0.4977, 0.6437, 1.0]], 111: [[0.0103, 0.0217, 0.0446], [0.539, 0.492, 0.6311, 1.0]], 112: [[0.0154, 0.0379, 0.045], [0.4229, 0.3995, 0.237, 1.0]], 113: [[0.0127, 0.0256, 0.0409], [0.1151, 0.6203, 0.5401, 1.0]], 114: [[0.0158, 0.0382, 0.0319], [0.5734, 0.3454, 0.7076, 1.0]], 115: [[0.0183, 0.0414, 0.043], [0.1459, 0.7983, 0.1768, 1.0]], 116: [[0.0209, 0.027, 0.0269], [0.8093, 0.2445, 0.6262, 1.0]], 117: [[0.0184, 0.0244, 0.0354], [0.5865, 0.1397, 0.1062, 1.0]], 118: [[0.009, 0.0375, 0.0361], [0.4566, 0.1599, 0.8376, 1.0]], 119: [[0.015, 0.0405, 0.0355], [0.1976, 0.5422, 0.1322, 1.0]], 120: [[0.0186, 0.0285, 0.0331], [0.1194, 0.6774, 0.8185, 1.0]], 121: [[0.0087, 0.0384, 0.0221], [0.204, 0.5089, 0.5555, 1.0]], 122: [[0.0198, 0.0201, 0.0248], [0.5485, 0.6, 0.8289, 1.0]], 123: [[0.0116, 0.0375, 0.0279], [0.3355, 0.1299, 0.2203, 1.0]], 124: [[0.0205, 0.0351, 0.0281], [0.8562, 0.5161, 0.3179, 1.0]], 125: [[0.0092, 0.0206, 0.0388], [0.8301, 0.7225, 0.644, 1.0]], 126: [[0.0094, 0.0394, 0.0274], [0.8076, 0.5332, 0.4689, 1.0]], 127: [[0.0209, 0.042, 0.0267], [0.189, 0.5581, 0.4131, 1.0]], 128: [[0.0081, 0.0368, 0.0257], [0.3423, 0.619, 0.279, 1.0]], 129: [[0.0131, 0.0395, 0.027], [0.3179, 0.3905, 0.7289, 1.0]], 130: [[0.0118, 0.044, 0.0324], [0.4052, 0.7151, 0.5217, 1.0]], 131: [[0.015, 0.0369, 0.0368], [0.6741, 0.689, 0.4182, 1.0]], 132: [[0.0194, 0.0301, 0.0436], [0.8841, 0.3257, 0.6886, 1.0]], 133: [[0.0206, 0.038, 0.026], [0.1673, 0.144, 0.7722, 1.0]], 134: [[0.0123, 0.0221, 0.0394], [0.4072, 0.5328, 0.1444, 1.0]], 135: [[0.015, 0.0331, 0.0268], [0.5753, 0.4666, 0.6661, 1.0]], 136: [[0.0158, 0.0224, 0.0308], [0.8457, 0.2743, 0.5386, 1.0]], 137: [[0.0203, 0.0203, 0.0206], [0.5707, 0.7995, 0.1143, 1.0]], 138: [[0.0108, 0.0444, 0.0302], [0.883, 0.2207, 0.7003, 1.0]], 139: [[0.0142, 0.0227, 0.0441], [0.5038, 0.1334, 0.4707, 1.0]], 140: [[0.0179, 0.0412, 0.0395], [0.1797, 0.8912, 0.7051, 1.0]], 141: [[0.022, 0.0277, 0.0245], [0.7404, 0.6499, 0.6006, 1.0]], 142: [[0.0151, 0.0384, 0.0247], [0.5983, 0.8277, 0.2456, 1.0]], 143: [[0.0158, 0.0318, 0.0285], [0.6195, 0.6399, 0.5961, 1.0]], 144: [[0.0157, 0.0389, 0.0347], [0.7169, 0.8966, 0.4938, 1.0]], 145: [[0.015, 0.0304, 0.0296], [0.4485, 0.3861, 0.6756, 1.0]], 146: [[0.0175, 0.0296, 0.0369], [0.3544, 0.5575, 0.8562, 1.0]], 147: [[0.0157, 0.0355, 0.0217], [0.7051, 0.5541, 0.2275, 1.0]], 148: [[0.0191, 0.0248, 0.041], [0.1171, 0.3022, 0.5573, 1.0]], 149: [[0.0095, 0.0394, 0.0327], [0.1485, 0.7252, 0.3156, 1.0]], 150: [[0.0092, 0.0369, 0.0416], [0.528, 0.8994, 0.1997, 1.0]], 151: [[0.0085, 0.0251, 0.0226], [0.159, 0.1596, 0.3521, 1.0]], 152: [[0.0132, 0.0381, 0.0248], [0.2871, 0.247, 0.4446, 1.0]], 153: [[0.0188, 0.0387, 0.0373], [0.3758, 0.1651, 0.2371, 1.0]], 154: [[0.0136, 0.0334, 0.0428], [0.3628, 0.6433, 0.8349, 1.0]], 155: [[0.0093, 0.0238, 0.0335], [0.6441, 0.4322, 0.1294, 1.0]], 156: [[0.0121, 0.0357, 0.031], [0.4834, 0.3508, 0.498, 1.0]], 157: [[0.0161, 0.0422, 0.0213], [0.4219, 0.6501, 0.1276, 1.0]], 158: [[0.0205, 0.0396, 0.0417], [0.1712, 0.7632, 0.5686, 1.0]], 159: [[0.0215, 0.03, 0.0323], [0.1548, 0.4639, 0.8891, 1.0]], 160: [[0.011, 0.0241, 0.0238], [0.3156, 0.6099, 0.358, 1.0]], 161: [[0.0172, 0.0426, 0.0292], [0.6982, 0.7319, 0.6483, 1.0]], 162: [[0.0206, 0.0229, 0.0336], [0.3875, 0.8807, 0.8088, 1.0]], 163: [[0.0213, 0.0221, 0.0433], [0.5799, 0.824, 0.7818, 1.0]], 164: [[0.0214, 0.0407, 0.0204], [0.6081, 0.4471, 0.2829, 1.0]], 165: [[0.0135, 0.0391, 0.0336], [0.145, 0.5886, 0.3284, 1.0]], 166: [[0.0083, 0.0323, 0.0298], [0.6364, 0.1223, 0.7308, 1.0]], 167: [[0.015, 0.0268, 0.0405], [0.8594, 0.5667, 0.6114, 1.0]], 168: [[0.0126, 0.0299, 0.0224], [0.1043, 0.3084, 0.2389, 1.0]], 169: [[0.019, 0.0387, 0.0295], [0.3984, 0.8755, 0.3889, 1.0]], 170: [[0.0135, 0.0333, 0.0278], [0.6523, 0.5869, 0.2271, 1.0]], 171: [[0.0145, 0.0427, 0.0284], [0.1204, 0.4681, 0.1915, 1.0]], 172: [[0.0215, 0.0331, 0.0201], [0.4475, 0.8037, 0.8816, 1.0]], 173: [[0.0164, 0.026, 0.0418], [0.4476, 0.7209, 0.5665, 1.0]], 174: [[0.0138, 0.0208, 0.0259], [0.4744, 0.2627, 0.8776, 1.0]], 175: [[0.0171, 0.0312, 0.0221], [0.3405, 0.2247, 0.4041, 1.0]], 176: [[0.0126, 0.0314, 0.0228], [0.3861, 0.6054, 0.2509, 1.0]], 177: [[0.0137, 0.0263, 0.0264], [0.5871, 0.4488, 0.1386, 1.0]], 178: [[0.0158, 0.0325, 0.0204], [0.4723, 0.7872, 0.3829, 1.0]], 179: [[0.018, 0.043, 0.0277], [0.1599, 0.818, 0.5206, 1.0]], 180: [[0.0085, 0.0264, 0.0422], [0.8142, 0.7123, 0.7632, 1.0]], 181: [[0.0219, 0.0211, 0.0207], [0.2742, 0.2027, 0.8496, 1.0]], 182: [[0.0156, 0.0212, 0.0356], [0.7043, 0.135, 0.2614, 1.0]], 183: [[0.0187, 0.0211, 0.0376], [0.8517, 0.4961, 0.2077, 1.0]], 184: [[0.0204, 0.0221, 0.0355], [0.5143, 0.5176, 0.6385, 1.0]], 185: [[0.0217, 0.0404, 0.0298], [0.6098, 0.2487, 0.8383, 1.0]], 186: [[0.0191, 0.043, 0.0282], [0.4476, 0.7411, 0.1112, 1.0]], 187: [[0.0159, 0.0293, 0.0377], [0.6904, 0.4926, 0.5994, 1.0]], 188: [[0.0091, 0.0262, 0.0417], [0.6368, 0.3511, 0.1941, 1.0]], 189: [[0.0093, 0.0297, 0.0244], [0.1454, 0.8709, 0.5172, 1.0]], 190: [[0.0183, 0.0413, 0.0418], [0.2906, 0.8729, 0.6184, 1.0]], 191: [[0.0089, 0.0218, 0.0381], [0.4896, 0.6824, 0.1759, 1.0]], 192: [[0.0112, 0.0434, 0.0427], [0.5026, 0.235, 0.7239, 1.0]], 193: [[0.0197, 0.0345, 0.0382], [0.5842, 0.7965, 0.4111, 1.0]], 194: [[0.0095, 0.0404, 0.0245], [0.1075, 0.8717, 0.2179, 1.0]], 195: [[0.0187, 0.0382, 0.043], [0.5261, 0.5464, 0.4336, 1.0]], 196: [[0.0162, 0.0373, 0.0389], [0.4062, 0.174, 0.2001, 1.0]], 197: [[0.0095, 0.0281, 0.0374], [0.7505, 0.4524, 0.8543, 1.0]], 198: [[0.0178, 0.0431, 0.0346], [0.7836, 0.507, 0.2946, 1.0]], 199: [[0.0086, 0.0206, 0.0255], [0.2519, 0.3512, 0.723, 1.0]], 200: [[0.0166, 0.0225, 0.0293], [0.3807, 0.7226, 0.5609, 1.0]], 201: [[0.0216, 0.0245, 0.0301], [0.1207, 0.4613, 0.7248, 1.0]], 202: [[0.0193, 0.0404, 0.0375], [0.2981, 0.8526, 0.134, 1.0]], 203: [[0.0138, 0.0241, 0.0389], [0.6076, 0.1337, 0.8627, 1.0]], 204: [[0.0113, 0.02, 0.0411], [0.1686, 0.7173, 0.1742, 1.0]], 205: [[0.0133, 0.0329, 0.0428], [0.1381, 0.5346, 0.7832, 1.0]], 206: [[0.022, 0.0331, 0.0301], [0.4032, 0.2363, 0.2722, 1.0]], 207: [[0.0195, 0.0306, 0.0281], [0.4538, 0.7098, 0.381, 1.0]], 208: [[0.0094, 0.0414, 0.042], [0.5582, 0.4801, 0.8864, 1.0]], 209: [[0.0085, 0.0286, 0.0346], [0.4346, 0.1496, 0.867, 1.0]], 210: [[0.0161, 0.0225, 0.0287], [0.6963, 0.4941, 0.4147, 1.0]], 211: [[0.0156, 0.0344, 0.0285], [0.7432, 0.6502, 0.7875, 1.0]], 212: [[0.0193, 0.0335, 0.0239], [0.7239, 0.8395, 0.6681, 1.0]], 213: [[0.0177, 0.021, 0.035], [0.2067, 0.115, 0.3646, 1.0]], 214: [[0.0212, 0.0276, 0.0447], [0.8377, 0.5035, 0.1801, 1.0]], 215: [[0.0217, 0.0431, 0.0297], [0.1662, 0.6649, 0.1192, 1.0]], 216: [[0.0161, 0.0381, 0.0243], [0.7918, 0.8644, 0.1653, 1.0]], 217: [[0.0144, 0.039, 0.0205], [0.2515, 0.4627, 0.1858, 1.0]], 218: [[0.0164, 0.0231, 0.0394], [0.8649, 0.6655, 0.6936, 1.0]], 219: [[0.021, 0.034, 0.0412], [0.6322, 0.3109, 0.1315, 1.0]], 220: [[0.0146, 0.0403, 0.0291], [0.4969, 0.6351, 0.7072, 1.0]], 221: [[0.0199, 0.0365, 0.0415], [0.1443, 0.7003, 0.3025, 1.0]], 222: [[0.0175, 0.0357, 0.0434], [0.8958, 0.1377, 0.5829, 1.0]], 223: [[0.0182, 0.0214, 0.0364], [0.7676, 0.4967, 0.6749, 1.0]], 224: [[0.0111, 0.0348, 0.0422], [0.1985, 0.4115, 0.2922, 1.0]], 225: [[0.0143, 0.0291, 0.0392], [0.1168, 0.8456, 0.4299, 1.0]], 226: [[0.013, 0.0323, 0.0267], [0.193, 0.2611, 0.5682, 1.0]], 227: [[0.0097, 0.0404, 0.0399], [0.486, 0.4811, 0.8595, 1.0]], 228: [[0.0163, 0.0266, 0.04], [0.364, 0.3567, 0.3217, 1.0]], 229: [[0.0207, 0.0286, 0.0214], [0.8925, 0.1563, 0.3139, 1.0]], 230: [[0.0097, 0.0375, 0.0255], [0.1164, 0.3525, 0.4542, 1.0]], 231: [[0.0119, 0.039, 0.0255], [0.3666, 0.7901, 0.4978, 1.0]], 232: [[0.0156, 0.0215, 0.0274], [0.8466, 0.1781, 0.6678, 1.0]], 233: [[0.0121, 0.0235, 0.0444], [0.6763, 0.3172, 0.3356, 1.0]], 234: [[0.0098, 0.0233, 0.0413], [0.8073, 0.5084, 0.8901, 1.0]], 235: [[0.0153, 0.0346, 0.0326], [0.5169, 0.5373, 0.6524, 1.0]], 236: [[0.0198, 0.0298, 0.0413], [0.7362, 0.3284, 0.1694, 1.0]], 237: [[0.0166, 0.0338, 0.0254], [0.5003, 0.1036, 0.675, 1.0]], 238: [[0.0195, 0.0284, 0.0218], [0.7115, 0.8238, 0.1391, 1.0]], 239: [[0.0207, 0.0352, 0.0397], [0.5007, 0.3275, 0.1935, 1.0]], 240: [[0.0121, 0.0337, 0.039], [0.4878, 0.7455, 0.3974, 1.0]], 241: [[0.0144, 0.0438, 0.0364], [0.669, 0.3398, 0.4691, 1.0]], 242: [[0.021, 0.0401, 0.0237], [0.655, 0.8149, 0.2127, 1.0]], 243: [[0.0146, 0.0413, 0.0303], [0.679, 0.5893, 0.4474, 1.0]], 244: [[0.0082, 0.0377, 0.0411], [0.688, 0.8363, 0.4124, 1.0]], 245: [[0.0172, 0.0439, 0.0397], [0.1834, 0.4685, 0.4009, 1.0]], 246: [[0.015, 0.0334, 0.0343], [0.3602, 0.6536, 0.5352, 1.0]], 247: [[0.0095, 0.0311, 0.0418], [0.3878, 0.3408, 0.1713, 1.0]], 248: [[0.0206, 0.0414, 0.0255], [0.7165, 0.2262, 0.6045, 1.0]], 249: [[0.0098, 0.0289, 0.0259], [0.8523, 0.179, 0.7258, 1.0]]}

        caps =  {0: [[0.0116, 0.0303, 0.0327], [0.3035, 0.8151, 0.1706, 1.0]], 1: [[0.0135, 0.0431, 0.027], [0.2176, 0.3692, 0.2129, 1.0]], 2: [[0.0211, 0.0425, 0.0369], [0.4259, 0.6632, 0.3458, 1.0]], 3: [[0.0157, 0.0271, 0.04], [0.2733, 0.7038, 0.8995, 1.0]], 4: [[0.0192, 0.0417, 0.0318], [0.4286, 0.4809, 0.8645, 1.0]], 5: [[0.0209, 0.0385, 0.0333], [0.7587, 0.4668, 0.4177, 1.0]], 6: [[0.0197, 0.0285, 0.0206], [0.2474, 0.6831, 0.4869, 1.0]], 7: [[0.0212, 0.0294, 0.0415], [0.5253, 0.4904, 0.8947, 1.0]], 8: [[0.0102, 0.0289, 0.0405], [0.2619, 0.1736, 0.7266, 1.0]], 9: [[0.0169, 0.0289, 0.0235], [0.3558, 0.3385, 0.149, 1.0]], 10: [[0.0196, 0.0215, 0.043], [0.1956, 0.2643, 0.4139, 1.0]], 11: [[0.0129, 0.0224, 0.0292], [0.1594, 0.8224, 0.1802, 1.0]], 12: [[0.0203, 0.0347, 0.0336], [0.1015, 0.1924, 0.7818, 1.0]], 13: [[0.0179, 0.0374, 0.0418], [0.4296, 0.2539, 0.5479, 1.0]], 14: [[0.021, 0.0438, 0.0232], [0.3144, 0.3437, 0.3675, 1.0]], 15: [[0.0175, 0.0292, 0.0202], [0.79, 0.7432, 0.2371, 1.0]], 16: [[0.0139, 0.0243, 0.0296], [0.6455, 0.715, 0.356, 1.0]], 17: [[0.0188, 0.0235, 0.0418], [0.8043, 0.6093, 0.609, 1.0]], 18: [[0.0141, 0.0402, 0.0338], [0.3568, 0.4025, 0.2161, 1.0]], 19: [[0.0198, 0.0225, 0.0278], [0.4036, 0.7152, 0.3897, 1.0]], 20: [[0.0162, 0.0363, 0.0384], [0.1466, 0.478, 0.6661, 1.0]], 21: [[0.021, 0.036, 0.0346], [0.7738, 0.8506, 0.8907, 1.0]], 22: [[0.0145, 0.0261, 0.0312], [0.2366, 0.1764, 0.8232, 1.0]], 23: [[0.0136, 0.0384, 0.0273], [0.6499, 0.2905, 0.4252, 1.0]], 24: [[0.0191, 0.0315, 0.0431], [0.7513, 0.8309, 0.1806, 1.0]], 25: [[0.0159, 0.0274, 0.0218], [0.2988, 0.6072, 0.7034, 1.0]], 26: [[0.0198, 0.0372, 0.0267], [0.5948, 0.4304, 0.3038, 1.0]], 27: [[0.0198, 0.0424, 0.0441], [0.7152, 0.7174, 0.3449, 1.0]], 28: [[0.0193, 0.0379, 0.0259], [0.6032, 0.3437, 0.8725, 1.0]], 29: [[0.0161, 0.0363, 0.0217], [0.2494, 0.595, 0.6631, 1.0]], 30: [[0.0167, 0.0403, 0.0245], [0.1006, 0.1567, 0.3473, 1.0]], 31: [[0.0173, 0.0217, 0.0409], [0.8376, 0.4187, 0.4746, 1.0]], 32: [[0.0162, 0.0391, 0.0438], [0.6416, 0.3594, 0.1724, 1.0]], 33: [[0.0142, 0.0385, 0.0401], [0.6182, 0.5726, 0.6438, 1.0]], 34: [[0.0184, 0.0333, 0.032], [0.8179, 0.4146, 0.531, 1.0]], 35: [[0.0194, 0.0427, 0.0374], [0.2303, 0.4759, 0.4294, 1.0]], 36: [[0.0102, 0.0207, 0.0367], [0.6047, 0.4501, 0.8338, 1.0]], 37: [[0.0152, 0.0282, 0.0407], [0.6631, 0.7613, 0.5563, 1.0]], 38: [[0.0199, 0.0262, 0.0347], [0.7692, 0.1686, 0.1261, 1.0]], 39: [[0.0185, 0.0245, 0.0259], [0.7763, 0.6939, 0.3496, 1.0]], 40: [[0.0179, 0.0324, 0.0317], [0.5161, 0.3995, 0.4301, 1.0]], 41: [[0.0199, 0.0379, 0.028], [0.4624, 0.6208, 0.8652, 1.0]], 42: [[0.0165, 0.0208, 0.0296], [0.201, 0.4577, 0.2458, 1.0]], 43: [[0.0113, 0.0264, 0.0277], [0.8419, 0.1653, 0.7582, 1.0]], 44: [[0.0101, 0.0212, 0.0321], [0.857, 0.8674, 0.8981, 1.0]], 45: [[0.0113, 0.042, 0.032], [0.4617, 0.3144, 0.5003, 1.0]], 46: [[0.0145, 0.0234, 0.0271], [0.566, 0.3929, 0.2595, 1.0]], 47: [[0.014, 0.0351, 0.0404], [0.2455, 0.3105, 0.6313, 1.0]], 48: [[0.0161, 0.0365, 0.0268], [0.6829, 0.483, 0.4624, 1.0]], 49: [[0.0174, 0.0222, 0.043], [0.1739, 0.7701, 0.5844, 1.0]], 50: [[0.0138, 0.0405, 0.0394], [0.4791, 0.582, 0.6131, 1.0]], 51: [[0.016, 0.0266, 0.0414], [0.219, 0.3952, 0.3512, 1.0]], 52: [[0.0193, 0.0374, 0.0285], [0.1733, 0.5253, 0.5952, 1.0]], 53: [[0.0195, 0.0305, 0.025], [0.2337, 0.2408, 0.2449, 1.0]], 54: [[0.0105, 0.0251, 0.0449], [0.3661, 0.8533, 0.2444, 1.0]], 55: [[0.0186, 0.041, 0.0409], [0.3865, 0.1818, 0.6864, 1.0]], 56: [[0.0192, 0.042, 0.0238], [0.5518, 0.7363, 0.1583, 1.0]], 57: [[0.0116, 0.0251, 0.0379], [0.2524, 0.8529, 0.3819, 1.0]], 58: [[0.021, 0.0371, 0.021], [0.1414, 0.384, 0.7155, 1.0]], 59: [[0.0213, 0.0265, 0.0395], [0.2749, 0.3127, 0.7909, 1.0]], 60: [[0.021, 0.0319, 0.0283], [0.5913, 0.6348, 0.5974, 1.0]], 61: [[0.0149, 0.0439, 0.0299], [0.8976, 0.4958, 0.1536, 1.0]], 62: [[0.0172, 0.0248, 0.0299], [0.1486, 0.2097, 0.2295, 1.0]], 63: [[0.0179, 0.0351, 0.0355], [0.8954, 0.4191, 0.8969, 1.0]], 64: [[0.0124, 0.0382, 0.0424], [0.5034, 0.4461, 0.1056, 1.0]], 65: [[0.0215, 0.0217, 0.0356], [0.5877, 0.1743, 0.4656, 1.0]], 66: [[0.0206, 0.0267, 0.0367], [0.855, 0.62, 0.1948, 1.0]], 67: [[0.0165, 0.0261, 0.0435], [0.4013, 0.3245, 0.5992, 1.0]], 68: [[0.0186, 0.0283, 0.0373], [0.2626, 0.2002, 0.8558, 1.0]], 69: [[0.0149, 0.0376, 0.0324], [0.3573, 0.691, 0.2245, 1.0]], 70: [[0.0177, 0.0345, 0.0291], [0.3173, 0.4388, 0.6166, 1.0]], 71: [[0.0161, 0.0401, 0.0311], [0.6692, 0.3539, 0.8887, 1.0]], 72: [[0.0137, 0.0416, 0.0388], [0.4142, 0.1037, 0.8764, 1.0]], 73: [[0.0214, 0.035, 0.0247], [0.2892, 0.4272, 0.5212, 1.0]], 74: [[0.0201, 0.0327, 0.0307], [0.4781, 0.1114, 0.8923, 1.0]], 75: [[0.0168, 0.0392, 0.0287], [0.4642, 0.8013, 0.873, 1.0]], 76: [[0.0154, 0.0246, 0.0297], [0.292, 0.4809, 0.3152, 1.0]], 77: [[0.02, 0.0412, 0.0386], [0.2382, 0.1954, 0.2127, 1.0]], 78: [[0.0115, 0.0286, 0.0446], [0.6675, 0.6493, 0.3174, 1.0]], 79: [[0.021, 0.0267, 0.0364], [0.4092, 0.7301, 0.4177, 1.0]], 80: [[0.0197, 0.036, 0.033], [0.7277, 0.137, 0.4312, 1.0]], 81: [[0.011, 0.0212, 0.024], [0.1985, 0.4726, 0.3116, 1.0]], 82: [[0.0148, 0.043, 0.0265], [0.8787, 0.3004, 0.3646, 1.0]], 83: [[0.0142, 0.0219, 0.0429], [0.3327, 0.3933, 0.1077, 1.0]], 84: [[0.0113, 0.0275, 0.0351], [0.12, 0.7265, 0.3532, 1.0]], 85: [[0.0105, 0.0441, 0.027], [0.1197, 0.6571, 0.298, 1.0]], 86: [[0.0104, 0.0272, 0.0389], [0.1791, 0.5694, 0.4313, 1.0]], 87: [[0.0169, 0.0287, 0.0275], [0.4656, 0.1723, 0.4079, 1.0]], 88: [[0.0182, 0.0292, 0.0267], [0.1662, 0.6236, 0.3404, 1.0]], 89: [[0.0158, 0.0357, 0.032], [0.8223, 0.2361, 0.3007, 1.0]], 90: [[0.0144, 0.0254, 0.0322], [0.4219, 0.7239, 0.5218, 1.0]], 91: [[0.0192, 0.0335, 0.0429], [0.8972, 0.4936, 0.1189, 1.0]], 92: [[0.0213, 0.0217, 0.0271], [0.4168, 0.6123, 0.1293, 1.0]], 93: [[0.0189, 0.0334, 0.0417], [0.5742, 0.5411, 0.1496, 1.0]], 94: [[0.0203, 0.0313, 0.035], [0.7883, 0.7032, 0.8112, 1.0]], 95: [[0.021, 0.0215, 0.0389], [0.7635, 0.167, 0.8432, 1.0]], 96: [[0.0194, 0.0237, 0.0261], [0.2485, 0.7071, 0.8598, 1.0]], 97: [[0.0177, 0.0341, 0.0236], [0.3404, 0.4298, 0.2811, 1.0]], 98: [[0.0141, 0.0237, 0.0399], [0.847, 0.2361, 0.6864, 1.0]], 99: [[0.0148, 0.0302, 0.0251], [0.7078, 0.2871, 0.125, 1.0]], 100: [[0.0201, 0.0372, 0.0237], [0.4901, 0.2259, 0.4513, 1.0]], 101: [[0.0135, 0.034, 0.0381], [0.626, 0.4381, 0.3036, 1.0]], 102: [[0.0115, 0.0209, 0.0369], [0.2645, 0.5309, 0.1583, 1.0]], 103: [[0.012, 0.0412, 0.0407], [0.4469, 0.6832, 0.5495, 1.0]], 104: [[0.0166, 0.0275, 0.0289], [0.7481, 0.4705, 0.2908, 1.0]], 105: [[0.0135, 0.0306, 0.0393], [0.2987, 0.5765, 0.8264, 1.0]], 106: [[0.0151, 0.0319, 0.028], [0.6073, 0.3117, 0.5346, 1.0]], 107: [[0.0111, 0.0299, 0.0264], [0.1762, 0.2882, 0.5015, 1.0]], 108: [[0.0121, 0.0274, 0.0258], [0.6427, 0.1698, 0.2819, 1.0]], 109: [[0.0191, 0.041, 0.0273], [0.8644, 0.7027, 0.4652, 1.0]], 110: [[0.0189, 0.0206, 0.0373], [0.6887, 0.1229, 0.7218, 1.0]], 111: [[0.01, 0.0426, 0.024], [0.8984, 0.5439, 0.8335, 1.0]], 112: [[0.0206, 0.0336, 0.0303], [0.6008, 0.8212, 0.6378, 1.0]], 113: [[0.0119, 0.0205, 0.0411], [0.1392, 0.3216, 0.5325, 1.0]], 114: [[0.0111, 0.0374, 0.0226], [0.497, 0.7804, 0.6547, 1.0]], 115: [[0.0184, 0.0374, 0.0253], [0.8878, 0.6867, 0.4106, 1.0]], 116: [[0.0186, 0.0382, 0.0427], [0.2857, 0.7416, 0.8699, 1.0]], 117: [[0.0153, 0.025, 0.0303], [0.8224, 0.7924, 0.7773, 1.0]], 118: [[0.0124, 0.0309, 0.0344], [0.7696, 0.7589, 0.4938, 1.0]], 119: [[0.016, 0.03, 0.035], [0.475, 0.3274, 0.6608, 1.0]], 120: [[0.0142, 0.0362, 0.025], [0.7899, 0.7263, 0.7698, 1.0]], 121: [[0.0202, 0.0213, 0.0249], [0.3437, 0.5714, 0.5224, 1.0]], 122: [[0.0114, 0.0344, 0.0268], [0.4585, 0.3573, 0.7211, 1.0]], 123: [[0.0107, 0.0202, 0.0315], [0.5748, 0.1965, 0.3126, 1.0]], 124: [[0.0114, 0.043, 0.0358], [0.762, 0.5276, 0.3324, 1.0]], 125: [[0.0187, 0.0345, 0.0342], [0.7973, 0.2488, 0.4734, 1.0]], 126: [[0.0157, 0.0442, 0.0308], [0.623, 0.1523, 0.4676, 1.0]], 127: [[0.017, 0.0314, 0.0259], [0.4546, 0.1244, 0.8692, 1.0]], 128: [[0.0152, 0.0292, 0.0205], [0.8583, 0.186, 0.6164, 1.0]], 129: [[0.0217, 0.0387, 0.0409], [0.7588, 0.2858, 0.4066, 1.0]], 130: [[0.0151, 0.0213, 0.0275], [0.1535, 0.1794, 0.2188, 1.0]], 131: [[0.0165, 0.0202, 0.0226], [0.6488, 0.5128, 0.7284, 1.0]], 132: [[0.0198, 0.0306, 0.0344], [0.8172, 0.1524, 0.5647, 1.0]], 133: [[0.012, 0.0202, 0.0431], [0.3239, 0.2815, 0.1753, 1.0]], 134: [[0.014, 0.0407, 0.029], [0.3331, 0.5083, 0.1023, 1.0]], 135: [[0.0129, 0.039, 0.0278], [0.8779, 0.8324, 0.5216, 1.0]], 136: [[0.0211, 0.0407, 0.033], [0.3116, 0.2184, 0.1684, 1.0]], 137: [[0.0215, 0.0237, 0.0446], [0.4445, 0.7085, 0.321, 1.0]], 138: [[0.021, 0.0431, 0.0259], [0.2011, 0.8037, 0.248, 1.0]], 139: [[0.018, 0.0385, 0.0354], [0.4092, 0.3509, 0.3809, 1.0]], 140: [[0.0192, 0.0435, 0.0387], [0.5259, 0.6084, 0.2877, 1.0]], 141: [[0.0146, 0.0442, 0.0437], [0.8129, 0.5661, 0.2454, 1.0]], 142: [[0.0122, 0.0318, 0.0204], [0.1138, 0.2666, 0.6396, 1.0]], 143: [[0.0149, 0.025, 0.0414], [0.8627, 0.1379, 0.792, 1.0]], 144: [[0.019, 0.0394, 0.0342], [0.8871, 0.7036, 0.7611, 1.0]], 145: [[0.013, 0.0405, 0.0222], [0.6197, 0.1222, 0.7248, 1.0]], 146: [[0.0118, 0.0286, 0.0435], [0.245, 0.8639, 0.4145, 1.0]], 147: [[0.0215, 0.0257, 0.0278], [0.5146, 0.4355, 0.1102, 1.0]], 148: [[0.0168, 0.0289, 0.0336], [0.5404, 0.4091, 0.4605, 1.0]], 149: [[0.0183, 0.0321, 0.0219], [0.3991, 0.5365, 0.1933, 1.0]], 150: [[0.0136, 0.0414, 0.0226], [0.2694, 0.6687, 0.1661, 1.0]], 151: [[0.0201, 0.0408, 0.0352], [0.8922, 0.5367, 0.3726, 1.0]], 152: [[0.0213, 0.0407, 0.0402], [0.6644, 0.8317, 0.2544, 1.0]], 153: [[0.0147, 0.0428, 0.0258], [0.8475, 0.6459, 0.42, 1.0]], 154: [[0.0127, 0.0217, 0.0264], [0.3133, 0.1272, 0.2741, 1.0]], 155: [[0.0176, 0.0326, 0.0372], [0.1748, 0.2515, 0.7733, 1.0]], 156: [[0.0206, 0.029, 0.0234], [0.5347, 0.6689, 0.5907, 1.0]], 157: [[0.02, 0.0377, 0.0282], [0.7088, 0.7381, 0.124, 1.0]], 158: [[0.0125, 0.0261, 0.0223], [0.5638, 0.1319, 0.2414, 1.0]], 159: [[0.0161, 0.0206, 0.0384], [0.7087, 0.8315, 0.1975, 1.0]], 160: [[0.0176, 0.0311, 0.0226], [0.2817, 0.6628, 0.1593, 1.0]], 161: [[0.0152, 0.0399, 0.0434], [0.7062, 0.806, 0.7353, 1.0]], 162: [[0.0126, 0.0419, 0.0367], [0.1241, 0.4649, 0.6977, 1.0]], 163: [[0.021, 0.0294, 0.0446], [0.274, 0.4746, 0.8286, 1.0]], 164: [[0.0166, 0.025, 0.0284], [0.4418, 0.4761, 0.4187, 1.0]], 165: [[0.0148, 0.0323, 0.0439], [0.1281, 0.2521, 0.2884, 1.0]], 166: [[0.0166, 0.037, 0.0249], [0.8255, 0.1862, 0.2677, 1.0]], 167: [[0.0158, 0.0344, 0.0354], [0.8431, 0.4058, 0.7443, 1.0]], 168: [[0.0161, 0.0433, 0.0431], [0.4497, 0.8796, 0.6064, 1.0]], 169: [[0.012, 0.0246, 0.0357], [0.8279, 0.5592, 0.5058, 1.0]], 170: [[0.012, 0.0328, 0.0301], [0.4054, 0.2396, 0.8119, 1.0]], 171: [[0.0207, 0.0285, 0.0449], [0.2582, 0.7108, 0.6562, 1.0]], 172: [[0.0198, 0.0407, 0.0406], [0.7306, 0.8985, 0.2908, 1.0]], 173: [[0.0116, 0.0432, 0.0367], [0.8991, 0.3089, 0.3308, 1.0]], 174: [[0.0176, 0.0316, 0.0209], [0.2477, 0.3467, 0.8524, 1.0]], 175: [[0.0165, 0.0304, 0.0219], [0.2946, 0.2616, 0.1959, 1.0]], 176: [[0.021, 0.0418, 0.0226], [0.3537, 0.7351, 0.877, 1.0]], 177: [[0.0172, 0.0448, 0.0285], [0.4011, 0.2376, 0.5996, 1.0]], 178: [[0.0204, 0.0305, 0.0347], [0.1494, 0.3643, 0.4896, 1.0]], 179: [[0.0174, 0.0297, 0.0284], [0.1526, 0.1234, 0.3979, 1.0]], 180: [[0.0156, 0.0356, 0.0265], [0.7139, 0.5448, 0.4136, 1.0]], 181: [[0.0142, 0.0307, 0.0289], [0.8888, 0.1212, 0.5144, 1.0]], 182: [[0.0146, 0.0374, 0.0376], [0.299, 0.6482, 0.2818, 1.0]], 183: [[0.0188, 0.035, 0.0298], [0.5388, 0.6216, 0.5589, 1.0]], 184: [[0.0139, 0.0411, 0.0392], [0.6896, 0.2936, 0.1024, 1.0]], 185: [[0.0156, 0.043, 0.0221], [0.7578, 0.1587, 0.4894, 1.0]], 186: [[0.0176, 0.0349, 0.0379], [0.8278, 0.2228, 0.4548, 1.0]], 187: [[0.016, 0.0359, 0.0345], [0.8741, 0.7924, 0.2308, 1.0]], 188: [[0.0152, 0.0271, 0.0202], [0.3433, 0.3167, 0.6326, 1.0]], 189: [[0.0115, 0.0368, 0.0403], [0.7768, 0.7014, 0.4184, 1.0]], 190: [[0.0165, 0.0352, 0.0243], [0.603, 0.6831, 0.155, 1.0]], 191: [[0.0172, 0.0436, 0.0227], [0.8262, 0.7358, 0.3365, 1.0]], 192: [[0.0124, 0.0269, 0.0377], [0.58, 0.3878, 0.7276, 1.0]], 193: [[0.0159, 0.0288, 0.0428], [0.6612, 0.2328, 0.3988, 1.0]], 194: [[0.0137, 0.0243, 0.0409], [0.7058, 0.7323, 0.8918, 1.0]], 195: [[0.0168, 0.0399, 0.0405], [0.3749, 0.891, 0.4413, 1.0]], 196: [[0.0171, 0.0282, 0.0323], [0.8426, 0.2891, 0.3984, 1.0]], 197: [[0.0165, 0.0402, 0.0215], [0.3392, 0.2928, 0.7445, 1.0]], 198: [[0.0114, 0.0256, 0.0442], [0.1266, 0.1478, 0.4688, 1.0]], 199: [[0.0202, 0.0225, 0.02], [0.7923, 0.8564, 0.5706, 1.0]], 200: [[0.0207, 0.0428, 0.0432], [0.8461, 0.395, 0.4055, 1.0]], 201: [[0.0165, 0.0201, 0.044], [0.1453, 0.4584, 0.6714, 1.0]], 202: [[0.0188, 0.0261, 0.0379], [0.3333, 0.1802, 0.444, 1.0]], 203: [[0.0183, 0.0285, 0.0449], [0.1724, 0.1512, 0.702, 1.0]], 204: [[0.0108, 0.0317, 0.023], [0.2897, 0.4708, 0.7926, 1.0]], 205: [[0.0165, 0.0332, 0.0368], [0.6661, 0.4885, 0.5519, 1.0]], 206: [[0.0151, 0.0255, 0.0348], [0.3782, 0.1494, 0.2391, 1.0]], 207: [[0.0164, 0.0208, 0.0212], [0.5459, 0.8448, 0.7581, 1.0]], 208: [[0.0125, 0.0423, 0.0228], [0.5731, 0.2101, 0.7136, 1.0]], 209: [[0.0157, 0.0434, 0.0228], [0.5913, 0.7602, 0.1991, 1.0]], 210: [[0.0102, 0.0383, 0.02], [0.4456, 0.3827, 0.7466, 1.0]], 211: [[0.0171, 0.0269, 0.0221], [0.148, 0.329, 0.2079, 1.0]], 212: [[0.0179, 0.0273, 0.0309], [0.8428, 0.7415, 0.6008, 1.0]], 213: [[0.014, 0.0265, 0.02], [0.4166, 0.8607, 0.3437, 1.0]], 214: [[0.0159, 0.0424, 0.0282], [0.2146, 0.6623, 0.7518, 1.0]], 215: [[0.0113, 0.0248, 0.0214], [0.2293, 0.8791, 0.8644, 1.0]], 216: [[0.0106, 0.0358, 0.0437], [0.8324, 0.4802, 0.4808, 1.0]], 217: [[0.0147, 0.04, 0.0201], [0.5684, 0.1587, 0.8058, 1.0]], 218: [[0.0204, 0.0264, 0.0332], [0.586, 0.546, 0.8121, 1.0]], 219: [[0.0202, 0.0444, 0.0354], [0.1814, 0.1217, 0.8219, 1.0]], 220: [[0.0202, 0.027, 0.0394], [0.2476, 0.165, 0.5762, 1.0]], 221: [[0.0213, 0.0419, 0.0288], [0.891, 0.7265, 0.3415, 1.0]], 222: [[0.018, 0.0265, 0.0213], [0.4445, 0.7071, 0.6048, 1.0]], 223: [[0.0133, 0.04, 0.0208], [0.2507, 0.2076, 0.5588, 1.0]], 224: [[0.0137, 0.0312, 0.0411], [0.7524, 0.5718, 0.6326, 1.0]], 225: [[0.019, 0.0398, 0.0295], [0.2345, 0.1574, 0.8824, 1.0]], 226: [[0.0199, 0.0424, 0.0407], [0.262, 0.4326, 0.2912, 1.0]], 227: [[0.0159, 0.0388, 0.0319], [0.8816, 0.1561, 0.8673, 1.0]], 228: [[0.0144, 0.0436, 0.0302], [0.6982, 0.1408, 0.122, 1.0]], 229: [[0.0171, 0.0229, 0.0331], [0.6688, 0.6537, 0.8223, 1.0]], 230: [[0.019, 0.035, 0.041], [0.2898, 0.2755, 0.3883, 1.0]], 231: [[0.0106, 0.0209, 0.0318], [0.3899, 0.4914, 0.4955, 1.0]], 232: [[0.0117, 0.0268, 0.0275], [0.5498, 0.6662, 0.236, 1.0]], 233: [[0.0171, 0.0338, 0.0323], [0.7855, 0.3418, 0.8884, 1.0]], 234: [[0.0106, 0.0322, 0.0206], [0.1908, 0.2949, 0.1147, 1.0]], 235: [[0.0104, 0.0408, 0.0394], [0.5442, 0.8103, 0.8033, 1.0]], 236: [[0.0146, 0.0374, 0.0234], [0.1747, 0.6012, 0.1749, 1.0]], 237: [[0.012, 0.0379, 0.0339], [0.3885, 0.4516, 0.2483, 1.0]], 238: [[0.0158, 0.0327, 0.0382], [0.5894, 0.7818, 0.1015, 1.0]], 239: [[0.0117, 0.0289, 0.0422], [0.4988, 0.4957, 0.278, 1.0]], 240: [[0.0114, 0.0378, 0.0282], [0.1492, 0.6123, 0.5169, 1.0]], 241: [[0.0129, 0.0259, 0.0323], [0.6051, 0.2834, 0.5271, 1.0]], 242: [[0.0157, 0.0404, 0.0397], [0.2497, 0.1461, 0.7958, 1.0]], 243: [[0.0113, 0.0409, 0.027], [0.5513, 0.1007, 0.6764, 1.0]], 244: [[0.0124, 0.0448, 0.0357], [0.2974, 0.878, 0.1887, 1.0]], 245: [[0.0105, 0.0279, 0.0406], [0.6258, 0.2195, 0.5361, 1.0]], 246: [[0.0159, 0.0381, 0.029], [0.3421, 0.3551, 0.2471, 1.0]], 247: [[0.022, 0.0368, 0.0404], [0.5323, 0.849, 0.3699, 1.0]], 248: [[0.0183, 0.0261, 0.0257], [0.3421, 0.3559, 0.4682, 1.0]], 249: [[0.018, 0.0308, 0.0279], [0.7701, 0.5332, 0.2324, 1.0]]}

        cyl = {0: [[0.0176, 0.0288, 0.0437], [0.1955, 0.8692, 0.1558, 1.0]], 1: [[0.0129, 0.0411, 0.0246], [0.5346, 0.4588, 0.3807, 1.0]], 2: [[0.0111, 0.0316, 0.027], [0.2722, 0.5271, 0.4063, 1.0]], 3: [[0.0164, 0.03, 0.0375], [0.8139, 0.4018, 0.3405, 1.0]], 4: [[0.0182, 0.0376, 0.0361], [0.4298, 0.654, 0.4991, 1.0]], 5: [[0.0183, 0.0436, 0.0389], [0.4267, 0.8494, 0.1413, 1.0]], 6: [[0.0151, 0.0226, 0.0325], [0.3963, 0.7956, 0.2608, 1.0]], 7: [[0.0163, 0.0401, 0.0378], [0.7232, 0.149, 0.2906, 1.0]], 8: [[0.0149, 0.0332, 0.0223], [0.8633, 0.3784, 0.8556, 1.0]], 9: [[0.0164, 0.0276, 0.0271], [0.6687, 0.6546, 0.5891, 1.0]], 10: [[0.016, 0.0411, 0.0241], [0.4137, 0.8747, 0.5776, 1.0]], 11: [[0.0205, 0.0272, 0.0208], [0.3739, 0.1976, 0.4307, 1.0]], 12: [[0.0137, 0.0213, 0.0314], [0.6686, 0.2704, 0.5239, 1.0]], 13: [[0.0121, 0.0208, 0.043], [0.7621, 0.7547, 0.6583, 1.0]], 14: [[0.0216, 0.0431, 0.042], [0.546, 0.3163, 0.5418, 1.0]], 15: [[0.0124, 0.0359, 0.0223], [0.8796, 0.6622, 0.6448, 1.0]], 16: [[0.012, 0.0317, 0.0355], [0.6591, 0.8639, 0.4593, 1.0]], 17: [[0.017, 0.0208, 0.0262], [0.3882, 0.3656, 0.7262, 1.0]], 18: [[0.0101, 0.039, 0.0385], [0.3666, 0.4205, 0.2599, 1.0]], 19: [[0.0175, 0.0295, 0.0365], [0.2311, 0.2364, 0.6033, 1.0]], 20: [[0.013, 0.0328, 0.0245], [0.6305, 0.6234, 0.8997, 1.0]], 21: [[0.0121, 0.0347, 0.0431], [0.212, 0.1089, 0.7685, 1.0]], 22: [[0.011, 0.0401, 0.0347], [0.4657, 0.6878, 0.4088, 1.0]], 23: [[0.0177, 0.0441, 0.0424], [0.7711, 0.8509, 0.5582, 1.0]], 24: [[0.0141, 0.0439, 0.0423], [0.1046, 0.2489, 0.5168, 1.0]], 25: [[0.0126, 0.0253, 0.0298], [0.8599, 0.4669, 0.6264, 1.0]], 26: [[0.0157, 0.0289, 0.0248], [0.7242, 0.3746, 0.1595, 1.0]], 27: [[0.0202, 0.0395, 0.0414], [0.1041, 0.5994, 0.7891, 1.0]], 28: [[0.0192, 0.0275, 0.0383], [0.1872, 0.5261, 0.8864, 1.0]], 29: [[0.016, 0.0316, 0.041], [0.6291, 0.29, 0.1883, 1.0]], 30: [[0.0156, 0.0338, 0.0396], [0.2847, 0.3636, 0.4507, 1.0]], 31: [[0.0188, 0.0365, 0.0324], [0.269, 0.1722, 0.8891, 1.0]], 32: [[0.0119, 0.0339, 0.023], [0.1936, 0.111, 0.5617, 1.0]], 33: [[0.0153, 0.0414, 0.0214], [0.3801, 0.6898, 0.3237, 1.0]], 34: [[0.0152, 0.0431, 0.0367], [0.4103, 0.1971, 0.116, 1.0]], 35: [[0.0124, 0.0386, 0.0205], [0.2076, 0.6064, 0.1946, 1.0]], 36: [[0.0115, 0.0403, 0.026], [0.6847, 0.8077, 0.4212, 1.0]], 37: [[0.0176, 0.0225, 0.0339], [0.1208, 0.3116, 0.4218, 1.0]], 38: [[0.0137, 0.0423, 0.0405], [0.671, 0.1136, 0.5476, 1.0]], 39: [[0.0212, 0.0359, 0.042], [0.5916, 0.6113, 0.5848, 1.0]], 40: [[0.0208, 0.0382, 0.0202], [0.7364, 0.443, 0.6715, 1.0]], 41: [[0.0196, 0.0301, 0.0303], [0.3232, 0.8992, 0.3064, 1.0]], 42: [[0.0183, 0.0223, 0.0359], [0.8289, 0.8171, 0.3384, 1.0]], 43: [[0.0156, 0.029, 0.0264], [0.7746, 0.4009, 0.6431, 1.0]], 44: [[0.0177, 0.0253, 0.0348], [0.1924, 0.6394, 0.3969, 1.0]], 45: [[0.0135, 0.0223, 0.0447], [0.5147, 0.3683, 0.5374, 1.0]], 46: [[0.0171, 0.0241, 0.0237], [0.8721, 0.7066, 0.3328, 1.0]], 47: [[0.0191, 0.0391, 0.026], [0.4021, 0.221, 0.1984, 1.0]], 48: [[0.0105, 0.0392, 0.0395], [0.4277, 0.1513, 0.864, 1.0]], 49: [[0.0174, 0.028, 0.0254], [0.48, 0.5795, 0.5829, 1.0]], 50: [[0.0157, 0.0321, 0.0306], [0.5567, 0.4042, 0.1932, 1.0]], 51: [[0.0156, 0.0249, 0.0244], [0.72, 0.1787, 0.3869, 1.0]], 52: [[0.014, 0.0367, 0.0352], [0.5161, 0.8056, 0.7292, 1.0]], 53: [[0.0165, 0.033, 0.0264], [0.7106, 0.4317, 0.1531, 1.0]], 54: [[0.0191, 0.0388, 0.0425], [0.148, 0.6173, 0.4453, 1.0]], 55: [[0.0178, 0.0371, 0.0443], [0.2542, 0.6349, 0.2528, 1.0]], 56: [[0.0154, 0.0322, 0.0273], [0.1614, 0.8514, 0.7885, 1.0]], 57: [[0.0217, 0.0371, 0.0434], [0.7396, 0.5852, 0.5149, 1.0]], 58: [[0.0208, 0.0312, 0.0283], [0.2846, 0.5234, 0.8809, 1.0]], 59: [[0.0175, 0.039, 0.0376], [0.835, 0.4662, 0.1119, 1.0]], 60: [[0.02, 0.0257, 0.0404], [0.4791, 0.7725, 0.582, 1.0]], 61: [[0.0215, 0.035, 0.0417], [0.6326, 0.2573, 0.3743, 1.0]], 62: [[0.0212, 0.0204, 0.0248], [0.7863, 0.65, 0.2481, 1.0]], 63: [[0.0179, 0.0429, 0.0362], [0.6587, 0.4708, 0.1628, 1.0]], 64: [[0.0161, 0.0305, 0.0221], [0.3585, 0.6368, 0.1015, 1.0]], 65: [[0.0132, 0.027, 0.0253], [0.6352, 0.2588, 0.4852, 1.0]], 66: [[0.0177, 0.0288, 0.0347], [0.3313, 0.36, 0.7129, 1.0]], 67: [[0.0138, 0.0418, 0.0213], [0.1237, 0.1154, 0.3153, 1.0]], 68: [[0.0167, 0.0424, 0.0415], [0.8498, 0.2445, 0.3854, 1.0]], 69: [[0.0169, 0.0275, 0.0385], [0.5836, 0.7205, 0.3492, 1.0]], 70: [[0.0155, 0.029, 0.0368], [0.5096, 0.4709, 0.5227, 1.0]], 71: [[0.0134, 0.0219, 0.0387], [0.798, 0.7968, 0.7894, 1.0]], 72: [[0.02, 0.0218, 0.0369], [0.3228, 0.1577, 0.7163, 1.0]], 73: [[0.018, 0.0394, 0.039], [0.6086, 0.8168, 0.7293, 1.0]], 74: [[0.0123, 0.0368, 0.0263], [0.3403, 0.6987, 0.54, 1.0]], 75: [[0.0193, 0.0227, 0.0383], [0.4518, 0.4197, 0.2259, 1.0]], 76: [[0.0123, 0.0411, 0.0221], [0.1367, 0.4653, 0.104, 1.0]], 77: [[0.0121, 0.0256, 0.0304], [0.7589, 0.8021, 0.852, 1.0]], 78: [[0.0157, 0.0414, 0.0222], [0.3245, 0.3431, 0.7576, 1.0]], 79: [[0.0215, 0.0211, 0.0293], [0.2951, 0.4129, 0.2645, 1.0]], 80: [[0.0161, 0.027, 0.0225], [0.7282, 0.7775, 0.6572, 1.0]], 81: [[0.0216, 0.0256, 0.038], [0.643, 0.5394, 0.2942, 1.0]], 82: [[0.0137, 0.0294, 0.0262], [0.7038, 0.5798, 0.6859, 1.0]], 83: [[0.0218, 0.0307, 0.0313], [0.43, 0.3723, 0.1487, 1.0]], 84: [[0.0146, 0.0202, 0.0423], [0.4138, 0.2301, 0.6285, 1.0]], 85: [[0.012, 0.0379, 0.039], [0.597, 0.8637, 0.6304, 1.0]], 86: [[0.0219, 0.0355, 0.0426], [0.7368, 0.6725, 0.6755, 1.0]], 87: [[0.0163, 0.0318, 0.0441], [0.7194, 0.8389, 0.7171, 1.0]], 88: [[0.015, 0.043, 0.0269], [0.692, 0.1812, 0.7279, 1.0]], 89: [[0.0157, 0.0412, 0.042], [0.8807, 0.3385, 0.791, 1.0]], 90: [[0.0122, 0.042, 0.0432], [0.8933, 0.4339, 0.4224, 1.0]], 91: [[0.011, 0.0448, 0.0282], [0.8607, 0.4587, 0.5929, 1.0]], 92: [[0.0181, 0.0354, 0.0294], [0.5487, 0.5111, 0.5091, 1.0]], 93: [[0.0139, 0.0339, 0.0448], [0.264, 0.6477, 0.7418, 1.0]], 94: [[0.0111, 0.0365, 0.0442], [0.842, 0.4455, 0.2316, 1.0]], 95: [[0.0122, 0.032, 0.0339], [0.2683, 0.1204, 0.8635, 1.0]], 96: [[0.0216, 0.0324, 0.0211], [0.1773, 0.128, 0.5108, 1.0]], 97: [[0.0118, 0.0417, 0.0241], [0.5898, 0.4796, 0.4019, 1.0]], 98: [[0.0187, 0.0437, 0.0261], [0.1737, 0.5121, 0.6015, 1.0]], 99: [[0.019, 0.0335, 0.0375], [0.801, 0.5307, 0.3033, 1.0]], 100: [[0.0108, 0.022, 0.0234], [0.358, 0.5698, 0.2648, 1.0]], 101: [[0.0205, 0.0244, 0.0431], [0.5208, 0.3639, 0.3299, 1.0]], 102: [[0.0102, 0.0359, 0.0264], [0.2407, 0.1642, 0.2944, 1.0]], 103: [[0.0123, 0.0276, 0.0381], [0.236, 0.195, 0.7314, 1.0]], 104: [[0.0218, 0.035, 0.0384], [0.6931, 0.8354, 0.2006, 1.0]], 105: [[0.0159, 0.0287, 0.0423], [0.5896, 0.5709, 0.6947, 1.0]], 106: [[0.0103, 0.0292, 0.0226], [0.5831, 0.4177, 0.6897, 1.0]], 107: [[0.0109, 0.0346, 0.0363], [0.6137, 0.2844, 0.346, 1.0]], 108: [[0.0105, 0.0383, 0.0322], [0.5307, 0.8274, 0.2829, 1.0]], 109: [[0.0164, 0.0394, 0.0283], [0.8993, 0.1718, 0.4503, 1.0]], 110: [[0.013, 0.0367, 0.0354], [0.5632, 0.6391, 0.3101, 1.0]], 111: [[0.0152, 0.0221, 0.0314], [0.6006, 0.8747, 0.3721, 1.0]], 112: [[0.0117, 0.0411, 0.0287], [0.3596, 0.4585, 0.762, 1.0]], 113: [[0.0152, 0.0205, 0.0381], [0.5618, 0.162, 0.6138, 1.0]], 114: [[0.0131, 0.0422, 0.0359], [0.288, 0.1458, 0.7632, 1.0]], 115: [[0.0118, 0.0372, 0.0361], [0.2635, 0.8376, 0.3854, 1.0]], 116: [[0.0144, 0.0284, 0.0415], [0.4601, 0.7825, 0.8615, 1.0]], 117: [[0.0129, 0.0229, 0.0205], [0.492, 0.2172, 0.1617, 1.0]], 118: [[0.0136, 0.0216, 0.0323], [0.3743, 0.34, 0.8424, 1.0]], 119: [[0.0118, 0.0255, 0.0243], [0.3073, 0.8209, 0.2258, 1.0]], 120: [[0.0197, 0.0269, 0.0429], [0.3923, 0.5821, 0.1846, 1.0]], 121: [[0.0157, 0.0369, 0.036], [0.5091, 0.7993, 0.4414, 1.0]], 122: [[0.0119, 0.0221, 0.0405], [0.199, 0.8498, 0.4097, 1.0]], 123: [[0.0202, 0.043, 0.0324], [0.1926, 0.3486, 0.7912, 1.0]], 124: [[0.0156, 0.0391, 0.0254], [0.4234, 0.2629, 0.3079, 1.0]], 125: [[0.0211, 0.0392, 0.0387], [0.8434, 0.8085, 0.3314, 1.0]], 126: [[0.0133, 0.0448, 0.0313], [0.3631, 0.1914, 0.5593, 1.0]], 127: [[0.0172, 0.0382, 0.0263], [0.5667, 0.8125, 0.5394, 1.0]], 128: [[0.0198, 0.0257, 0.027], [0.2081, 0.8628, 0.5272, 1.0]], 129: [[0.0214, 0.0289, 0.0404], [0.5107, 0.5484, 0.8734, 1.0]], 130: [[0.0217, 0.0342, 0.024], [0.1731, 0.6679, 0.7242, 1.0]], 131: [[0.0197, 0.0341, 0.0324], [0.3754, 0.7483, 0.5249, 1.0]], 132: [[0.0213, 0.0422, 0.0298], [0.4306, 0.5199, 0.6132, 1.0]], 133: [[0.0164, 0.0424, 0.0254], [0.5267, 0.1447, 0.102, 1.0]], 134: [[0.015, 0.0415, 0.04], [0.6816, 0.3856, 0.2989, 1.0]], 135: [[0.0101, 0.0284, 0.0386], [0.8479, 0.7401, 0.4474, 1.0]], 136: [[0.0172, 0.0417, 0.0405], [0.2205, 0.4545, 0.2398, 1.0]], 137: [[0.0115, 0.032, 0.042], [0.1161, 0.3205, 0.2477, 1.0]], 138: [[0.0127, 0.0211, 0.0381], [0.5818, 0.6509, 0.3561, 1.0]], 139: [[0.0209, 0.0213, 0.0293], [0.7306, 0.3184, 0.1439, 1.0]], 140: [[0.0107, 0.0251, 0.0405], [0.6492, 0.2104, 0.4461, 1.0]], 141: [[0.0201, 0.045, 0.0342], [0.4241, 0.2191, 0.4656, 1.0]], 142: [[0.0119, 0.0394, 0.0362], [0.7678, 0.7518, 0.7368, 1.0]], 143: [[0.016, 0.0273, 0.0213], [0.7487, 0.3251, 0.4032, 1.0]], 144: [[0.0208, 0.0236, 0.0336], [0.5525, 0.2326, 0.4606, 1.0]], 145: [[0.0218, 0.0362, 0.0236], [0.5637, 0.5772, 0.5434, 1.0]], 146: [[0.017, 0.0388, 0.0355], [0.3889, 0.4015, 0.8094, 1.0]], 147: [[0.0115, 0.0414, 0.0405], [0.7109, 0.7992, 0.1502, 1.0]], 148: [[0.0162, 0.0333, 0.0211], [0.2807, 0.7799, 0.1244, 1.0]], 149: [[0.0207, 0.0299, 0.0224], [0.4421, 0.6508, 0.5639, 1.0]], 150: [[0.0192, 0.0412, 0.0342], [0.1031, 0.4742, 0.7186, 1.0]], 151: [[0.0103, 0.0327, 0.0252], [0.71, 0.6124, 0.2956, 1.0]], 152: [[0.0194, 0.0401, 0.031], [0.3148, 0.5322, 0.3538, 1.0]], 153: [[0.0216, 0.033, 0.0395], [0.5694, 0.4188, 0.6095, 1.0]], 154: [[0.0144, 0.0228, 0.044], [0.689, 0.8871, 0.399, 1.0]], 155: [[0.021, 0.0381, 0.0225], [0.6817, 0.5315, 0.2325, 1.0]], 156: [[0.0137, 0.0308, 0.0405], [0.2367, 0.7663, 0.5155, 1.0]], 157: [[0.0139, 0.0404, 0.0227], [0.4541, 0.8084, 0.1453, 1.0]], 158: [[0.0135, 0.0346, 0.0251], [0.6241, 0.6556, 0.4949, 1.0]], 159: [[0.022, 0.0413, 0.0217], [0.6033, 0.7087, 0.53, 1.0]], 160: [[0.0132, 0.0314, 0.0284], [0.5949, 0.6618, 0.2694, 1.0]], 161: [[0.0127, 0.0265, 0.0312], [0.6692, 0.219, 0.5189, 1.0]], 162: [[0.0101, 0.0436, 0.0339], [0.4773, 0.2023, 0.4044, 1.0]], 163: [[0.016, 0.0342, 0.0276], [0.585, 0.5593, 0.2596, 1.0]], 164: [[0.0164, 0.0241, 0.0413], [0.5503, 0.6184, 0.3801, 1.0]], 165: [[0.0168, 0.0416, 0.0389], [0.1073, 0.2302, 0.2206, 1.0]], 166: [[0.0151, 0.0261, 0.0202], [0.5729, 0.368, 0.3013, 1.0]], 167: [[0.0135, 0.0383, 0.0439], [0.3511, 0.2122, 0.6319, 1.0]], 168: [[0.0138, 0.0398, 0.0386], [0.7694, 0.3461, 0.252, 1.0]], 169: [[0.0154, 0.0423, 0.043], [0.1542, 0.2746, 0.7864, 1.0]], 170: [[0.0206, 0.0386, 0.0388], [0.8024, 0.7605, 0.2797, 1.0]], 171: [[0.0218, 0.0338, 0.0303], [0.8404, 0.6186, 0.7619, 1.0]], 172: [[0.0165, 0.038, 0.0269], [0.8242, 0.781, 0.8995, 1.0]], 173: [[0.0172, 0.0294, 0.0254], [0.4594, 0.6199, 0.882, 1.0]], 174: [[0.0182, 0.0306, 0.0364], [0.1616, 0.1437, 0.8758, 1.0]], 175: [[0.0202, 0.0398, 0.0412], [0.4723, 0.6764, 0.6574, 1.0]], 176: [[0.0187, 0.0343, 0.0332], [0.6391, 0.271, 0.6761, 1.0]], 177: [[0.0214, 0.0383, 0.0429], [0.2966, 0.7352, 0.899, 1.0]], 178: [[0.0165, 0.0448, 0.0394], [0.2997, 0.3067, 0.8112, 1.0]], 179: [[0.02, 0.0337, 0.0368], [0.4485, 0.2599, 0.3203, 1.0]], 180: [[0.0171, 0.0417, 0.0281], [0.7062, 0.6693, 0.518, 1.0]], 181: [[0.0212, 0.0402, 0.0443], [0.1907, 0.8191, 0.6024, 1.0]], 182: [[0.0149, 0.0299, 0.0283], [0.3886, 0.4874, 0.3695, 1.0]], 183: [[0.0207, 0.0311, 0.0281], [0.2276, 0.4357, 0.7311, 1.0]], 184: [[0.0193, 0.0323, 0.0233], [0.3985, 0.6813, 0.7162, 1.0]], 185: [[0.0155, 0.0336, 0.0336], [0.2722, 0.4393, 0.5203, 1.0]], 186: [[0.0144, 0.0396, 0.0357], [0.4398, 0.545, 0.8333, 1.0]], 187: [[0.0165, 0.0214, 0.0331], [0.191, 0.7097, 0.6747, 1.0]], 188: [[0.0103, 0.0241, 0.0213], [0.173, 0.2551, 0.1535, 1.0]], 189: [[0.018, 0.0398, 0.0234], [0.8608, 0.4749, 0.4799, 1.0]], 190: [[0.0166, 0.022, 0.039], [0.5312, 0.7164, 0.2489, 1.0]], 191: [[0.0103, 0.0272, 0.021], [0.8534, 0.1164, 0.7616, 1.0]], 192: [[0.011, 0.0382, 0.0221], [0.556, 0.4619, 0.7268, 1.0]], 193: [[0.0217, 0.0278, 0.034], [0.4169, 0.3591, 0.8646, 1.0]], 194: [[0.0199, 0.0296, 0.0293], [0.8043, 0.4033, 0.3037, 1.0]], 195: [[0.0173, 0.0306, 0.0243], [0.807, 0.5735, 0.858, 1.0]], 196: [[0.0112, 0.0434, 0.0413], [0.6777, 0.4382, 0.1137, 1.0]], 197: [[0.0106, 0.0246, 0.0303], [0.8758, 0.8195, 0.8418, 1.0]], 198: [[0.021, 0.0256, 0.0316], [0.8448, 0.7903, 0.1238, 1.0]], 199: [[0.0157, 0.042, 0.0224], [0.1269, 0.8075, 0.1688, 1.0]], 200: [[0.0197, 0.0309, 0.0221], [0.3874, 0.7454, 0.8182, 1.0]], 201: [[0.0189, 0.0203, 0.0346], [0.6335, 0.5233, 0.2571, 1.0]], 202: [[0.0218, 0.0366, 0.038], [0.2219, 0.3082, 0.1361, 1.0]], 203: [[0.0208, 0.0201, 0.0357], [0.1636, 0.27, 0.5034, 1.0]], 204: [[0.0208, 0.0365, 0.0345], [0.6328, 0.1272, 0.6168, 1.0]], 205: [[0.0124, 0.0373, 0.0311], [0.3344, 0.2031, 0.2468, 1.0]], 206: [[0.0143, 0.0379, 0.0395], [0.8305, 0.5423, 0.8664, 1.0]], 207: [[0.0133, 0.0212, 0.0414], [0.1625, 0.36, 0.7254, 1.0]], 208: [[0.0145, 0.0219, 0.0234], [0.4912, 0.2509, 0.1665, 1.0]], 209: [[0.0121, 0.0234, 0.0369], [0.5579, 0.378, 0.1668, 1.0]], 210: [[0.0143, 0.0292, 0.0339], [0.2489, 0.6209, 0.7502, 1.0]], 211: [[0.0142, 0.0432, 0.0371], [0.8218, 0.1552, 0.1052, 1.0]], 212: [[0.0115, 0.0289, 0.0284], [0.2255, 0.6463, 0.4094, 1.0]], 213: [[0.0176, 0.0255, 0.0202], [0.2328, 0.6855, 0.2289, 1.0]], 214: [[0.0112, 0.0363, 0.0287], [0.3388, 0.3265, 0.2721, 1.0]], 215: [[0.0132, 0.0264, 0.0209], [0.3365, 0.5819, 0.4295, 1.0]], 216: [[0.016, 0.0427, 0.0237], [0.1908, 0.1575, 0.1963, 1.0]], 217: [[0.0141, 0.0356, 0.0327], [0.8393, 0.4881, 0.2408, 1.0]], 218: [[0.0185, 0.0327, 0.043], [0.6225, 0.3706, 0.6238, 1.0]], 219: [[0.0112, 0.0244, 0.0243], [0.2414, 0.2091, 0.7657, 1.0]], 220: [[0.0219, 0.0268, 0.0365], [0.4134, 0.3872, 0.5928, 1.0]], 221: [[0.015, 0.0355, 0.0415], [0.578, 0.8624, 0.6922, 1.0]], 222: [[0.0175, 0.0289, 0.0376], [0.4067, 0.1945, 0.4169, 1.0]], 223: [[0.011, 0.0293, 0.0362], [0.8496, 0.1236, 0.8933, 1.0]], 224: [[0.0147, 0.0416, 0.0211], [0.5137, 0.7411, 0.7558, 1.0]], 225: [[0.014, 0.0239, 0.0287], [0.3363, 0.434, 0.5731, 1.0]], 226: [[0.0112, 0.0313, 0.0293], [0.2666, 0.1082, 0.2759, 1.0]], 227: [[0.0152, 0.0401, 0.0409], [0.7486, 0.6091, 0.1218, 1.0]], 228: [[0.0135, 0.0427, 0.0376], [0.7317, 0.1704, 0.453, 1.0]], 229: [[0.0109, 0.0276, 0.0354], [0.387, 0.22, 0.7152, 1.0]], 230: [[0.0154, 0.0259, 0.0311], [0.2042, 0.2908, 0.8036, 1.0]], 231: [[0.0215, 0.0315, 0.0381], [0.5352, 0.3633, 0.5082, 1.0]], 232: [[0.0142, 0.0398, 0.0334], [0.317, 0.4417, 0.4846, 1.0]], 233: [[0.0108, 0.0374, 0.0288], [0.1945, 0.164, 0.3555, 1.0]], 234: [[0.0158, 0.0415, 0.0369], [0.7297, 0.5111, 0.7063, 1.0]], 235: [[0.0156, 0.0337, 0.0439], [0.1131, 0.3021, 0.5014, 1.0]], 236: [[0.018, 0.039, 0.0319], [0.6088, 0.5688, 0.1823, 1.0]], 237: [[0.0216, 0.0395, 0.0342], [0.6259, 0.7867, 0.1727, 1.0]], 238: [[0.0118, 0.0293, 0.0296], [0.4397, 0.4899, 0.4579, 1.0]], 239: [[0.0152, 0.0302, 0.0318], [0.1486, 0.3765, 0.1784, 1.0]], 240: [[0.0197, 0.0409, 0.026], [0.2573, 0.5809, 0.7757, 1.0]], 241: [[0.0199, 0.0276, 0.0364], [0.5127, 0.3823, 0.1573, 1.0]], 242: [[0.0199, 0.0242, 0.0333], [0.47, 0.4244, 0.7773, 1.0]], 243: [[0.0197, 0.0323, 0.0424], [0.7977, 0.5848, 0.7968, 1.0]], 244: [[0.0107, 0.0395, 0.0273], [0.6682, 0.2293, 0.6209, 1.0]], 245: [[0.0119, 0.0267, 0.0323], [0.7545, 0.7158, 0.8866, 1.0]], 246: [[0.0168, 0.0298, 0.024], [0.4238, 0.7473, 0.4003, 1.0]], 247: [[0.0168, 0.0244, 0.0427], [0.5894, 0.3424, 0.6723, 1.0]], 248: [[0.0102, 0.0229, 0.0348], [0.4247, 0.8039, 0.5861, 1.0]], 249: [[0.0193, 0.0349, 0.0441], [0.5251, 0.2718, 0.4503, 1.0]]}

        box = {0: [[0.0188, 0.0192, 0.0238], [0.1157, 0.5468, 0.5298, 1.0]], 1: [[0.0166, 0.0154, 0.0161], [0.8362, 0.492, 0.5342, 1.0]], 2: [[0.0172, 0.017, 0.0229], [0.5015, 0.8978, 0.7618, 1.0]], 3: [[0.0231, 0.0211, 0.0213], [0.349, 0.2852, 0.5659, 1.0]], 4: [[0.0151, 0.0228, 0.0153], [0.8342, 0.5794, 0.2301, 1.0]], 5: [[0.0182, 0.0191, 0.0188], [0.54, 0.6492, 0.5547, 1.0]], 6: [[0.019, 0.0237, 0.0235], [0.6246, 0.3044, 0.7597, 1.0]], 7: [[0.0167, 0.0155, 0.0207], [0.1901, 0.7238, 0.3158, 1.0]], 8: [[0.0217, 0.0162, 0.0216], [0.2811, 0.6427, 0.4474, 1.0]], 9: [[0.0204, 0.0237, 0.0242], [0.3724, 0.2969, 0.2009, 1.0]], 10: [[0.0207, 0.0244, 0.0221], [0.103, 0.3814, 0.3197, 1.0]], 11: [[0.0212, 0.0165, 0.0225], [0.5088, 0.7749, 0.5548, 1.0]], 12: [[0.0202, 0.0249, 0.0191], [0.6839, 0.7958, 0.7363, 1.0]], 13: [[0.0235, 0.015, 0.0237], [0.7952, 0.7226, 0.4611, 1.0]], 14: [[0.0168, 0.0227, 0.0178], [0.4187, 0.7883, 0.3948, 1.0]], 15: [[0.0156, 0.0164, 0.0159], [0.3567, 0.3675, 0.3958, 1.0]], 16: [[0.019, 0.02, 0.0218], [0.2052, 0.5397, 0.8996, 1.0]], 17: [[0.02, 0.0239, 0.0227], [0.6051, 0.8375, 0.893, 1.0]], 18: [[0.0198, 0.0204, 0.0178], [0.855, 0.2184, 0.6497, 1.0]], 19: [[0.0197, 0.0195, 0.0246], [0.4698, 0.2316, 0.6538, 1.0]], 20: [[0.0166, 0.0191, 0.0181], [0.4624, 0.6063, 0.7248, 1.0]], 21: [[0.017, 0.0153, 0.0203], [0.4032, 0.5508, 0.3714, 1.0]], 22: [[0.0236, 0.0183, 0.0169], [0.8093, 0.7768, 0.3895, 1.0]], 23: [[0.0227, 0.0178, 0.0228], [0.4312, 0.8795, 0.3573, 1.0]], 24: [[0.0222, 0.0176, 0.0218], [0.4119, 0.2034, 0.1935, 1.0]], 25: [[0.0249, 0.0241, 0.0209], [0.7022, 0.2689, 0.6458, 1.0]], 26: [[0.0177, 0.0246, 0.0169], [0.291, 0.4589, 0.1552, 1.0]], 27: [[0.0179, 0.0245, 0.022], [0.5256, 0.7501, 0.5228, 1.0]], 28: [[0.0236, 0.0219, 0.0218], [0.2704, 0.301, 0.887, 1.0]], 29: [[0.0161, 0.0236, 0.0202], [0.259, 0.5184, 0.6071, 1.0]], 30: [[0.0228, 0.0204, 0.0165], [0.7792, 0.3918, 0.2399, 1.0]], 31: [[0.0151, 0.0224, 0.0237], [0.1878, 0.3899, 0.6565, 1.0]], 32: [[0.0206, 0.0202, 0.0245], [0.1409, 0.4541, 0.5315, 1.0]], 33: [[0.0182, 0.0225, 0.0163], [0.2971, 0.8615, 0.6326, 1.0]], 34: [[0.0154, 0.0158, 0.0153], [0.7879, 0.7887, 0.1791, 1.0]], 35: [[0.0227, 0.0204, 0.0225], [0.7103, 0.187, 0.7408, 1.0]], 36: [[0.0187, 0.0189, 0.0248], [0.8541, 0.4555, 0.3461, 1.0]], 37: [[0.0167, 0.0233, 0.0247], [0.6725, 0.6654, 0.6883, 1.0]], 38: [[0.021, 0.0187, 0.0216], [0.2053, 0.5668, 0.8892, 1.0]], 39: [[0.0207, 0.0188, 0.0192], [0.7457, 0.2201, 0.7194, 1.0]], 40: [[0.0207, 0.0232, 0.0207], [0.7449, 0.7471, 0.3682, 1.0]], 41: [[0.0231, 0.0221, 0.019], [0.8289, 0.4981, 0.6716, 1.0]], 42: [[0.0198, 0.0194, 0.0174], [0.3994, 0.8011, 0.2073, 1.0]], 43: [[0.0239, 0.017, 0.0207], [0.3184, 0.5842, 0.7444, 1.0]], 44: [[0.0177, 0.0211, 0.0183], [0.4592, 0.3165, 0.4915, 1.0]], 45: [[0.02, 0.0188, 0.0167], [0.4009, 0.2913, 0.8605, 1.0]], 46: [[0.0197, 0.0249, 0.0217], [0.1467, 0.2308, 0.1109, 1.0]], 47: [[0.0247, 0.0221, 0.0248], [0.8624, 0.578, 0.2853, 1.0]], 48: [[0.0172, 0.0219, 0.0249], [0.1328, 0.4253, 0.2437, 1.0]], 49: [[0.0207, 0.0194, 0.0229], [0.1358, 0.6328, 0.3772, 1.0]], 50: [[0.0209, 0.0164, 0.0201], [0.1579, 0.6973, 0.5214, 1.0]], 51: [[0.0212, 0.0173, 0.0184], [0.8681, 0.3725, 0.1858, 1.0]], 52: [[0.0207, 0.0225, 0.0156], [0.8406, 0.7461, 0.2469, 1.0]], 53: [[0.017, 0.0153, 0.0213], [0.5882, 0.6247, 0.5005, 1.0]], 54: [[0.022, 0.0187, 0.0189], [0.1203, 0.5383, 0.7762, 1.0]], 55: [[0.0219, 0.0217, 0.0191], [0.7271, 0.1392, 0.4995, 1.0]], 56: [[0.019, 0.0212, 0.0173], [0.6879, 0.8504, 0.8998, 1.0]], 57: [[0.0192, 0.0152, 0.0207], [0.3944, 0.1462, 0.1537, 1.0]], 58: [[0.0234, 0.0246, 0.0176], [0.7466, 0.1077, 0.4363, 1.0]], 59: [[0.0166, 0.0172, 0.0153], [0.3044, 0.2311, 0.123, 1.0]], 60: [[0.0218, 0.0202, 0.0211], [0.43, 0.6779, 0.6962, 1.0]], 61: [[0.0215, 0.0249, 0.02], [0.3305, 0.4902, 0.3456, 1.0]], 62: [[0.018, 0.0245, 0.019], [0.1645, 0.4244, 0.1102, 1.0]], 63: [[0.0229, 0.0162, 0.0242], [0.4765, 0.6324, 0.5216, 1.0]], 64: [[0.0201, 0.0246, 0.0152], [0.6425, 0.7387, 0.7668, 1.0]], 65: [[0.0189, 0.0239, 0.0166], [0.6015, 0.2813, 0.6271, 1.0]], 66: [[0.0157, 0.0243, 0.0182], [0.3029, 0.6347, 0.1559, 1.0]], 67: [[0.0222, 0.0223, 0.0241], [0.4463, 0.5335, 0.8092, 1.0]], 68: [[0.0227, 0.0168, 0.0174], [0.8336, 0.6726, 0.4362, 1.0]], 69: [[0.0178, 0.0221, 0.0194], [0.1254, 0.4113, 0.805, 1.0]], 70: [[0.0225, 0.018, 0.0185], [0.5456, 0.1449, 0.4684, 1.0]], 71: [[0.0202, 0.018, 0.0206], [0.3386, 0.619, 0.7237, 1.0]], 72: [[0.0198, 0.016, 0.0201], [0.3958, 0.2983, 0.3007, 1.0]], 73: [[0.0221, 0.016, 0.0174], [0.8759, 0.7854, 0.5189, 1.0]], 74: [[0.0224, 0.0208, 0.0197], [0.5449, 0.3217, 0.8611, 1.0]], 75: [[0.0163, 0.0184, 0.0192], [0.3695, 0.8234, 0.2277, 1.0]], 76: [[0.0238, 0.0215, 0.0198], [0.1953, 0.2573, 0.4841, 1.0]], 77: [[0.0219, 0.0186, 0.0166], [0.1623, 0.1582, 0.3637, 1.0]], 78: [[0.0229, 0.0234, 0.0208], [0.8211, 0.8311, 0.816, 1.0]], 79: [[0.0203, 0.0217, 0.0217], [0.1982, 0.4906, 0.8791, 1.0]], 80: [[0.019, 0.0185, 0.0187], [0.4395, 0.7735, 0.7449, 1.0]], 81: [[0.0152, 0.019, 0.0224], [0.3388, 0.4858, 0.8543, 1.0]], 82: [[0.0208, 0.023, 0.0186], [0.3971, 0.3404, 0.7106, 1.0]], 83: [[0.0156, 0.017, 0.0233], [0.5972, 0.4844, 0.8015, 1.0]], 84: [[0.0217, 0.0159, 0.0245], [0.4526, 0.6549, 0.6808, 1.0]], 85: [[0.0207, 0.0224, 0.0249], [0.1635, 0.1974, 0.5968, 1.0]], 86: [[0.0198, 0.0198, 0.0249], [0.8498, 0.4652, 0.127, 1.0]], 87: [[0.0212, 0.0191, 0.0186], [0.6708, 0.6304, 0.829, 1.0]], 88: [[0.0245, 0.0155, 0.023], [0.1101, 0.2157, 0.2597, 1.0]], 89: [[0.0159, 0.0181, 0.0213], [0.5378, 0.5534, 0.6585, 1.0]], 90: [[0.0159, 0.0157, 0.0205], [0.3187, 0.6664, 0.7561, 1.0]], 91: [[0.0215, 0.0213, 0.021], [0.1618, 0.8518, 0.7655, 1.0]], 92: [[0.0217, 0.0167, 0.025], [0.8264, 0.1042, 0.5996, 1.0]], 93: [[0.023, 0.0188, 0.0214], [0.1535, 0.3703, 0.8753, 1.0]], 94: [[0.0233, 0.0241, 0.0209], [0.7126, 0.2512, 0.8931, 1.0]], 95: [[0.0157, 0.02, 0.0184], [0.5809, 0.3672, 0.805, 1.0]], 96: [[0.0218, 0.0215, 0.0161], [0.7363, 0.5192, 0.3646, 1.0]], 97: [[0.0167, 0.0154, 0.0236], [0.791, 0.5462, 0.4804, 1.0]], 98: [[0.0205, 0.0213, 0.0209], [0.5034, 0.6078, 0.584, 1.0]], 99: [[0.02, 0.021, 0.0203], [0.2591, 0.4827, 0.2348, 1.0]], 100: [[0.0226, 0.016, 0.0244], [0.6201, 0.8978, 0.4718, 1.0]], 101: [[0.017, 0.022, 0.0156], [0.121, 0.7951, 0.6697, 1.0]], 102: [[0.0232, 0.0158, 0.0246], [0.5097, 0.3134, 0.3802, 1.0]], 103: [[0.0187, 0.0205, 0.0242], [0.8249, 0.1654, 0.5373, 1.0]], 104: [[0.0243, 0.0167, 0.022], [0.885, 0.2512, 0.5524, 1.0]], 105: [[0.0163, 0.0207, 0.0186], [0.8041, 0.6267, 0.4802, 1.0]], 106: [[0.0214, 0.0179, 0.0157], [0.1829, 0.4316, 0.3659, 1.0]], 107: [[0.0178, 0.0168, 0.0236], [0.3865, 0.6586, 0.3085, 1.0]], 108: [[0.0197, 0.0157, 0.0151], [0.2064, 0.3052, 0.481, 1.0]], 109: [[0.0216, 0.0245, 0.0189], [0.795, 0.5807, 0.5138, 1.0]], 110: [[0.0228, 0.0206, 0.0212], [0.7062, 0.3873, 0.6848, 1.0]], 111: [[0.0174, 0.0231, 0.0168], [0.8666, 0.7965, 0.2914, 1.0]], 112: [[0.0209, 0.0227, 0.0151], [0.4429, 0.4441, 0.3847, 1.0]], 113: [[0.016, 0.0165, 0.0241], [0.6086, 0.3449, 0.6998, 1.0]], 114: [[0.0172, 0.019, 0.0248], [0.496, 0.6682, 0.6098, 1.0]], 115: [[0.0214, 0.0174, 0.0167], [0.6637, 0.5879, 0.2356, 1.0]], 116: [[0.0239, 0.0183, 0.0231], [0.3474, 0.467, 0.1889, 1.0]], 117: [[0.0246, 0.0154, 0.0193], [0.6473, 0.7742, 0.1641, 1.0]], 118: [[0.0157, 0.0162, 0.0182], [0.8136, 0.646, 0.5016, 1.0]], 119: [[0.0154, 0.0196, 0.0238], [0.7429, 0.8963, 0.8465, 1.0]], 120: [[0.0173, 0.0207, 0.0244], [0.7942, 0.5455, 0.3962, 1.0]], 121: [[0.0193, 0.0152, 0.0168], [0.1819, 0.8512, 0.4593, 1.0]], 122: [[0.0234, 0.0197, 0.018], [0.3722, 0.414, 0.1006, 1.0]], 123: [[0.0177, 0.0208, 0.0232], [0.2606, 0.3974, 0.3894, 1.0]], 124: [[0.0178, 0.0172, 0.0229], [0.8913, 0.3903, 0.6651, 1.0]], 125: [[0.0204, 0.0169, 0.0199], [0.3235, 0.8712, 0.3666, 1.0]], 126: [[0.0162, 0.024, 0.0201], [0.6081, 0.1862, 0.8895, 1.0]], 127: [[0.0239, 0.021, 0.0157], [0.1972, 0.2049, 0.703, 1.0]], 128: [[0.0218, 0.0238, 0.0191], [0.6624, 0.4706, 0.2322, 1.0]], 129: [[0.0183, 0.0184, 0.0173], [0.5123, 0.3576, 0.3863, 1.0]], 130: [[0.0185, 0.0243, 0.0249], [0.7913, 0.8203, 0.4, 1.0]], 131: [[0.0161, 0.0192, 0.0201], [0.1152, 0.8901, 0.4979, 1.0]], 132: [[0.0214, 0.0205, 0.0244], [0.5835, 0.8189, 0.617, 1.0]], 133: [[0.0151, 0.0168, 0.0151], [0.1735, 0.1128, 0.1499, 1.0]], 134: [[0.0186, 0.0155, 0.0184], [0.4947, 0.8172, 0.7077, 1.0]], 135: [[0.0218, 0.0163, 0.0245], [0.8951, 0.2786, 0.2947, 1.0]], 136: [[0.0191, 0.0204, 0.0225], [0.7025, 0.7, 0.1905, 1.0]], 137: [[0.0235, 0.0243, 0.0222], [0.7864, 0.5797, 0.6, 1.0]], 138: [[0.0161, 0.0189, 0.0187], [0.6994, 0.4672, 0.6465, 1.0]], 139: [[0.0219, 0.0197, 0.0226], [0.2589, 0.4121, 0.7096, 1.0]], 140: [[0.0205, 0.0206, 0.0243], [0.747, 0.8136, 0.8908, 1.0]], 141: [[0.0152, 0.0198, 0.0166], [0.4414, 0.3087, 0.5194, 1.0]], 142: [[0.0162, 0.0184, 0.0186], [0.4307, 0.4766, 0.2516, 1.0]], 143: [[0.0245, 0.0216, 0.0248], [0.2028, 0.5749, 0.4629, 1.0]], 144: [[0.02, 0.0216, 0.0178], [0.8497, 0.7481, 0.2985, 1.0]], 145: [[0.0233, 0.0202, 0.0236], [0.6327, 0.2855, 0.5918, 1.0]], 146: [[0.0243, 0.0224, 0.0236], [0.4561, 0.6034, 0.5216, 1.0]], 147: [[0.0218, 0.0201, 0.0193], [0.2356, 0.5694, 0.4694, 1.0]], 148: [[0.0228, 0.0219, 0.0241], [0.11, 0.7373, 0.1046, 1.0]], 149: [[0.0183, 0.0242, 0.021], [0.2571, 0.2436, 0.8815, 1.0]], 150: [[0.0179, 0.0245, 0.0196], [0.8699, 0.5128, 0.4463, 1.0]], 151: [[0.0166, 0.0181, 0.0204], [0.2349, 0.8111, 0.6923, 1.0]], 152: [[0.0236, 0.0215, 0.0198], [0.6455, 0.8282, 0.3836, 1.0]], 153: [[0.0159, 0.0182, 0.0205], [0.4966, 0.723, 0.8175, 1.0]], 154: [[0.0166, 0.0174, 0.0216], [0.8929, 0.718, 0.8695, 1.0]], 155: [[0.0198, 0.0222, 0.0228], [0.6348, 0.7252, 0.566, 1.0]], 156: [[0.0206, 0.0188, 0.0209], [0.4951, 0.7481, 0.5229, 1.0]], 157: [[0.0154, 0.0249, 0.0231], [0.8488, 0.4081, 0.8447, 1.0]], 158: [[0.0181, 0.0229, 0.0151], [0.7611, 0.875, 0.2098, 1.0]], 159: [[0.0183, 0.0233, 0.0192], [0.5327, 0.8806, 0.4014, 1.0]], 160: [[0.02, 0.0166, 0.0189], [0.3005, 0.1474, 0.6373, 1.0]], 161: [[0.0247, 0.0235, 0.0172], [0.734, 0.1927, 0.1626, 1.0]], 162: [[0.0163, 0.0225, 0.0234], [0.2544, 0.143, 0.4345, 1.0]], 163: [[0.0204, 0.0175, 0.0171], [0.4171, 0.5949, 0.5354, 1.0]], 164: [[0.0226, 0.0234, 0.0156], [0.819, 0.7545, 0.7482, 1.0]], 165: [[0.0198, 0.0214, 0.0249], [0.7458, 0.6964, 0.4539, 1.0]], 166: [[0.0205, 0.0242, 0.0222], [0.8448, 0.4132, 0.4392, 1.0]], 167: [[0.0211, 0.022, 0.0188], [0.4747, 0.34, 0.7601, 1.0]], 168: [[0.0165, 0.0196, 0.022], [0.2946, 0.696, 0.6032, 1.0]], 169: [[0.02, 0.0156, 0.0233], [0.7002, 0.4049, 0.5347, 1.0]], 170: [[0.0197, 0.0174, 0.0156], [0.4793, 0.4806, 0.6855, 1.0]], 171: [[0.0231, 0.0171, 0.0235], [0.1064, 0.6366, 0.4003, 1.0]], 172: [[0.02, 0.0236, 0.0227], [0.1286, 0.1423, 0.2565, 1.0]], 173: [[0.0201, 0.0157, 0.0213], [0.6695, 0.6928, 0.1007, 1.0]], 174: [[0.0235, 0.0225, 0.021], [0.8023, 0.7059, 0.3136, 1.0]], 175: [[0.02, 0.0219, 0.0228], [0.2672, 0.7438, 0.7211, 1.0]], 176: [[0.0227, 0.0167, 0.0157], [0.1625, 0.1078, 0.5148, 1.0]], 177: [[0.0188, 0.0237, 0.0246], [0.2965, 0.4671, 0.5822, 1.0]], 178: [[0.0153, 0.0152, 0.0234], [0.7245, 0.7919, 0.8613, 1.0]], 179: [[0.0172, 0.0184, 0.0171], [0.1694, 0.2967, 0.1001, 1.0]], 180: [[0.0202, 0.0187, 0.0153], [0.2622, 0.5813, 0.8903, 1.0]], 181: [[0.0157, 0.0205, 0.0225], [0.4663, 0.2546, 0.6237, 1.0]], 182: [[0.0213, 0.0165, 0.0155], [0.6339, 0.8072, 0.223, 1.0]], 183: [[0.0249, 0.0157, 0.0191], [0.1859, 0.883, 0.3162, 1.0]], 184: [[0.0193, 0.024, 0.0178], [0.5108, 0.6493, 0.8669, 1.0]], 185: [[0.0201, 0.0199, 0.0196], [0.3163, 0.4088, 0.5704, 1.0]], 186: [[0.019, 0.022, 0.0246], [0.4674, 0.4149, 0.1426, 1.0]], 187: [[0.0209, 0.0188, 0.0214], [0.7427, 0.2838, 0.8177, 1.0]], 188: [[0.0212, 0.0238, 0.0221], [0.817, 0.3788, 0.6606, 1.0]], 189: [[0.0226, 0.0212, 0.0172], [0.406, 0.3903, 0.1421, 1.0]], 190: [[0.0157, 0.0189, 0.0179], [0.5673, 0.231, 0.4941, 1.0]], 191: [[0.0243, 0.0224, 0.0167], [0.6421, 0.6707, 0.2522, 1.0]], 192: [[0.0185, 0.0214, 0.0201], [0.4481, 0.4325, 0.514, 1.0]], 193: [[0.0175, 0.0154, 0.0209], [0.4171, 0.7212, 0.7224, 1.0]], 194: [[0.0232, 0.0163, 0.0222], [0.6242, 0.1554, 0.6223, 1.0]], 195: [[0.0189, 0.0198, 0.0194], [0.4561, 0.3317, 0.733, 1.0]], 196: [[0.0163, 0.0223, 0.0161], [0.8438, 0.4038, 0.3435, 1.0]], 197: [[0.0152, 0.0228, 0.0152], [0.6622, 0.4895, 0.5117, 1.0]], 198: [[0.0241, 0.0227, 0.0216], [0.6316, 0.8634, 0.1768, 1.0]], 199: [[0.0166, 0.016, 0.0194], [0.728, 0.3733, 0.1028, 1.0]], 200: [[0.023, 0.023, 0.0209], [0.3375, 0.3126, 0.8713, 1.0]], 201: [[0.0184, 0.0161, 0.0156], [0.7476, 0.7174, 0.7739, 1.0]], 202: [[0.0233, 0.0182, 0.0202], [0.2979, 0.7325, 0.4604, 1.0]], 203: [[0.016, 0.0216, 0.0211], [0.1558, 0.3887, 0.2349, 1.0]], 204: [[0.022, 0.0198, 0.023], [0.5106, 0.3487, 0.418, 1.0]], 205: [[0.0231, 0.0218, 0.0238], [0.4113, 0.8085, 0.5832, 1.0]], 206: [[0.019, 0.0202, 0.0208], [0.441, 0.3086, 0.4229, 1.0]], 207: [[0.0184, 0.0225, 0.0203], [0.7053, 0.3388, 0.8543, 1.0]], 208: [[0.0248, 0.0211, 0.0229], [0.3469, 0.3735, 0.5464, 1.0]], 209: [[0.0248, 0.0236, 0.0166], [0.1506, 0.8515, 0.4782, 1.0]], 210: [[0.0162, 0.0183, 0.0229], [0.3541, 0.3448, 0.5782, 1.0]], 211: [[0.0181, 0.0182, 0.0198], [0.762, 0.6647, 0.898, 1.0]], 212: [[0.0206, 0.0169, 0.0152], [0.2578, 0.4881, 0.6495, 1.0]], 213: [[0.0186, 0.0156, 0.0207], [0.1742, 0.6029, 0.2419, 1.0]], 214: [[0.0169, 0.0159, 0.0153], [0.7548, 0.2567, 0.1852, 1.0]], 215: [[0.0165, 0.0248, 0.0172], [0.1056, 0.4079, 0.5451, 1.0]], 216: [[0.021, 0.0169, 0.0192], [0.5669, 0.6307, 0.1621, 1.0]], 217: [[0.0238, 0.0155, 0.0174], [0.6802, 0.6323, 0.5451, 1.0]], 218: [[0.0233, 0.0191, 0.0176], [0.7547, 0.1679, 0.4837, 1.0]], 219: [[0.0219, 0.0181, 0.0236], [0.7421, 0.6871, 0.6387, 1.0]], 220: [[0.0194, 0.0235, 0.0185], [0.7106, 0.4854, 0.1913, 1.0]], 221: [[0.0162, 0.0159, 0.0229], [0.3346, 0.2574, 0.4671, 1.0]], 222: [[0.0216, 0.0246, 0.0179], [0.3851, 0.8005, 0.3999, 1.0]], 223: [[0.0183, 0.0175, 0.0169], [0.2824, 0.3015, 0.693, 1.0]], 224: [[0.0226, 0.0236, 0.0219], [0.2421, 0.1436, 0.3191, 1.0]], 225: [[0.0151, 0.0247, 0.0238], [0.6006, 0.7322, 0.3906, 1.0]], 226: [[0.016, 0.0226, 0.0168], [0.5646, 0.8537, 0.2265, 1.0]], 227: [[0.0217, 0.0175, 0.0155], [0.8814, 0.5999, 0.6421, 1.0]], 228: [[0.0223, 0.0155, 0.0188], [0.3754, 0.7213, 0.6549, 1.0]], 229: [[0.02, 0.0152, 0.0229], [0.1865, 0.8622, 0.5378, 1.0]], 230: [[0.0205, 0.0247, 0.0225], [0.8969, 0.4254, 0.7325, 1.0]], 231: [[0.0182, 0.0182, 0.021], [0.7619, 0.2873, 0.2591, 1.0]], 232: [[0.018, 0.0167, 0.0166], [0.6019, 0.4609, 0.7598, 1.0]], 233: [[0.0248, 0.0228, 0.018], [0.5112, 0.1142, 0.2955, 1.0]], 234: [[0.0164, 0.0232, 0.0199], [0.2676, 0.332, 0.4466, 1.0]], 235: [[0.0206, 0.0244, 0.0167], [0.8465, 0.6413, 0.419, 1.0]], 236: [[0.0198, 0.0195, 0.0196], [0.8642, 0.8079, 0.2992, 1.0]], 237: [[0.0195, 0.0235, 0.0157], [0.418, 0.8809, 0.89, 1.0]], 238: [[0.0176, 0.0248, 0.0212], [0.2692, 0.3212, 0.3103, 1.0]], 239: [[0.0177, 0.0164, 0.024], [0.3752, 0.1648, 0.4115, 1.0]], 240: [[0.0178, 0.0197, 0.0236], [0.815, 0.3113, 0.4348, 1.0]], 241: [[0.021, 0.0246, 0.0153], [0.1793, 0.5382, 0.8984, 1.0]], 242: [[0.0162, 0.0196, 0.0228], [0.7266, 0.8403, 0.1349, 1.0]], 243: [[0.0232, 0.021, 0.0155], [0.3308, 0.4696, 0.7282, 1.0]], 244: [[0.0186, 0.0188, 0.0213], [0.6601, 0.7555, 0.3684, 1.0]], 245: [[0.0171, 0.025, 0.0157], [0.4995, 0.2881, 0.3414, 1.0]], 246: [[0.0226, 0.0194, 0.0155], [0.8502, 0.2585, 0.4709, 1.0]], 247: [[0.0214, 0.0158, 0.0184], [0.4867, 0.6682, 0.7835, 1.0]], 248: [[0.0232, 0.018, 0.017], [0.2365, 0.3333, 0.5871, 1.0]], 249: [[0.0185, 0.0201, 0.0246], [0.1041, 0.7671, 0.6034, 1.0]]}

        # randomize target
        self.obj_gid = self.sim.model.geom_name2id('obj')
        self.tar_gid = self.sim.model.geom_name2id('target')

        self.obj_t_gid = self.sim.model.geom_name2id('top')
        self.obj_b_gid = self.sim.model.geom_name2id('bot')
        self.tar_t_gid = self.sim.model.geom_name2id('t_top')
        self.tar_b_gid = self.sim.model.geom_name2id('t_bot')

        geom_type = self.np_random.choice([3,4,5,6])

        if geom_type == 3:
            size = caps[self.np_random.choice(range(len(caps)))][0]
            color = caps[self.np_random.choice(range(len(caps)))][1]
            top_pos = np.array([0, 0, 1.3*size[1]])
            bot_pos = np.array([0, 0, 1.3*-size[1]])
        elif geom_type == 4:
            size = ellips[self.np_random.choice(range(len(ellips)))][0]
            color = ellips[self.np_random.choice(range(len(ellips)))][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])
        elif geom_type == 5:
            size = cyl[self.np_random.choice(range(len(cyl)))][0]
            color = cyl[self.np_random.choice(range(len(cyl)))][1]
            top_pos = np.array([0, 0, size[1]])
            bot_pos = np.array([0, 0, -size[1]])
        else:
            size = box[self.np_random.choice(range(len(box)))][0]
            color = box[self.np_random.choice(range(len(box)))][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])

        color = [x/255 for x in [38, 194, 129, 255]]

        self.sim.model.geom_size[self.obj_gid] = size
        self.sim.model.geom_type[self.obj_gid] = geom_type
        self.sim.model.geom_rgba[self.obj_gid] = color
        self.sim.model.geom_pos[self.obj_t_gid] = top_pos
        self.sim.model.geom_pos[self.obj_b_gid] = bot_pos

        self.sim.model.body_mass[self.obj_bid] = 1.2

        self.sim.model.geom_size[self.tar_gid] = size
        self.sim.model.geom_type[self.tar_gid] = geom_type
        self.sim.model.geom_rgba[self.tar_gid] = color
        self.sim.model.geom_pos[self.tar_t_gid] = top_pos
        self.sim.model.geom_pos[self.tar_b_gid] = bot_pos

        self.sim.model.geom_condim[self.obj_gid] = 3
        desired_orien = np.zeros(3)
        desired_orien[0] = self.np_random.uniform(low=-1, high=1)
        desired_orien[1] = self.np_random.uniform(low=-.8, high=1.2)
        self.sim.model.body_quat[self.target_obj_bid] = euler2quat(desired_orien)

        self.sim.model.site_rgba[self.success_indicator_sid, :2] = np.array([2, 0])

        self.robot.sync_sims(self.sim, self.sim_obsd)
        obs = super().reset()
        return obs

class OutofDistribution(ProprioceptiveEnvV0):
    def reset(self):
        ellips = {0: [[0.0192, 0.0452, 0.0499], [0.5267, 0.4574, 0.8141, 1.0]], 1: [[0.0139, 0.0151, 0.0196], [0.732, 0.149, 0.7571, 1.0]], 2: [[0.0199, 0.0193, 0.0195], [0.8499, 0.8535, 0.5279, 1.0]], 3: [[0.013, 0.0498, 0.0495], [0.4756, 0.6692, 0.2101, 1.0]], 4: [[0.0101, 0.0184, 0.017], [0.2398, 0.142, 0.2351, 1.0]], 5: [[0.0163, 0.0181, 0.0158], [0.7945, 0.3168, 0.8549, 1.0]], 6: [[0.0123, 0.016, 0.0178], [0.2792, 0.164, 0.1963, 1.0]], 7: [[0.0207, 0.0151, 0.0191], [0.7214, 0.2926, 0.1578, 1.0]], 8: [[0.0101, 0.0178, 0.0184], [0.2483, 0.829, 0.394, 1.0]], 9: [[0.0152, 0.0168, 0.0184], [0.4377, 0.3745, 0.6418, 1.0]], 10: [[0.0181, 0.049, 0.0478], [0.6609, 0.7573, 0.3534, 1.0]], 11: [[0.0209, 0.0472, 0.0461], [0.8082, 0.6887, 0.5339, 1.0]], 12: [[0.0207, 0.0458, 0.0479], [0.2804, 0.6847, 0.2201, 1.0]], 13: [[0.021, 0.0464, 0.0483], [0.4936, 0.2515, 0.1887, 1.0]], 14: [[0.0208, 0.0171, 0.0162], [0.5238, 0.8167, 0.4633, 1.0]], 15: [[0.0191, 0.0173, 0.0152], [0.1136, 0.1179, 0.4157, 1.0]], 16: [[0.0164, 0.0186, 0.0168], [0.1566, 0.4181, 0.4308, 1.0]], 17: [[0.0218, 0.046, 0.0452], [0.7179, 0.2007, 0.6449, 1.0]], 18: [[0.013, 0.0455, 0.049], [0.2808, 0.8917, 0.8089, 1.0]], 19: [[0.0172, 0.0151, 0.0184], [0.6515, 0.5491, 0.1872, 1.0]], 20: [[0.0147, 0.0196, 0.0198], [0.4904, 0.2028, 0.1975, 1.0]], 21: [[0.018, 0.048, 0.0471], [0.1999, 0.5807, 0.2375, 1.0]], 22: [[0.021, 0.0478, 0.0459], [0.4834, 0.2712, 0.8256, 1.0]], 23: [[0.0215, 0.0174, 0.0194], [0.876, 0.1412, 0.5414, 1.0]], 24: [[0.0138, 0.0499, 0.0451], [0.2633, 0.2109, 0.1106, 1.0]], 25: [[0.0127, 0.0498, 0.0465], [0.217, 0.5914, 0.4403, 1.0]], 26: [[0.0205, 0.0477, 0.0479], [0.6415, 0.5312, 0.1625, 1.0]], 27: [[0.0166, 0.016, 0.0188], [0.3175, 0.7062, 0.3952, 1.0]], 28: [[0.0153, 0.0176, 0.0176], [0.4573, 0.1392, 0.3729, 1.0]], 29: [[0.018, 0.0175, 0.0161], [0.6205, 0.5582, 0.4211, 1.0]], 30: [[0.0148, 0.0186, 0.0172], [0.6425, 0.8501, 0.547, 1.0]], 31: [[0.0106, 0.0152, 0.0171], [0.5701, 0.5896, 0.423, 1.0]], 32: [[0.0159, 0.0166, 0.0159], [0.4597, 0.5511, 0.5647, 1.0]], 33: [[0.0197, 0.0158, 0.0159], [0.5407, 0.8049, 0.5615, 1.0]], 34: [[0.0123, 0.0499, 0.0479], [0.2203, 0.4778, 0.6365, 1.0]], 35: [[0.0192, 0.0151, 0.0158], [0.2206, 0.8939, 0.2668, 1.0]], 36: [[0.0169, 0.0196, 0.0172], [0.8576, 0.2538, 0.2064, 1.0]], 37: [[0.0158, 0.0182, 0.0151], [0.4524, 0.197, 0.5453, 1.0]], 38: [[0.0207, 0.0493, 0.0478], [0.6939, 0.1908, 0.4939, 1.0]], 39: [[0.0175, 0.0156, 0.0197], [0.2389, 0.5661, 0.5583, 1.0]], 40: [[0.0166, 0.0459, 0.0465], [0.7683, 0.2604, 0.1557, 1.0]], 41: [[0.0208, 0.0186, 0.0156], [0.3433, 0.4539, 0.1055, 1.0]], 42: [[0.0205, 0.0477, 0.0456], [0.1845, 0.5981, 0.1483, 1.0]], 43: [[0.0124, 0.0489, 0.0499], [0.5175, 0.2887, 0.3097, 1.0]], 44: [[0.0113, 0.0485, 0.0461], [0.3786, 0.8318, 0.1601, 1.0]], 45: [[0.0215, 0.016, 0.0166], [0.3173, 0.1281, 0.1034, 1.0]], 46: [[0.0144, 0.0186, 0.0179], [0.1642, 0.5316, 0.5814, 1.0]], 47: [[0.0193, 0.0487, 0.0457], [0.3788, 0.4253, 0.8784, 1.0]], 48: [[0.0185, 0.0197, 0.0198], [0.1222, 0.1569, 0.3243, 1.0]], 49: [[0.0107, 0.0196, 0.0186], [0.7762, 0.2969, 0.1212, 1.0]], 50: [[0.0185, 0.0475, 0.0464], [0.4258, 0.596, 0.1708, 1.0]], 51: [[0.0219, 0.0463, 0.0453], [0.6682, 0.2874, 0.8677, 1.0]], 52: [[0.0216, 0.0184, 0.0167], [0.2306, 0.5239, 0.6937, 1.0]], 53: [[0.0142, 0.0153, 0.0172], [0.4521, 0.3318, 0.8279, 1.0]], 54: [[0.0205, 0.0163, 0.0187], [0.7757, 0.2893, 0.5572, 1.0]], 55: [[0.0111, 0.0487, 0.0458], [0.8859, 0.8771, 0.1853, 1.0]], 56: [[0.0123, 0.0494, 0.046], [0.2689, 0.5928, 0.3019, 1.0]], 57: [[0.0181, 0.0495, 0.046], [0.7198, 0.1171, 0.6706, 1.0]], 58: [[0.0101, 0.0157, 0.0152], [0.7989, 0.6149, 0.3325, 1.0]], 59: [[0.0119, 0.0199, 0.0164], [0.2336, 0.7501, 0.1612, 1.0]], 60: [[0.0138, 0.0165, 0.0183], [0.7101, 0.7613, 0.5269, 1.0]], 61: [[0.0149, 0.0193, 0.0181], [0.2383, 0.8573, 0.657, 1.0]], 62: [[0.0205, 0.0462, 0.0483], [0.8779, 0.876, 0.5988, 1.0]], 63: [[0.0126, 0.0485, 0.0498], [0.2322, 0.7434, 0.3378, 1.0]], 64: [[0.0208, 0.0166, 0.0175], [0.5481, 0.3963, 0.2458, 1.0]], 65: [[0.0123, 0.0458, 0.0473], [0.8363, 0.584, 0.8345, 1.0]], 66: [[0.0152, 0.0461, 0.0488], [0.2852, 0.6593, 0.4106, 1.0]], 67: [[0.0175, 0.018, 0.0156], [0.1591, 0.8999, 0.1738, 1.0]], 68: [[0.0104, 0.0167, 0.0164], [0.8758, 0.6222, 0.7088, 1.0]], 69: [[0.0161, 0.0456, 0.0488], [0.2751, 0.6347, 0.1566, 1.0]], 70: [[0.0118, 0.0183, 0.0177], [0.2599, 0.4178, 0.5191, 1.0]], 71: [[0.0169, 0.0484, 0.0471], [0.6084, 0.1645, 0.8656, 1.0]], 72: [[0.0144, 0.0191, 0.0158], [0.3274, 0.1633, 0.1479, 1.0]], 73: [[0.0196, 0.0182, 0.0178], [0.2689, 0.1003, 0.857, 1.0]], 74: [[0.0176, 0.047, 0.0498], [0.2583, 0.8694, 0.2129, 1.0]], 75: [[0.0216, 0.0166, 0.0178], [0.5812, 0.4521, 0.2085, 1.0]], 76: [[0.0193, 0.0184, 0.0176], [0.2953, 0.1088, 0.1919, 1.0]], 77: [[0.0181, 0.0194, 0.0178], [0.2713, 0.1443, 0.2227, 1.0]], 78: [[0.0164, 0.0156, 0.0177], [0.2967, 0.2742, 0.7264, 1.0]], 79: [[0.0219, 0.049, 0.0496], [0.58, 0.3954, 0.3932, 1.0]], 80: [[0.0128, 0.049, 0.0457], [0.7774, 0.1565, 0.7805, 1.0]], 81: [[0.0124, 0.0489, 0.0468], [0.4525, 0.3835, 0.832, 1.0]], 82: [[0.0186, 0.0486, 0.0464], [0.8187, 0.6362, 0.1601, 1.0]], 83: [[0.0206, 0.046, 0.0488], [0.623, 0.8552, 0.8165, 1.0]], 84: [[0.0117, 0.0165, 0.0153], [0.4809, 0.7749, 0.8775, 1.0]], 85: [[0.0213, 0.0459, 0.0464], [0.4702, 0.4813, 0.3434, 1.0]], 86: [[0.019, 0.045, 0.0475], [0.7441, 0.8441, 0.8382, 1.0]], 87: [[0.0122, 0.0151, 0.0156], [0.2572, 0.4567, 0.6492, 1.0]], 88: [[0.0215, 0.017, 0.0172], [0.2149, 0.479, 0.4043, 1.0]], 89: [[0.0115, 0.0182, 0.0173], [0.1044, 0.857, 0.6485, 1.0]], 90: [[0.0153, 0.0463, 0.0498], [0.3174, 0.8098, 0.6281, 1.0]], 91: [[0.0149, 0.0485, 0.0462], [0.811, 0.6085, 0.3801, 1.0]], 92: [[0.0164, 0.0167, 0.0152], [0.4186, 0.4523, 0.6883, 1.0]], 93: [[0.012, 0.045, 0.048], [0.7487, 0.1029, 0.1244, 1.0]], 94: [[0.012, 0.0158, 0.0167], [0.6503, 0.2879, 0.7908, 1.0]], 95: [[0.0139, 0.0474, 0.045], [0.5876, 0.8936, 0.6603, 1.0]], 96: [[0.0161, 0.0473, 0.0474], [0.8403, 0.7127, 0.4783, 1.0]], 97: [[0.0194, 0.0184, 0.0176], [0.2173, 0.1284, 0.533, 1.0]], 98: [[0.0213, 0.0478, 0.0486], [0.6218, 0.1434, 0.7881, 1.0]], 99: [[0.0187, 0.0457, 0.0454], [0.701, 0.5286, 0.1528, 1.0]], 100: [[0.0212, 0.0164, 0.0173], [0.1359, 0.3418, 0.1702, 1.0]], 101: [[0.0115, 0.0494, 0.0474], [0.1115, 0.6422, 0.2154, 1.0]], 102: [[0.0201, 0.0482, 0.0467], [0.3541, 0.5037, 0.1197, 1.0]], 103: [[0.0113, 0.018, 0.018], [0.2665, 0.6401, 0.5914, 1.0]], 104: [[0.0186, 0.0452, 0.0464], [0.5232, 0.5509, 0.4378, 1.0]], 105: [[0.016, 0.0453, 0.0474], [0.3037, 0.1085, 0.4448, 1.0]], 106: [[0.0173, 0.0455, 0.0477], [0.7182, 0.1459, 0.3942, 1.0]], 107: [[0.0202, 0.0184, 0.0151], [0.4666, 0.5228, 0.4178, 1.0]], 108: [[0.0172, 0.018, 0.0193], [0.2233, 0.215, 0.2582, 1.0]], 109: [[0.0167, 0.0175, 0.0187], [0.8508, 0.2018, 0.5478, 1.0]], 110: [[0.0108, 0.0175, 0.0179], [0.5524, 0.4684, 0.4966, 1.0]], 111: [[0.0143, 0.0191, 0.0193], [0.7332, 0.8646, 0.5665, 1.0]], 112: [[0.0139, 0.0482, 0.0454], [0.3081, 0.1629, 0.2863, 1.0]], 113: [[0.0147, 0.0473, 0.0491], [0.3088, 0.7483, 0.6019, 1.0]], 114: [[0.0185, 0.0466, 0.05], [0.116, 0.1138, 0.7998, 1.0]], 115: [[0.011, 0.0192, 0.0193], [0.7837, 0.3943, 0.6835, 1.0]], 116: [[0.0131, 0.0178, 0.0177], [0.8052, 0.3264, 0.2248, 1.0]], 117: [[0.0117, 0.0182, 0.02], [0.3734, 0.845, 0.856, 1.0]], 118: [[0.0141, 0.0185, 0.017], [0.7165, 0.2071, 0.2128, 1.0]], 119: [[0.0138, 0.0462, 0.0474], [0.1744, 0.39, 0.2223, 1.0]], 120: [[0.0171, 0.0461, 0.0475], [0.3088, 0.2349, 0.4867, 1.0]], 121: [[0.0117, 0.016, 0.0181], [0.2226, 0.8378, 0.3225, 1.0]], 122: [[0.0105, 0.0158, 0.0171], [0.8364, 0.7014, 0.8305, 1.0]], 123: [[0.0189, 0.0165, 0.0187], [0.297, 0.3524, 0.7998, 1.0]], 124: [[0.0215, 0.0461, 0.045], [0.4239, 0.1262, 0.3183, 1.0]], 125: [[0.0172, 0.0153, 0.0164], [0.1557, 0.6216, 0.3463, 1.0]], 126: [[0.0103, 0.0156, 0.0183], [0.7072, 0.1275, 0.6285, 1.0]], 127: [[0.0181, 0.0159, 0.016], [0.237, 0.2653, 0.556, 1.0]], 128: [[0.0173, 0.0464, 0.0495], [0.4488, 0.2617, 0.806, 1.0]], 129: [[0.0189, 0.0179, 0.0169], [0.7203, 0.2278, 0.7525, 1.0]], 130: [[0.0123, 0.0181, 0.0176], [0.2859, 0.6673, 0.7392, 1.0]], 131: [[0.0195, 0.019, 0.0177], [0.437, 0.4429, 0.1932, 1.0]], 132: [[0.0116, 0.019, 0.0183], [0.4918, 0.1073, 0.5199, 1.0]], 133: [[0.0104, 0.0486, 0.0455], [0.4454, 0.6188, 0.4804, 1.0]], 134: [[0.0147, 0.0182, 0.0187], [0.8536, 0.2603, 0.6406, 1.0]], 135: [[0.0101, 0.0155, 0.0154], [0.6714, 0.4946, 0.3238, 1.0]], 136: [[0.021, 0.0174, 0.0163], [0.6446, 0.837, 0.5574, 1.0]], 137: [[0.0213, 0.049, 0.0485], [0.4923, 0.1998, 0.6318, 1.0]], 138: [[0.0216, 0.0158, 0.0157], [0.7656, 0.8456, 0.4701, 1.0]], 139: [[0.0211, 0.0457, 0.0455], [0.8384, 0.2285, 0.1857, 1.0]], 140: [[0.0153, 0.0462, 0.049], [0.4427, 0.8192, 0.1261, 1.0]], 141: [[0.0173, 0.0486, 0.0458], [0.1271, 0.7625, 0.6444, 1.0]], 142: [[0.0111, 0.0167, 0.0181], [0.2497, 0.3426, 0.5949, 1.0]], 143: [[0.0129, 0.0162, 0.0176], [0.1535, 0.4215, 0.1287, 1.0]], 144: [[0.0173, 0.05, 0.0454], [0.7078, 0.6966, 0.7814, 1.0]], 145: [[0.0173, 0.0173, 0.0189], [0.8446, 0.8726, 0.1454, 1.0]], 146: [[0.0175, 0.0467, 0.0452], [0.3159, 0.2075, 0.4135, 1.0]], 147: [[0.0189, 0.0472, 0.0489], [0.3509, 0.7609, 0.247, 1.0]], 148: [[0.011, 0.0473, 0.0477], [0.5892, 0.7222, 0.7374, 1.0]], 149: [[0.0129, 0.0474, 0.0453], [0.4001, 0.3362, 0.3461, 1.0]], 150: [[0.0147, 0.0492, 0.0495], [0.3652, 0.3845, 0.7267, 1.0]], 151: [[0.0177, 0.0455, 0.048], [0.3723, 0.8321, 0.8714, 1.0]], 152: [[0.0166, 0.0487, 0.0466], [0.4703, 0.3864, 0.4177, 1.0]], 153: [[0.0153, 0.0499, 0.0499], [0.7648, 0.7416, 0.7357, 1.0]], 154: [[0.0134, 0.0493, 0.0457], [0.3237, 0.6859, 0.2325, 1.0]], 155: [[0.018, 0.018, 0.0197], [0.7054, 0.213, 0.2625, 1.0]], 156: [[0.0162, 0.0452, 0.0477], [0.6377, 0.6928, 0.5753, 1.0]], 157: [[0.0148, 0.046, 0.0469], [0.5586, 0.8162, 0.6545, 1.0]], 158: [[0.0211, 0.0157, 0.0184], [0.8362, 0.3753, 0.7235, 1.0]], 159: [[0.0214, 0.0196, 0.0156], [0.5081, 0.4477, 0.5708, 1.0]], 160: [[0.0205, 0.0492, 0.0497], [0.1427, 0.7907, 0.8017, 1.0]], 161: [[0.0133, 0.0166, 0.0178], [0.7213, 0.3542, 0.781, 1.0]], 162: [[0.0174, 0.0483, 0.0453], [0.449, 0.4152, 0.215, 1.0]], 163: [[0.0118, 0.0183, 0.0173], [0.8446, 0.4294, 0.1347, 1.0]], 164: [[0.0127, 0.0193, 0.0175], [0.5537, 0.6098, 0.4473, 1.0]], 165: [[0.0109, 0.046, 0.045], [0.8556, 0.5505, 0.5512, 1.0]], 166: [[0.0147, 0.0457, 0.0481], [0.2713, 0.7413, 0.2877, 1.0]], 167: [[0.0144, 0.0462, 0.0459], [0.165, 0.5973, 0.8201, 1.0]], 168: [[0.0212, 0.0157, 0.0171], [0.214, 0.1838, 0.1647, 1.0]], 169: [[0.019, 0.0468, 0.0494], [0.6828, 0.5465, 0.1629, 1.0]], 170: [[0.0185, 0.0177, 0.0197], [0.5079, 0.5769, 0.7325, 1.0]], 171: [[0.0203, 0.0477, 0.0491], [0.2193, 0.22, 0.6761, 1.0]], 172: [[0.0103, 0.0186, 0.0178], [0.7868, 0.5255, 0.352, 1.0]], 173: [[0.0196, 0.0184, 0.0155], [0.4391, 0.3064, 0.5309, 1.0]], 174: [[0.0189, 0.0452, 0.0459], [0.6549, 0.5861, 0.727, 1.0]], 175: [[0.0117, 0.0485, 0.0461], [0.1796, 0.2544, 0.4843, 1.0]], 176: [[0.0138, 0.0162, 0.0169], [0.7049, 0.7908, 0.8643, 1.0]], 177: [[0.0165, 0.0452, 0.0477], [0.6733, 0.1488, 0.204, 1.0]], 178: [[0.0166, 0.0184, 0.0165], [0.788, 0.1421, 0.3831, 1.0]], 179: [[0.013, 0.0183, 0.0167], [0.2977, 0.5449, 0.3888, 1.0]], 180: [[0.0154, 0.0172, 0.0179], [0.4412, 0.8275, 0.5608, 1.0]], 181: [[0.0104, 0.0196, 0.016], [0.1381, 0.1568, 0.6259, 1.0]], 182: [[0.0133, 0.0475, 0.0488], [0.1533, 0.3559, 0.1507, 1.0]], 183: [[0.0155, 0.0185, 0.0164], [0.5281, 0.6594, 0.1726, 1.0]], 184: [[0.0211, 0.0485, 0.0467], [0.8173, 0.4514, 0.5882, 1.0]], 185: [[0.0143, 0.0153, 0.0158], [0.8264, 0.3948, 0.6738, 1.0]], 186: [[0.0109, 0.0493, 0.0474], [0.8409, 0.4085, 0.5155, 1.0]], 187: [[0.0128, 0.0192, 0.0157], [0.5373, 0.433, 0.8742, 1.0]], 188: [[0.0209, 0.0455, 0.0498], [0.1971, 0.8868, 0.2746, 1.0]], 189: [[0.018, 0.0454, 0.0469], [0.4346, 0.538, 0.2588, 1.0]], 190: [[0.0116, 0.0165, 0.0163], [0.4802, 0.811, 0.4269, 1.0]], 191: [[0.0109, 0.018, 0.0163], [0.7326, 0.4583, 0.8561, 1.0]], 192: [[0.014, 0.0166, 0.0167], [0.2543, 0.5965, 0.3042, 1.0]], 193: [[0.013, 0.0166, 0.0165], [0.5034, 0.7119, 0.4114, 1.0]], 194: [[0.0177, 0.0182, 0.0182], [0.4691, 0.8266, 0.8231, 1.0]], 195: [[0.018, 0.0498, 0.0485], [0.426, 0.4928, 0.5865, 1.0]], 196: [[0.0108, 0.0481, 0.0478], [0.4718, 0.469, 0.2384, 1.0]], 197: [[0.0208, 0.0499, 0.0488], [0.5592, 0.7773, 0.8967, 1.0]], 198: [[0.0115, 0.0458, 0.0462], [0.4334, 0.1846, 0.3883, 1.0]], 199: [[0.0206, 0.0456, 0.0495], [0.6058, 0.7777, 0.429, 1.0]], 200: [[0.0177, 0.0464, 0.0472], [0.1362, 0.1824, 0.5835, 1.0]], 201: [[0.014, 0.0191, 0.0181], [0.7206, 0.3437, 0.37, 1.0]], 202: [[0.0118, 0.0474, 0.0455], [0.1148, 0.262, 0.4485, 1.0]], 203: [[0.0188, 0.0455, 0.0477], [0.5114, 0.8115, 0.7676, 1.0]], 204: [[0.0172, 0.0499, 0.0462], [0.5308, 0.8111, 0.4337, 1.0]], 205: [[0.0103, 0.0178, 0.0162], [0.1149, 0.2917, 0.5876, 1.0]], 206: [[0.0153, 0.0476, 0.0499], [0.4999, 0.5089, 0.4157, 1.0]], 207: [[0.0102, 0.0171, 0.0199], [0.2822, 0.6051, 0.1381, 1.0]], 208: [[0.0117, 0.0166, 0.0161], [0.628, 0.8254, 0.2762, 1.0]], 209: [[0.0214, 0.0191, 0.0198], [0.7102, 0.6143, 0.7504, 1.0]], 210: [[0.011, 0.0197, 0.0181], [0.4121, 0.5831, 0.6519, 1.0]], 211: [[0.0112, 0.0482, 0.0489], [0.5244, 0.4626, 0.3611, 1.0]], 212: [[0.0207, 0.0452, 0.049], [0.5354, 0.5679, 0.6929, 1.0]], 213: [[0.0181, 0.0178, 0.0199], [0.1985, 0.8685, 0.6615, 1.0]], 214: [[0.0161, 0.018, 0.0184], [0.8486, 0.384, 0.3626, 1.0]], 215: [[0.014, 0.0186, 0.0164], [0.8014, 0.1281, 0.8958, 1.0]], 216: [[0.0186, 0.0174, 0.0167], [0.2922, 0.3107, 0.4267, 1.0]], 217: [[0.0146, 0.0156, 0.0177], [0.8846, 0.8364, 0.6047, 1.0]], 218: [[0.0112, 0.047, 0.0465], [0.5951, 0.6992, 0.1147, 1.0]], 219: [[0.0158, 0.0489, 0.0452], [0.2499, 0.7388, 0.8561, 1.0]], 220: [[0.0106, 0.048, 0.0469], [0.7771, 0.1326, 0.4109, 1.0]], 221: [[0.0214, 0.0464, 0.0482], [0.3484, 0.482, 0.6405, 1.0]], 222: [[0.0114, 0.0465, 0.0492], [0.2858, 0.3632, 0.6027, 1.0]], 223: [[0.0153, 0.0173, 0.0154], [0.4912, 0.2048, 0.5629, 1.0]], 224: [[0.0109, 0.016, 0.0168], [0.3627, 0.4122, 0.7033, 1.0]], 225: [[0.0105, 0.0492, 0.046], [0.5845, 0.6587, 0.7785, 1.0]], 226: [[0.0215, 0.0199, 0.0167], [0.6111, 0.4582, 0.5161, 1.0]], 227: [[0.0177, 0.0452, 0.0487], [0.887, 0.7108, 0.8265, 1.0]], 228: [[0.0182, 0.0454, 0.0477], [0.4018, 0.2768, 0.6919, 1.0]], 229: [[0.0141, 0.0466, 0.0481], [0.6269, 0.4912, 0.5564, 1.0]], 230: [[0.0191, 0.0195, 0.0175], [0.428, 0.3875, 0.1871, 1.0]], 231: [[0.0174, 0.0491, 0.048], [0.3444, 0.1266, 0.8867, 1.0]], 232: [[0.0186, 0.0159, 0.0187], [0.2639, 0.3621, 0.5915, 1.0]], 233: [[0.0209, 0.0193, 0.0156], [0.7967, 0.8694, 0.8187, 1.0]], 234: [[0.0154, 0.0191, 0.0199], [0.2638, 0.3083, 0.3342, 1.0]], 235: [[0.0143, 0.0464, 0.0469], [0.7041, 0.4356, 0.8954, 1.0]], 236: [[0.0161, 0.0185, 0.0166], [0.1647, 0.6814, 0.3447, 1.0]], 237: [[0.0113, 0.0493, 0.0467], [0.6876, 0.7429, 0.7811, 1.0]], 238: [[0.0156, 0.0473, 0.0485], [0.5644, 0.27, 0.7321, 1.0]], 239: [[0.0163, 0.0461, 0.049], [0.1809, 0.2318, 0.6899, 1.0]], 240: [[0.0152, 0.0492, 0.0493], [0.2864, 0.6311, 0.2679, 1.0]], 241: [[0.015, 0.0162, 0.0169], [0.7129, 0.8796, 0.4358, 1.0]], 242: [[0.0131, 0.0483, 0.0494], [0.1308, 0.4755, 0.1302, 1.0]], 243: [[0.0208, 0.0183, 0.0167], [0.5402, 0.816, 0.8953, 1.0]], 244: [[0.0219, 0.0188, 0.0188], [0.4618, 0.2172, 0.1317, 1.0]], 245: [[0.0144, 0.0157, 0.0165], [0.7438, 0.3295, 0.4762, 1.0]], 246: [[0.0214, 0.0479, 0.047], [0.2621, 0.1049, 0.6837, 1.0]], 247: [[0.015, 0.0486, 0.0472], [0.1136, 0.1078, 0.1676, 1.0]], 248: [[0.0202, 0.0474, 0.0472], [0.5658, 0.7594, 0.5825, 1.0]], 249: [[0.0143, 0.0171, 0.017], [0.1856, 0.2551, 0.2272, 1.0]]}

        caps =  {0: [[0.014, 0.0462, 0.0456], [0.7552, 0.5999, 0.1518, 1.0]], 1: [[0.0216, 0.0457, 0.0467], [0.2093, 0.2803, 0.6963, 1.0]], 2: [[0.0164, 0.0488, 0.0469], [0.533, 0.7576, 0.3658, 1.0]], 3: [[0.0155, 0.0467, 0.0474], [0.7339, 0.4907, 0.8917, 1.0]], 4: [[0.018, 0.05, 0.0487], [0.7503, 0.278, 0.1356, 1.0]], 5: [[0.0124, 0.0484, 0.0491], [0.6518, 0.8439, 0.2416, 1.0]], 6: [[0.0197, 0.0483, 0.0459], [0.1324, 0.4663, 0.3485, 1.0]], 7: [[0.0129, 0.0476, 0.0495], [0.2285, 0.6781, 0.8076, 1.0]], 8: [[0.0135, 0.0455, 0.0458], [0.6364, 0.4566, 0.7004, 1.0]], 9: [[0.0193, 0.0485, 0.0487], [0.2589, 0.7651, 0.819, 1.0]], 10: [[0.0159, 0.0461, 0.0485], [0.3405, 0.7015, 0.4264, 1.0]], 11: [[0.0124, 0.0465, 0.0454], [0.4863, 0.8691, 0.3991, 1.0]], 12: [[0.0129, 0.046, 0.0493], [0.3887, 0.4909, 0.2819, 1.0]], 13: [[0.0152, 0.0494, 0.048], [0.336, 0.572, 0.237, 1.0]], 14: [[0.017, 0.0463, 0.049], [0.1482, 0.6936, 0.3843, 1.0]], 15: [[0.0212, 0.0489, 0.0472], [0.1987, 0.7815, 0.2925, 1.0]], 16: [[0.0159, 0.0454, 0.0494], [0.4086, 0.5416, 0.2677, 1.0]], 17: [[0.0209, 0.0484, 0.0461], [0.6409, 0.6521, 0.4668, 1.0]], 18: [[0.0117, 0.0461, 0.0493], [0.1992, 0.4997, 0.1159, 1.0]], 19: [[0.0167, 0.0474, 0.0474], [0.2572, 0.8054, 0.2026, 1.0]], 20: [[0.0176, 0.0479, 0.0453], [0.8732, 0.411, 0.7781, 1.0]], 21: [[0.0131, 0.0483, 0.05], [0.7602, 0.7548, 0.5154, 1.0]], 22: [[0.0127, 0.047, 0.0495], [0.1458, 0.7376, 0.6436, 1.0]], 23: [[0.0185, 0.0454, 0.048], [0.4824, 0.429, 0.2692, 1.0]], 24: [[0.0183, 0.0467, 0.0467], [0.2042, 0.1897, 0.5394, 1.0]], 25: [[0.0166, 0.0491, 0.0487], [0.2161, 0.8852, 0.116, 1.0]], 26: [[0.0155, 0.0456, 0.0485], [0.4469, 0.7486, 0.2039, 1.0]], 27: [[0.017, 0.0489, 0.0491], [0.2923, 0.296, 0.463, 1.0]], 28: [[0.0172, 0.0469, 0.0488], [0.5067, 0.8173, 0.2359, 1.0]], 29: [[0.0114, 0.0467, 0.0487], [0.7463, 0.4096, 0.2019, 1.0]], 30: [[0.0156, 0.0491, 0.046], [0.5904, 0.1744, 0.3366, 1.0]], 31: [[0.0176, 0.0478, 0.0451], [0.3127, 0.827, 0.4654, 1.0]], 32: [[0.0164, 0.0476, 0.0455], [0.5944, 0.7672, 0.201, 1.0]], 33: [[0.0198, 0.0459, 0.0474], [0.1264, 0.7821, 0.2798, 1.0]], 34: [[0.0191, 0.0456, 0.0482], [0.8076, 0.7139, 0.8896, 1.0]], 35: [[0.0165, 0.0494, 0.0481], [0.8064, 0.5783, 0.2943, 1.0]], 36: [[0.0153, 0.0464, 0.0482], [0.7696, 0.8593, 0.8418, 1.0]], 37: [[0.013, 0.0491, 0.0486], [0.7956, 0.5799, 0.2945, 1.0]], 38: [[0.0164, 0.0481, 0.0478], [0.1812, 0.6424, 0.6937, 1.0]], 39: [[0.0207, 0.0492, 0.048], [0.4824, 0.3024, 0.5251, 1.0]], 40: [[0.0189, 0.0496, 0.0465], [0.1554, 0.181, 0.8387, 1.0]], 41: [[0.0143, 0.0473, 0.0488], [0.2187, 0.5529, 0.1564, 1.0]], 42: [[0.0135, 0.0489, 0.0485], [0.2515, 0.7067, 0.6655, 1.0]], 43: [[0.0152, 0.0465, 0.0491], [0.1965, 0.7194, 0.3763, 1.0]], 44: [[0.0178, 0.0467, 0.0491], [0.6523, 0.6147, 0.4241, 1.0]], 45: [[0.0163, 0.0497, 0.0474], [0.106, 0.4336, 0.751, 1.0]], 46: [[0.013, 0.0455, 0.0497], [0.127, 0.8312, 0.266, 1.0]], 47: [[0.0144, 0.0491, 0.0497], [0.4803, 0.174, 0.3265, 1.0]], 48: [[0.0165, 0.0495, 0.0468], [0.3985, 0.1923, 0.5643, 1.0]], 49: [[0.0108, 0.0464, 0.0464], [0.8164, 0.1074, 0.6519, 1.0]], 50: [[0.0125, 0.0496, 0.0476], [0.3726, 0.3997, 0.7442, 1.0]], 51: [[0.0202, 0.049, 0.0452], [0.6754, 0.1224, 0.4845, 1.0]], 52: [[0.0152, 0.0491, 0.047], [0.4577, 0.1851, 0.8076, 1.0]], 53: [[0.0157, 0.047, 0.0469], [0.7069, 0.7766, 0.7214, 1.0]], 54: [[0.0164, 0.0491, 0.0478], [0.3279, 0.7244, 0.7102, 1.0]], 55: [[0.0215, 0.0495, 0.0483], [0.432, 0.3687, 0.2987, 1.0]], 56: [[0.0111, 0.0474, 0.0478], [0.4287, 0.5973, 0.2854, 1.0]], 57: [[0.0198, 0.047, 0.047], [0.6354, 0.6947, 0.3978, 1.0]], 58: [[0.0157, 0.045, 0.0482], [0.7976, 0.8864, 0.5672, 1.0]], 59: [[0.0145, 0.0473, 0.0467], [0.7185, 0.634, 0.3194, 1.0]], 60: [[0.0136, 0.0457, 0.0498], [0.6684, 0.8659, 0.6903, 1.0]], 61: [[0.0126, 0.0458, 0.0499], [0.582, 0.6602, 0.855, 1.0]], 62: [[0.0112, 0.0471, 0.0475], [0.7666, 0.4851, 0.2647, 1.0]], 63: [[0.0131, 0.0476, 0.0482], [0.8042, 0.4006, 0.1974, 1.0]], 64: [[0.0209, 0.0452, 0.05], [0.68, 0.7969, 0.8836, 1.0]], 65: [[0.0161, 0.0473, 0.048], [0.6047, 0.4247, 0.859, 1.0]], 66: [[0.0125, 0.0494, 0.048], [0.1788, 0.8538, 0.4675, 1.0]], 67: [[0.0118, 0.0461, 0.049], [0.7032, 0.7573, 0.5692, 1.0]], 68: [[0.0135, 0.049, 0.0475], [0.1252, 0.7411, 0.6948, 1.0]], 69: [[0.019, 0.0477, 0.0451], [0.6571, 0.5464, 0.3748, 1.0]], 70: [[0.0169, 0.0491, 0.0474], [0.4295, 0.4868, 0.5209, 1.0]], 71: [[0.0145, 0.0468, 0.0468], [0.4492, 0.7508, 0.8496, 1.0]], 72: [[0.0118, 0.0485, 0.0469], [0.7431, 0.3396, 0.4076, 1.0]], 73: [[0.021, 0.0482, 0.0476], [0.1269, 0.2122, 0.6914, 1.0]], 74: [[0.0179, 0.0479, 0.0475], [0.2867, 0.8822, 0.2719, 1.0]], 75: [[0.0155, 0.0473, 0.0485], [0.1142, 0.8501, 0.6702, 1.0]], 76: [[0.0106, 0.0493, 0.0461], [0.7289, 0.8452, 0.7046, 1.0]], 77: [[0.0105, 0.0451, 0.0481], [0.7829, 0.3362, 0.4289, 1.0]], 78: [[0.0111, 0.0452, 0.0451], [0.8546, 0.3946, 0.6141, 1.0]], 79: [[0.013, 0.0473, 0.0464], [0.7705, 0.1568, 0.4001, 1.0]], 80: [[0.0109, 0.0476, 0.0464], [0.7354, 0.5219, 0.8647, 1.0]], 81: [[0.0139, 0.047, 0.0487], [0.656, 0.3886, 0.8112, 1.0]], 82: [[0.0191, 0.0496, 0.0454], [0.2665, 0.2847, 0.8008, 1.0]], 83: [[0.013, 0.0486, 0.0474], [0.2069, 0.1691, 0.6395, 1.0]], 84: [[0.0179, 0.0463, 0.0495], [0.4257, 0.7388, 0.7933, 1.0]], 85: [[0.0194, 0.047, 0.0455], [0.8254, 0.4672, 0.3752, 1.0]], 86: [[0.0196, 0.0451, 0.0464], [0.3792, 0.8519, 0.6723, 1.0]], 87: [[0.015, 0.0492, 0.0453], [0.2409, 0.1773, 0.7596, 1.0]], 88: [[0.0172, 0.0456, 0.0471], [0.3939, 0.6507, 0.3653, 1.0]], 89: [[0.0186, 0.0496, 0.0486], [0.5724, 0.1403, 0.5072, 1.0]], 90: [[0.0169, 0.0484, 0.0484], [0.1981, 0.3563, 0.7455, 1.0]], 91: [[0.0216, 0.0461, 0.0472], [0.7385, 0.2926, 0.8827, 1.0]], 92: [[0.0172, 0.0487, 0.0476], [0.1522, 0.3485, 0.5212, 1.0]], 93: [[0.0113, 0.0455, 0.0489], [0.1023, 0.8846, 0.1536, 1.0]], 94: [[0.0162, 0.046, 0.0458], [0.671, 0.378, 0.2139, 1.0]], 95: [[0.0189, 0.0452, 0.0465], [0.5306, 0.4105, 0.8137, 1.0]], 96: [[0.0177, 0.0466, 0.0457], [0.4245, 0.1615, 0.7385, 1.0]], 97: [[0.019, 0.0471, 0.046], [0.4411, 0.3306, 0.8051, 1.0]], 98: [[0.0213, 0.0463, 0.0461], [0.6587, 0.8677, 0.6339, 1.0]], 99: [[0.0129, 0.0498, 0.0495], [0.2374, 0.1145, 0.2975, 1.0]], 100: [[0.0119, 0.0466, 0.0457], [0.1556, 0.7113, 0.3381, 1.0]], 101: [[0.0155, 0.0454, 0.0468], [0.3652, 0.3653, 0.1902, 1.0]], 102: [[0.0119, 0.0488, 0.0498], [0.8815, 0.2709, 0.3494, 1.0]], 103: [[0.0142, 0.0455, 0.0473], [0.5871, 0.6631, 0.3971, 1.0]], 104: [[0.0123, 0.0473, 0.0489], [0.2607, 0.5424, 0.5473, 1.0]], 105: [[0.0161, 0.0482, 0.0461], [0.6337, 0.3421, 0.5175, 1.0]], 106: [[0.018, 0.0455, 0.0494], [0.2013, 0.3448, 0.8066, 1.0]], 107: [[0.0166, 0.0479, 0.0462], [0.4087, 0.2224, 0.206, 1.0]], 108: [[0.0149, 0.0469, 0.0474], [0.5612, 0.6575, 0.8316, 1.0]], 109: [[0.0142, 0.0474, 0.0489], [0.214, 0.8244, 0.3106, 1.0]], 110: [[0.0141, 0.049, 0.0475], [0.355, 0.3569, 0.2908, 1.0]], 111: [[0.0138, 0.0494, 0.0484], [0.508, 0.2495, 0.7607, 1.0]], 112: [[0.0162, 0.0476, 0.0496], [0.113, 0.7339, 0.1047, 1.0]], 113: [[0.0142, 0.0457, 0.0487], [0.897, 0.5763, 0.7634, 1.0]], 114: [[0.0149, 0.0464, 0.0482], [0.5559, 0.3195, 0.2474, 1.0]], 115: [[0.0142, 0.0462, 0.0452], [0.3724, 0.3881, 0.5863, 1.0]], 116: [[0.0164, 0.0459, 0.0466], [0.4296, 0.3586, 0.2608, 1.0]], 117: [[0.018, 0.049, 0.0472], [0.3819, 0.6104, 0.47, 1.0]], 118: [[0.0118, 0.0499, 0.0475], [0.345, 0.5759, 0.5034, 1.0]], 119: [[0.017, 0.0474, 0.0471], [0.1843, 0.7297, 0.5133, 1.0]], 120: [[0.0109, 0.0491, 0.0466], [0.1814, 0.7445, 0.3425, 1.0]], 121: [[0.0116, 0.0491, 0.0495], [0.4409, 0.6216, 0.359, 1.0]], 122: [[0.0138, 0.0495, 0.0492], [0.3301, 0.1702, 0.5307, 1.0]], 123: [[0.012, 0.0475, 0.0497], [0.1439, 0.1612, 0.4065, 1.0]], 124: [[0.0214, 0.0485, 0.0469], [0.5616, 0.2181, 0.7746, 1.0]], 125: [[0.0121, 0.0493, 0.0465], [0.801, 0.4947, 0.2195, 1.0]], 126: [[0.0211, 0.0456, 0.0457], [0.7768, 0.3856, 0.8558, 1.0]], 127: [[0.019, 0.0483, 0.0464], [0.4195, 0.4983, 0.3311, 1.0]], 128: [[0.019, 0.0454, 0.0484], [0.1451, 0.2257, 0.3421, 1.0]], 129: [[0.0144, 0.0463, 0.0474], [0.275, 0.7622, 0.4729, 1.0]], 130: [[0.0191, 0.0474, 0.0495], [0.8404, 0.1813, 0.4804, 1.0]], 131: [[0.0161, 0.0465, 0.048], [0.8094, 0.7558, 0.363, 1.0]], 132: [[0.0198, 0.0463, 0.0487], [0.4962, 0.5524, 0.6939, 1.0]], 133: [[0.0132, 0.0459, 0.0481], [0.7799, 0.3286, 0.59, 1.0]], 134: [[0.019, 0.0456, 0.0468], [0.5001, 0.7448, 0.146, 1.0]], 135: [[0.0216, 0.0499, 0.0476], [0.681, 0.1947, 0.5377, 1.0]], 136: [[0.0171, 0.0478, 0.0495], [0.2267, 0.2009, 0.3166, 1.0]], 137: [[0.0168, 0.0478, 0.0476], [0.1241, 0.8835, 0.1198, 1.0]], 138: [[0.0166, 0.0477, 0.0466], [0.8359, 0.8665, 0.589, 1.0]], 139: [[0.0136, 0.0497, 0.0484], [0.641, 0.8936, 0.1003, 1.0]], 140: [[0.0124, 0.048, 0.0458], [0.8078, 0.3435, 0.7498, 1.0]], 141: [[0.0107, 0.0464, 0.0455], [0.5088, 0.6734, 0.2235, 1.0]], 142: [[0.0211, 0.0452, 0.0461], [0.8815, 0.8621, 0.3036, 1.0]], 143: [[0.0143, 0.047, 0.0489], [0.5841, 0.5472, 0.5041, 1.0]], 144: [[0.0163, 0.0474, 0.0493], [0.3747, 0.4394, 0.3849, 1.0]], 145: [[0.0146, 0.0495, 0.0453], [0.5791, 0.8076, 0.8573, 1.0]], 146: [[0.0134, 0.0477, 0.0459], [0.8958, 0.3424, 0.784, 1.0]], 147: [[0.0137, 0.0476, 0.0451], [0.5203, 0.5384, 0.1478, 1.0]], 148: [[0.0137, 0.046, 0.0485], [0.4583, 0.8113, 0.1408, 1.0]], 149: [[0.0138, 0.0466, 0.0492], [0.2218, 0.6236, 0.4278, 1.0]], 150: [[0.0195, 0.0486, 0.0481], [0.8044, 0.8925, 0.5465, 1.0]], 151: [[0.0107, 0.0489, 0.0459], [0.2298, 0.5506, 0.7442, 1.0]], 152: [[0.0131, 0.049, 0.0496], [0.5106, 0.5685, 0.2025, 1.0]], 153: [[0.0179, 0.0462, 0.05], [0.5488, 0.4512, 0.5557, 1.0]], 154: [[0.0167, 0.0452, 0.0468], [0.2355, 0.5682, 0.3305, 1.0]], 155: [[0.0145, 0.0494, 0.046], [0.7485, 0.5168, 0.5872, 1.0]], 156: [[0.0161, 0.0486, 0.05], [0.7493, 0.2481, 0.7278, 1.0]], 157: [[0.013, 0.0478, 0.0468], [0.7072, 0.558, 0.6206, 1.0]], 158: [[0.021, 0.0454, 0.0476], [0.3211, 0.7554, 0.1626, 1.0]], 159: [[0.021, 0.0463, 0.046], [0.1345, 0.2543, 0.8641, 1.0]], 160: [[0.0159, 0.0481, 0.0458], [0.8549, 0.3534, 0.7343, 1.0]], 161: [[0.0167, 0.0463, 0.0482], [0.1375, 0.2357, 0.7441, 1.0]], 162: [[0.0196, 0.0465, 0.0463], [0.7306, 0.1254, 0.4492, 1.0]], 163: [[0.0112, 0.0458, 0.0462], [0.3003, 0.211, 0.472, 1.0]], 164: [[0.0107, 0.048, 0.0462], [0.3184, 0.1849, 0.8337, 1.0]], 165: [[0.0101, 0.0487, 0.0451], [0.1582, 0.3694, 0.8754, 1.0]], 166: [[0.0126, 0.0459, 0.047], [0.8508, 0.4215, 0.384, 1.0]], 167: [[0.0185, 0.0453, 0.0499], [0.8908, 0.4287, 0.7101, 1.0]], 168: [[0.01, 0.048, 0.05], [0.1278, 0.6523, 0.1568, 1.0]], 169: [[0.0216, 0.0477, 0.0493], [0.5271, 0.4004, 0.7093, 1.0]], 170: [[0.0156, 0.0485, 0.0491], [0.5416, 0.4668, 0.8501, 1.0]], 171: [[0.0189, 0.0465, 0.0489], [0.1752, 0.8665, 0.4717, 1.0]], 172: [[0.0134, 0.0488, 0.0497], [0.1015, 0.5939, 0.2164, 1.0]], 173: [[0.0129, 0.0462, 0.0454], [0.4864, 0.7432, 0.5212, 1.0]], 174: [[0.0154, 0.049, 0.0461], [0.4718, 0.836, 0.8369, 1.0]], 175: [[0.0119, 0.046, 0.0465], [0.4722, 0.5093, 0.448, 1.0]], 176: [[0.0176, 0.049, 0.0471], [0.5645, 0.1044, 0.1248, 1.0]], 177: [[0.0176, 0.0487, 0.0495], [0.5378, 0.2277, 0.4826, 1.0]], 178: [[0.0118, 0.05, 0.0489], [0.1466, 0.5063, 0.7052, 1.0]], 179: [[0.0206, 0.049, 0.0459], [0.6002, 0.7016, 0.157, 1.0]], 180: [[0.0198, 0.0471, 0.0461], [0.3864, 0.7201, 0.4741, 1.0]], 181: [[0.0195, 0.046, 0.0457], [0.1372, 0.6083, 0.8012, 1.0]], 182: [[0.0207, 0.0463, 0.0488], [0.2979, 0.2253, 0.1311, 1.0]], 183: [[0.0137, 0.0498, 0.0487], [0.6296, 0.5709, 0.6801, 1.0]], 184: [[0.0123, 0.0462, 0.0482], [0.7159, 0.2452, 0.5007, 1.0]], 185: [[0.0144, 0.0474, 0.0495], [0.5067, 0.2733, 0.4707, 1.0]], 186: [[0.0161, 0.047, 0.0499], [0.2278, 0.6304, 0.3455, 1.0]], 187: [[0.0202, 0.0467, 0.0493], [0.3487, 0.6855, 0.2026, 1.0]], 188: [[0.0211, 0.0495, 0.0493], [0.7012, 0.4327, 0.5554, 1.0]], 189: [[0.0164, 0.0491, 0.0474], [0.7327, 0.6326, 0.4928, 1.0]], 190: [[0.0123, 0.0487, 0.0483], [0.1216, 0.8714, 0.3256, 1.0]], 191: [[0.0125, 0.0464, 0.0465], [0.5306, 0.8276, 0.4073, 1.0]], 192: [[0.0189, 0.0452, 0.0466], [0.5741, 0.4903, 0.6112, 1.0]], 193: [[0.0209, 0.0492, 0.0461], [0.3885, 0.3919, 0.7096, 1.0]], 194: [[0.0183, 0.045, 0.049], [0.5015, 0.2727, 0.1527, 1.0]], 195: [[0.0161, 0.0467, 0.0457], [0.8348, 0.8837, 0.803, 1.0]], 196: [[0.0197, 0.0463, 0.0487], [0.4166, 0.2679, 0.4325, 1.0]], 197: [[0.0112, 0.0484, 0.0458], [0.3721, 0.7998, 0.1675, 1.0]], 198: [[0.0209, 0.0483, 0.049], [0.7851, 0.2541, 0.6033, 1.0]], 199: [[0.0164, 0.0465, 0.0463], [0.5876, 0.8114, 0.655, 1.0]], 200: [[0.0178, 0.0455, 0.0457], [0.4663, 0.1764, 0.7163, 1.0]], 201: [[0.0105, 0.0499, 0.048], [0.6529, 0.6654, 0.3025, 1.0]], 202: [[0.0129, 0.0485, 0.0469], [0.5254, 0.738, 0.7322, 1.0]], 203: [[0.0128, 0.0454, 0.0496], [0.6982, 0.7002, 0.372, 1.0]], 204: [[0.0179, 0.0498, 0.0462], [0.5059, 0.5297, 0.4374, 1.0]], 205: [[0.0158, 0.047, 0.0491], [0.2421, 0.4905, 0.3823, 1.0]], 206: [[0.0164, 0.0459, 0.046], [0.879, 0.2691, 0.4236, 1.0]], 207: [[0.0168, 0.0479, 0.0493], [0.6304, 0.1239, 0.1815, 1.0]], 208: [[0.0177, 0.05, 0.0497], [0.2669, 0.6729, 0.1417, 1.0]], 209: [[0.0109, 0.0467, 0.0472], [0.6868, 0.8308, 0.1593, 1.0]], 210: [[0.0163, 0.0468, 0.0483], [0.355, 0.3442, 0.494, 1.0]], 211: [[0.0125, 0.0488, 0.0463], [0.7516, 0.3278, 0.3894, 1.0]], 212: [[0.0201, 0.0486, 0.0478], [0.7729, 0.6964, 0.1113, 1.0]], 213: [[0.0144, 0.0488, 0.0498], [0.5506, 0.4388, 0.1844, 1.0]], 214: [[0.0128, 0.047, 0.0462], [0.2999, 0.2076, 0.79, 1.0]], 215: [[0.0184, 0.0482, 0.0492], [0.1001, 0.616, 0.1964, 1.0]], 216: [[0.0209, 0.0477, 0.0491], [0.299, 0.4896, 0.7637, 1.0]], 217: [[0.0212, 0.0456, 0.0452], [0.2345, 0.863, 0.1339, 1.0]], 218: [[0.0136, 0.047, 0.0484], [0.6714, 0.2591, 0.2313, 1.0]], 219: [[0.013, 0.0484, 0.0496], [0.2711, 0.198, 0.564, 1.0]], 220: [[0.0145, 0.0464, 0.0486], [0.1741, 0.7011, 0.3521, 1.0]], 221: [[0.018, 0.0497, 0.0486], [0.1392, 0.4709, 0.4622, 1.0]], 222: [[0.0158, 0.0489, 0.0462], [0.6808, 0.7848, 0.6046, 1.0]], 223: [[0.019, 0.0499, 0.0484], [0.4008, 0.627, 0.5309, 1.0]], 224: [[0.0182, 0.0475, 0.0471], [0.6352, 0.7478, 0.5901, 1.0]], 225: [[0.0161, 0.0453, 0.0488], [0.4907, 0.5478, 0.322, 1.0]], 226: [[0.0112, 0.0451, 0.0467], [0.6312, 0.8163, 0.7085, 1.0]], 227: [[0.016, 0.0494, 0.0464], [0.4536, 0.2015, 0.3697, 1.0]], 228: [[0.0182, 0.0459, 0.0495], [0.2824, 0.4644, 0.4464, 1.0]], 229: [[0.0129, 0.0459, 0.0469], [0.2873, 0.2099, 0.2296, 1.0]], 230: [[0.0127, 0.0462, 0.0497], [0.8869, 0.2588, 0.1459, 1.0]], 231: [[0.0188, 0.047, 0.0483], [0.3265, 0.4428, 0.7746, 1.0]], 232: [[0.0163, 0.0474, 0.0463], [0.7222, 0.7784, 0.179, 1.0]], 233: [[0.0215, 0.0454, 0.0496], [0.4895, 0.7109, 0.5859, 1.0]], 234: [[0.0205, 0.0475, 0.0491], [0.379, 0.1502, 0.4715, 1.0]], 235: [[0.0177, 0.0493, 0.0493], [0.2987, 0.8402, 0.4573, 1.0]], 236: [[0.0131, 0.0451, 0.0466], [0.7328, 0.2241, 0.1718, 1.0]], 237: [[0.0216, 0.0479, 0.048], [0.6541, 0.8914, 0.8037, 1.0]], 238: [[0.0198, 0.0482, 0.05], [0.6548, 0.1264, 0.2441, 1.0]], 239: [[0.0141, 0.0475, 0.0474], [0.7782, 0.7148, 0.4431, 1.0]], 240: [[0.0186, 0.0458, 0.0492], [0.1186, 0.8537, 0.2014, 1.0]], 241: [[0.0197, 0.0471, 0.0498], [0.2241, 0.227, 0.3461, 1.0]], 242: [[0.0192, 0.0473, 0.0481], [0.7069, 0.7175, 0.5007, 1.0]], 243: [[0.0218, 0.0466, 0.0459], [0.6437, 0.7739, 0.5479, 1.0]], 244: [[0.018, 0.0462, 0.0499], [0.7862, 0.1087, 0.8509, 1.0]], 245: [[0.0188, 0.0499, 0.0468], [0.4566, 0.4582, 0.5993, 1.0]], 246: [[0.0187, 0.0471, 0.0462], [0.4948, 0.6884, 0.8971, 1.0]], 247: [[0.0181, 0.0488, 0.0465], [0.1297, 0.3699, 0.8934, 1.0]], 248: [[0.0156, 0.047, 0.0461], [0.7562, 0.5196, 0.331, 1.0]], 249: [[0.0101, 0.0456, 0.0455], [0.1511, 0.8475, 0.322, 1.0]]}

        cyl = {0: [[0.0136, 0.0453, 0.0465], [0.3748, 0.7979, 0.3027, 1.0]], 1: [[0.0216, 0.0467, 0.0453], [0.8871, 0.7031, 0.3587, 1.0]], 2: [[0.0115, 0.019, 0.0198], [0.4045, 0.3236, 0.5272, 1.0]], 3: [[0.0155, 0.0462, 0.0458], [0.4078, 0.8078, 0.8977, 1.0]], 4: [[0.015, 0.0475, 0.0464], [0.8567, 0.6917, 0.6805, 1.0]], 5: [[0.0202, 0.0153, 0.0151], [0.1327, 0.8283, 0.6056, 1.0]], 6: [[0.0113, 0.0167, 0.0184], [0.4724, 0.2611, 0.7643, 1.0]], 7: [[0.0105, 0.0155, 0.0183], [0.3037, 0.1116, 0.4612, 1.0]], 8: [[0.0183, 0.0165, 0.0194], [0.8587, 0.5076, 0.4942, 1.0]], 9: [[0.0106, 0.0187, 0.0193], [0.8848, 0.3242, 0.3264, 1.0]], 10: [[0.0209, 0.048, 0.0465], [0.5808, 0.6403, 0.6915, 1.0]], 11: [[0.0173, 0.018, 0.0159], [0.2226, 0.4985, 0.7465, 1.0]], 12: [[0.0211, 0.0183, 0.0167], [0.3708, 0.6494, 0.3519, 1.0]], 13: [[0.0106, 0.0461, 0.0473], [0.6228, 0.5324, 0.3074, 1.0]], 14: [[0.0138, 0.0151, 0.0181], [0.5768, 0.5588, 0.7342, 1.0]], 15: [[0.0162, 0.0484, 0.0494], [0.7756, 0.6015, 0.3484, 1.0]], 16: [[0.0134, 0.0487, 0.0493], [0.2497, 0.3642, 0.3717, 1.0]], 17: [[0.014, 0.0498, 0.0488], [0.2395, 0.845, 0.1984, 1.0]], 18: [[0.0176, 0.0157, 0.0197], [0.4929, 0.3595, 0.5476, 1.0]], 19: [[0.0118, 0.0487, 0.047], [0.2546, 0.4479, 0.4248, 1.0]], 20: [[0.0162, 0.0186, 0.0182], [0.129, 0.7677, 0.7907, 1.0]], 21: [[0.0155, 0.016, 0.0176], [0.4439, 0.5233, 0.4009, 1.0]], 22: [[0.0176, 0.0174, 0.0167], [0.7972, 0.5682, 0.7325, 1.0]], 23: [[0.0113, 0.0468, 0.0481], [0.4106, 0.3357, 0.6396, 1.0]], 24: [[0.018, 0.016, 0.0158], [0.7593, 0.1941, 0.4112, 1.0]], 25: [[0.0174, 0.0195, 0.0191], [0.4793, 0.5692, 0.5898, 1.0]], 26: [[0.0157, 0.048, 0.0456], [0.4152, 0.7345, 0.8086, 1.0]], 27: [[0.0144, 0.0496, 0.0461], [0.1571, 0.8837, 0.2644, 1.0]], 28: [[0.0167, 0.0454, 0.0455], [0.2708, 0.7488, 0.186, 1.0]], 29: [[0.0211, 0.0183, 0.0191], [0.6811, 0.8403, 0.3378, 1.0]], 30: [[0.0153, 0.0463, 0.047], [0.3878, 0.7574, 0.2985, 1.0]], 31: [[0.0165, 0.0168, 0.0191], [0.1893, 0.4611, 0.4676, 1.0]], 32: [[0.0219, 0.0493, 0.0471], [0.7561, 0.8639, 0.2819, 1.0]], 33: [[0.0213, 0.019, 0.0162], [0.7951, 0.4273, 0.757, 1.0]], 34: [[0.0147, 0.0474, 0.0456], [0.6613, 0.7031, 0.7934, 1.0]], 35: [[0.0195, 0.0188, 0.0179], [0.7112, 0.2417, 0.8843, 1.0]], 36: [[0.012, 0.0195, 0.0157], [0.7884, 0.7618, 0.4011, 1.0]], 37: [[0.0155, 0.0178, 0.017], [0.2318, 0.2223, 0.3326, 1.0]], 38: [[0.0126, 0.0458, 0.0473], [0.4536, 0.2954, 0.6068, 1.0]], 39: [[0.0181, 0.0168, 0.017], [0.6729, 0.7699, 0.4198, 1.0]], 40: [[0.0198, 0.0176, 0.0191], [0.3455, 0.8455, 0.3532, 1.0]], 41: [[0.0137, 0.0462, 0.0465], [0.1889, 0.2106, 0.2626, 1.0]], 42: [[0.0179, 0.0167, 0.0198], [0.6908, 0.1125, 0.7255, 1.0]], 43: [[0.0192, 0.0456, 0.0476], [0.568, 0.6211, 0.5615, 1.0]], 44: [[0.0198, 0.0479, 0.0473], [0.1722, 0.3925, 0.714, 1.0]], 45: [[0.0108, 0.0456, 0.0497], [0.8997, 0.5026, 0.4454, 1.0]], 46: [[0.0198, 0.0478, 0.0492], [0.19, 0.2314, 0.6352, 1.0]], 47: [[0.0106, 0.0498, 0.0479], [0.4222, 0.3215, 0.6336, 1.0]], 48: [[0.0123, 0.0167, 0.0196], [0.5462, 0.8333, 0.6013, 1.0]], 49: [[0.013, 0.018, 0.0195], [0.8953, 0.6054, 0.4694, 1.0]], 50: [[0.0175, 0.0451, 0.0468], [0.1662, 0.6229, 0.6202, 1.0]], 51: [[0.0108, 0.0155, 0.018], [0.2623, 0.3395, 0.6575, 1.0]], 52: [[0.0183, 0.0153, 0.0166], [0.4428, 0.1253, 0.8992, 1.0]], 53: [[0.0116, 0.0473, 0.0455], [0.3141, 0.5985, 0.8164, 1.0]], 54: [[0.0111, 0.0175, 0.0161], [0.8752, 0.3452, 0.1244, 1.0]], 55: [[0.0215, 0.0186, 0.0198], [0.6631, 0.3464, 0.5824, 1.0]], 56: [[0.0183, 0.0455, 0.0461], [0.3822, 0.6385, 0.5034, 1.0]], 57: [[0.0157, 0.0191, 0.0178], [0.1726, 0.6767, 0.5074, 1.0]], 58: [[0.0118, 0.0152, 0.0184], [0.2454, 0.5554, 0.1053, 1.0]], 59: [[0.0192, 0.0158, 0.0162], [0.5347, 0.8503, 0.6721, 1.0]], 60: [[0.0197, 0.0166, 0.0196], [0.7361, 0.7123, 0.7099, 1.0]], 61: [[0.0111, 0.018, 0.0176], [0.8479, 0.5385, 0.5338, 1.0]], 62: [[0.0113, 0.0484, 0.0493], [0.8344, 0.1699, 0.1653, 1.0]], 63: [[0.0189, 0.0462, 0.0499], [0.2908, 0.3325, 0.4644, 1.0]], 64: [[0.0176, 0.0456, 0.0494], [0.2908, 0.8575, 0.1616, 1.0]], 65: [[0.0198, 0.0493, 0.0481], [0.6792, 0.8347, 0.8449, 1.0]], 66: [[0.0105, 0.0193, 0.0174], [0.2344, 0.7014, 0.8164, 1.0]], 67: [[0.0122, 0.0486, 0.0475], [0.7926, 0.3015, 0.1006, 1.0]], 68: [[0.0174, 0.0186, 0.0182], [0.8619, 0.2421, 0.6301, 1.0]], 69: [[0.0123, 0.045, 0.0476], [0.6971, 0.2918, 0.1315, 1.0]], 70: [[0.0186, 0.0467, 0.0489], [0.8719, 0.2628, 0.5912, 1.0]], 71: [[0.0128, 0.0184, 0.017], [0.6068, 0.294, 0.6821, 1.0]], 72: [[0.0206, 0.0477, 0.047], [0.3023, 0.334, 0.3004, 1.0]], 73: [[0.0209, 0.0462, 0.0472], [0.3413, 0.408, 0.2517, 1.0]], 74: [[0.0126, 0.0451, 0.0458], [0.4991, 0.135, 0.1745, 1.0]], 75: [[0.01, 0.0456, 0.0463], [0.444, 0.6526, 0.8371, 1.0]], 76: [[0.0132, 0.0473, 0.0476], [0.6716, 0.5891, 0.413, 1.0]], 77: [[0.0197, 0.0177, 0.0199], [0.8096, 0.1534, 0.1838, 1.0]], 78: [[0.0207, 0.0154, 0.0193], [0.2925, 0.1442, 0.7184, 1.0]], 79: [[0.019, 0.0168, 0.016], [0.6438, 0.6782, 0.1347, 1.0]], 80: [[0.0197, 0.0495, 0.0487], [0.2494, 0.1925, 0.7452, 1.0]], 81: [[0.0162, 0.0187, 0.0171], [0.8802, 0.8363, 0.2678, 1.0]], 82: [[0.0158, 0.0189, 0.0164], [0.4709, 0.1114, 0.4535, 1.0]], 83: [[0.0212, 0.0455, 0.0453], [0.6133, 0.1836, 0.7338, 1.0]], 84: [[0.0184, 0.0193, 0.0155], [0.256, 0.5637, 0.4474, 1.0]], 85: [[0.0204, 0.0168, 0.0159], [0.8652, 0.7129, 0.3266, 1.0]], 86: [[0.0171, 0.0169, 0.0151], [0.6231, 0.2797, 0.1848, 1.0]], 87: [[0.0177, 0.0453, 0.0467], [0.4235, 0.346, 0.3701, 1.0]], 88: [[0.0166, 0.0475, 0.0454], [0.8074, 0.6102, 0.3561, 1.0]], 89: [[0.0176, 0.0191, 0.0187], [0.6713, 0.4377, 0.1849, 1.0]], 90: [[0.0144, 0.0458, 0.0489], [0.6503, 0.8789, 0.5924, 1.0]], 91: [[0.0147, 0.0473, 0.0466], [0.3362, 0.5114, 0.8767, 1.0]], 92: [[0.0216, 0.0484, 0.0485], [0.866, 0.8917, 0.6477, 1.0]], 93: [[0.0154, 0.017, 0.0169], [0.4121, 0.3731, 0.4281, 1.0]], 94: [[0.0211, 0.049, 0.0456], [0.116, 0.2635, 0.8146, 1.0]], 95: [[0.019, 0.0193, 0.0195], [0.5679, 0.1996, 0.4843, 1.0]], 96: [[0.0208, 0.0455, 0.047], [0.6802, 0.7661, 0.7532, 1.0]], 97: [[0.0181, 0.0465, 0.0451], [0.7782, 0.7574, 0.3937, 1.0]], 98: [[0.0155, 0.0459, 0.0471], [0.1022, 0.4936, 0.3575, 1.0]], 99: [[0.0199, 0.0499, 0.0484], [0.1005, 0.4705, 0.6336, 1.0]], 100: [[0.0185, 0.0473, 0.0459], [0.2766, 0.6181, 0.6827, 1.0]], 101: [[0.0156, 0.0489, 0.0485], [0.7419, 0.5151, 0.1375, 1.0]], 102: [[0.0169, 0.0467, 0.0495], [0.1598, 0.3039, 0.3623, 1.0]], 103: [[0.0132, 0.0492, 0.0451], [0.5516, 0.6943, 0.7989, 1.0]], 104: [[0.0218, 0.0465, 0.046], [0.8309, 0.6258, 0.5871, 1.0]], 105: [[0.0191, 0.0197, 0.0151], [0.3479, 0.6324, 0.6539, 1.0]], 106: [[0.0134, 0.0474, 0.0487], [0.7873, 0.4069, 0.6138, 1.0]], 107: [[0.0108, 0.018, 0.0188], [0.2213, 0.6136, 0.6595, 1.0]], 108: [[0.0183, 0.0463, 0.0498], [0.2309, 0.6564, 0.1953, 1.0]], 109: [[0.0217, 0.02, 0.0164], [0.3696, 0.3274, 0.8075, 1.0]], 110: [[0.0208, 0.0188, 0.0195], [0.5793, 0.7412, 0.5842, 1.0]], 111: [[0.0123, 0.0472, 0.0462], [0.2745, 0.6986, 0.6367, 1.0]], 112: [[0.0109, 0.0169, 0.0187], [0.846, 0.2854, 0.817, 1.0]], 113: [[0.0197, 0.0462, 0.0451], [0.6611, 0.8956, 0.6017, 1.0]], 114: [[0.0126, 0.0482, 0.0466], [0.1515, 0.75, 0.4982, 1.0]], 115: [[0.0144, 0.0461, 0.0483], [0.2334, 0.439, 0.4271, 1.0]], 116: [[0.0113, 0.017, 0.015], [0.6978, 0.2366, 0.6323, 1.0]], 117: [[0.0121, 0.0497, 0.0453], [0.6231, 0.5773, 0.8073, 1.0]], 118: [[0.0134, 0.0482, 0.0477], [0.8612, 0.6541, 0.5015, 1.0]], 119: [[0.0172, 0.0477, 0.0473], [0.6113, 0.3086, 0.5542, 1.0]], 120: [[0.0203, 0.016, 0.0162], [0.5004, 0.4939, 0.6544, 1.0]], 121: [[0.0187, 0.0481, 0.05], [0.2052, 0.2524, 0.5169, 1.0]], 122: [[0.0127, 0.0192, 0.0167], [0.8749, 0.8374, 0.6564, 1.0]], 123: [[0.0166, 0.019, 0.0178], [0.2808, 0.8304, 0.7394, 1.0]], 124: [[0.0217, 0.0452, 0.0497], [0.1606, 0.1125, 0.7687, 1.0]], 125: [[0.0206, 0.0151, 0.0198], [0.1354, 0.3074, 0.8946, 1.0]], 126: [[0.0155, 0.0484, 0.0478], [0.1494, 0.3161, 0.5858, 1.0]], 127: [[0.0189, 0.049, 0.0476], [0.7759, 0.8747, 0.3178, 1.0]], 128: [[0.0132, 0.0485, 0.05], [0.1897, 0.7151, 0.3964, 1.0]], 129: [[0.0206, 0.0158, 0.0162], [0.5965, 0.6827, 0.6767, 1.0]], 130: [[0.0111, 0.017, 0.0163], [0.89, 0.8836, 0.3453, 1.0]], 131: [[0.0126, 0.0181, 0.0172], [0.4132, 0.6712, 0.4386, 1.0]], 132: [[0.0109, 0.0153, 0.0189], [0.7646, 0.602, 0.6782, 1.0]], 133: [[0.0115, 0.0493, 0.0468], [0.6764, 0.1377, 0.465, 1.0]], 134: [[0.0205, 0.0178, 0.0178], [0.8301, 0.4982, 0.6208, 1.0]], 135: [[0.0178, 0.0477, 0.0467], [0.7852, 0.5497, 0.4871, 1.0]], 136: [[0.0137, 0.0465, 0.0468], [0.1821, 0.853, 0.2829, 1.0]], 137: [[0.0155, 0.0191, 0.0199], [0.6496, 0.1858, 0.8941, 1.0]], 138: [[0.0162, 0.0158, 0.0173], [0.142, 0.2526, 0.5855, 1.0]], 139: [[0.0165, 0.0495, 0.0499], [0.2778, 0.4025, 0.3015, 1.0]], 140: [[0.0116, 0.0168, 0.0152], [0.7704, 0.7059, 0.3592, 1.0]], 141: [[0.0209, 0.0176, 0.0184], [0.2291, 0.5791, 0.6604, 1.0]], 142: [[0.0218, 0.0157, 0.0161], [0.4174, 0.852, 0.6206, 1.0]], 143: [[0.0219, 0.0474, 0.0486], [0.2593, 0.6737, 0.214, 1.0]], 144: [[0.0157, 0.0464, 0.0479], [0.8564, 0.8092, 0.8674, 1.0]], 145: [[0.0165, 0.0154, 0.0163], [0.3021, 0.1334, 0.2628, 1.0]], 146: [[0.0197, 0.0152, 0.0198], [0.6966, 0.2052, 0.2024, 1.0]], 147: [[0.0125, 0.017, 0.0176], [0.4436, 0.3441, 0.3479, 1.0]], 148: [[0.0203, 0.0481, 0.0461], [0.2574, 0.6637, 0.1716, 1.0]], 149: [[0.0102, 0.0198, 0.0198], [0.7608, 0.3312, 0.4526, 1.0]], 150: [[0.0157, 0.0186, 0.0161], [0.5963, 0.8301, 0.2927, 1.0]], 151: [[0.0133, 0.0153, 0.0182], [0.8432, 0.4611, 0.2456, 1.0]], 152: [[0.0128, 0.0166, 0.0157], [0.1976, 0.3311, 0.64, 1.0]], 153: [[0.0213, 0.0472, 0.0478], [0.1181, 0.6357, 0.8259, 1.0]], 154: [[0.0126, 0.0482, 0.0471], [0.1803, 0.1877, 0.5213, 1.0]], 155: [[0.0136, 0.016, 0.0174], [0.3999, 0.1447, 0.6851, 1.0]], 156: [[0.011, 0.0452, 0.0481], [0.6586, 0.293, 0.7292, 1.0]], 157: [[0.0123, 0.0468, 0.0451], [0.3359, 0.5545, 0.5047, 1.0]], 158: [[0.0216, 0.0154, 0.0171], [0.658, 0.2362, 0.511, 1.0]], 159: [[0.0184, 0.0174, 0.0166], [0.7972, 0.7666, 0.323, 1.0]], 160: [[0.0136, 0.0196, 0.0191], [0.5821, 0.5891, 0.1253, 1.0]], 161: [[0.0206, 0.0472, 0.0499], [0.6396, 0.4526, 0.8318, 1.0]], 162: [[0.0136, 0.0457, 0.045], [0.701, 0.8266, 0.3977, 1.0]], 163: [[0.0132, 0.0495, 0.0473], [0.6122, 0.4861, 0.2338, 1.0]], 164: [[0.0152, 0.0458, 0.0456], [0.3905, 0.3974, 0.8327, 1.0]], 165: [[0.0191, 0.0172, 0.0179], [0.2514, 0.3555, 0.7059, 1.0]], 166: [[0.012, 0.0172, 0.0157], [0.3152, 0.3787, 0.1999, 1.0]], 167: [[0.0211, 0.0155, 0.0173], [0.2141, 0.473, 0.5786, 1.0]], 168: [[0.0132, 0.0479, 0.0453], [0.8847, 0.1754, 0.5034, 1.0]], 169: [[0.0204, 0.049, 0.0486], [0.3426, 0.2197, 0.6171, 1.0]], 170: [[0.0155, 0.0495, 0.0458], [0.5176, 0.6022, 0.3693, 1.0]], 171: [[0.0131, 0.0188, 0.0198], [0.7074, 0.2283, 0.1959, 1.0]], 172: [[0.0191, 0.0195, 0.0161], [0.3126, 0.2926, 0.316, 1.0]], 173: [[0.0197, 0.0472, 0.0467], [0.7152, 0.6113, 0.6974, 1.0]], 174: [[0.0159, 0.0487, 0.0454], [0.2813, 0.1605, 0.4436, 1.0]], 175: [[0.0191, 0.0194, 0.017], [0.42, 0.5316, 0.8872, 1.0]], 176: [[0.0122, 0.047, 0.0485], [0.4035, 0.3233, 0.7167, 1.0]], 177: [[0.0201, 0.0491, 0.046], [0.7328, 0.6765, 0.4075, 1.0]], 178: [[0.0142, 0.0174, 0.0192], [0.7001, 0.1587, 0.5301, 1.0]], 179: [[0.016, 0.0192, 0.0172], [0.7941, 0.2497, 0.4538, 1.0]], 180: [[0.011, 0.046, 0.0474], [0.4089, 0.5942, 0.8117, 1.0]], 181: [[0.0158, 0.0493, 0.0488], [0.6709, 0.6229, 0.1599, 1.0]], 182: [[0.0188, 0.0165, 0.0188], [0.8725, 0.731, 0.2593, 1.0]], 183: [[0.0216, 0.0198, 0.0172], [0.4314, 0.4511, 0.5129, 1.0]], 184: [[0.0103, 0.0193, 0.0197], [0.4794, 0.1068, 0.8258, 1.0]], 185: [[0.0212, 0.0158, 0.0175], [0.6356, 0.5141, 0.518, 1.0]], 186: [[0.0204, 0.0179, 0.0156], [0.6822, 0.6685, 0.1364, 1.0]], 187: [[0.0214, 0.0178, 0.0161], [0.5418, 0.3319, 0.4074, 1.0]], 188: [[0.0168, 0.0163, 0.0152], [0.6582, 0.5392, 0.6434, 1.0]], 189: [[0.0196, 0.0451, 0.048], [0.5922, 0.7509, 0.8286, 1.0]], 190: [[0.0191, 0.0175, 0.0163], [0.5524, 0.808, 0.8212, 1.0]], 191: [[0.0146, 0.0479, 0.0476], [0.2852, 0.3818, 0.4151, 1.0]], 192: [[0.0145, 0.0489, 0.0473], [0.3713, 0.3106, 0.7304, 1.0]], 193: [[0.0193, 0.0482, 0.0462], [0.4943, 0.1897, 0.6195, 1.0]], 194: [[0.0193, 0.0158, 0.0173], [0.3981, 0.1664, 0.8238, 1.0]], 195: [[0.0129, 0.0498, 0.0466], [0.151, 0.4217, 0.1918, 1.0]], 196: [[0.0134, 0.0487, 0.047], [0.7262, 0.293, 0.1777, 1.0]], 197: [[0.0183, 0.0195, 0.0178], [0.5601, 0.5973, 0.8701, 1.0]], 198: [[0.0113, 0.0494, 0.0497], [0.2797, 0.2278, 0.6123, 1.0]], 199: [[0.0119, 0.047, 0.0493], [0.778, 0.5574, 0.8248, 1.0]], 200: [[0.0157, 0.0164, 0.0178], [0.4218, 0.8019, 0.4802, 1.0]], 201: [[0.0161, 0.0171, 0.0153], [0.7662, 0.6513, 0.834, 1.0]], 202: [[0.0132, 0.0165, 0.0175], [0.5475, 0.4114, 0.572, 1.0]], 203: [[0.0182, 0.0151, 0.0186], [0.889, 0.4841, 0.3023, 1.0]], 204: [[0.0133, 0.047, 0.048], [0.516, 0.468, 0.556, 1.0]], 205: [[0.0159, 0.0493, 0.0488], [0.3457, 0.3227, 0.8008, 1.0]], 206: [[0.0163, 0.048, 0.05], [0.1953, 0.5315, 0.2236, 1.0]], 207: [[0.0119, 0.0479, 0.0474], [0.3902, 0.7467, 0.5039, 1.0]], 208: [[0.0137, 0.05, 0.0492], [0.3128, 0.3642, 0.8588, 1.0]], 209: [[0.0182, 0.0474, 0.0497], [0.2686, 0.1469, 0.6976, 1.0]], 210: [[0.0124, 0.0452, 0.0468], [0.3935, 0.6924, 0.8082, 1.0]], 211: [[0.0111, 0.0489, 0.0469], [0.231, 0.2743, 0.4259, 1.0]], 212: [[0.0131, 0.0155, 0.0161], [0.4258, 0.6261, 0.672, 1.0]], 213: [[0.0147, 0.0455, 0.0457], [0.2577, 0.8078, 0.7995, 1.0]], 214: [[0.0173, 0.0152, 0.0152], [0.3563, 0.8776, 0.5738, 1.0]], 215: [[0.0102, 0.0494, 0.0461], [0.2283, 0.2928, 0.2659, 1.0]], 216: [[0.0111, 0.045, 0.0484], [0.8343, 0.2985, 0.3979, 1.0]], 217: [[0.0138, 0.019, 0.0163], [0.5333, 0.3226, 0.7198, 1.0]], 218: [[0.0211, 0.0169, 0.0184], [0.5943, 0.1519, 0.1221, 1.0]], 219: [[0.0196, 0.0183, 0.0181], [0.152, 0.868, 0.7424, 1.0]], 220: [[0.015, 0.0182, 0.0198], [0.5884, 0.8706, 0.7795, 1.0]], 221: [[0.0204, 0.0191, 0.0162], [0.4559, 0.7612, 0.6992, 1.0]], 222: [[0.0161, 0.0477, 0.0486], [0.8128, 0.2952, 0.4113, 1.0]], 223: [[0.0192, 0.0177, 0.0155], [0.5765, 0.8737, 0.6464, 1.0]], 224: [[0.0197, 0.0182, 0.0198], [0.4668, 0.2507, 0.5802, 1.0]], 225: [[0.0169, 0.0193, 0.019], [0.8713, 0.2197, 0.3935, 1.0]], 226: [[0.0133, 0.0161, 0.0187], [0.3672, 0.8112, 0.7703, 1.0]], 227: [[0.0102, 0.0181, 0.0166], [0.1516, 0.6819, 0.6346, 1.0]], 228: [[0.0138, 0.0166, 0.0158], [0.4226, 0.3507, 0.8957, 1.0]], 229: [[0.0183, 0.0472, 0.0496], [0.6886, 0.7564, 0.2884, 1.0]], 230: [[0.0101, 0.0162, 0.0183], [0.5962, 0.1275, 0.298, 1.0]], 231: [[0.0145, 0.0175, 0.0161], [0.7147, 0.4422, 0.4434, 1.0]], 232: [[0.0155, 0.0182, 0.0183], [0.7283, 0.7949, 0.155, 1.0]], 233: [[0.0144, 0.0199, 0.0191], [0.592, 0.8102, 0.6046, 1.0]], 234: [[0.0151, 0.0183, 0.0167], [0.841, 0.4077, 0.7135, 1.0]], 235: [[0.0162, 0.0156, 0.016], [0.1617, 0.6185, 0.8451, 1.0]], 236: [[0.0154, 0.0176, 0.0199], [0.2528, 0.7324, 0.4479, 1.0]], 237: [[0.014, 0.0173, 0.0153], [0.2075, 0.7459, 0.2714, 1.0]], 238: [[0.0113, 0.0486, 0.0478], [0.3511, 0.4506, 0.4231, 1.0]], 239: [[0.0102, 0.0177, 0.0197], [0.4801, 0.5203, 0.1062, 1.0]], 240: [[0.0154, 0.0181, 0.0198], [0.7822, 0.3277, 0.727, 1.0]], 241: [[0.0157, 0.0464, 0.0481], [0.5016, 0.4648, 0.8669, 1.0]], 242: [[0.0163, 0.0154, 0.0162], [0.2544, 0.4654, 0.1192, 1.0]], 243: [[0.0152, 0.0175, 0.0151], [0.7921, 0.1791, 0.2835, 1.0]], 244: [[0.0177, 0.0471, 0.0494], [0.7089, 0.7851, 0.7274, 1.0]], 245: [[0.0217, 0.0159, 0.0176], [0.4895, 0.5296, 0.3363, 1.0]], 246: [[0.0137, 0.0178, 0.0159], [0.2279, 0.179, 0.1894, 1.0]], 247: [[0.0133, 0.0173, 0.0191], [0.8349, 0.7864, 0.2583, 1.0]], 248: [[0.0211, 0.0498, 0.0465], [0.5685, 0.5647, 0.6536, 1.0]], 249: [[0.0135, 0.0162, 0.0175], [0.6919, 0.8295, 0.6216, 1.0]]}

        box = {0: [[0.0128, 0.0137, 0.0125], [0.1262, 0.4485, 0.5911, 1.0]], 1: [[0.0101, 0.013, 0.0135], [0.8879, 0.8189, 0.6094, 1.0]], 2: [[0.0263, 0.0265, 0.0262], [0.1836, 0.3486, 0.8583, 1.0]], 3: [[0.0267, 0.0277, 0.0259], [0.5334, 0.7891, 0.6696, 1.0]], 4: [[0.0126, 0.0127, 0.014], [0.4511, 0.291, 0.5827, 1.0]], 5: [[0.0256, 0.0259, 0.0299], [0.5044, 0.6057, 0.7387, 1.0]], 6: [[0.0131, 0.0106, 0.0127], [0.3935, 0.3898, 0.5982, 1.0]], 7: [[0.0286, 0.0289, 0.0272], [0.4006, 0.6326, 0.218, 1.0]], 8: [[0.0112, 0.0141, 0.0142], [0.1367, 0.4518, 0.1999, 1.0]], 9: [[0.0285, 0.026, 0.0279], [0.662, 0.3339, 0.4512, 1.0]], 10: [[0.0124, 0.0137, 0.0117], [0.3106, 0.7111, 0.6745, 1.0]], 11: [[0.0115, 0.0115, 0.0147], [0.334, 0.5556, 0.3631, 1.0]], 12: [[0.03, 0.0271, 0.0261], [0.1791, 0.5839, 0.7973, 1.0]], 13: [[0.0288, 0.0289, 0.0295], [0.1496, 0.8241, 0.7307, 1.0]], 14: [[0.0138, 0.0127, 0.0105], [0.4867, 0.3897, 0.2802, 1.0]], 15: [[0.0113, 0.0137, 0.0142], [0.1835, 0.6865, 0.4083, 1.0]], 16: [[0.0136, 0.0136, 0.0118], [0.3375, 0.4376, 0.5449, 1.0]], 17: [[0.0133, 0.0129, 0.013], [0.2291, 0.2999, 0.6796, 1.0]], 18: [[0.0255, 0.027, 0.0268], [0.3764, 0.8595, 0.112, 1.0]], 19: [[0.0129, 0.0132, 0.0107], [0.6853, 0.7254, 0.4691, 1.0]], 20: [[0.0139, 0.0134, 0.0101], [0.8053, 0.5342, 0.7035, 1.0]], 21: [[0.0278, 0.0282, 0.0289], [0.8352, 0.2037, 0.4103, 1.0]], 22: [[0.0274, 0.029, 0.0251], [0.3175, 0.4848, 0.7251, 1.0]], 23: [[0.0285, 0.027, 0.0277], [0.8741, 0.2172, 0.8168, 1.0]], 24: [[0.0278, 0.0278, 0.0285], [0.3457, 0.8998, 0.5409, 1.0]], 25: [[0.0263, 0.0287, 0.0261], [0.3879, 0.6013, 0.888, 1.0]], 26: [[0.0253, 0.0297, 0.0256], [0.6693, 0.2169, 0.7129, 1.0]], 27: [[0.0148, 0.0138, 0.0126], [0.2763, 0.2673, 0.4364, 1.0]], 28: [[0.0142, 0.0125, 0.0119], [0.725, 0.846, 0.6398, 1.0]], 29: [[0.0279, 0.029, 0.0296], [0.6528, 0.2828, 0.1484, 1.0]], 30: [[0.0149, 0.0101, 0.01], [0.6548, 0.4679, 0.7466, 1.0]], 31: [[0.013, 0.0122, 0.0127], [0.8046, 0.8899, 0.4473, 1.0]], 32: [[0.0298, 0.0296, 0.0269], [0.6014, 0.6499, 0.4345, 1.0]], 33: [[0.0118, 0.0138, 0.0144], [0.877, 0.1957, 0.4091, 1.0]], 34: [[0.0128, 0.0149, 0.0101], [0.6243, 0.274, 0.4457, 1.0]], 35: [[0.0134, 0.0102, 0.0109], [0.7476, 0.1233, 0.3714, 1.0]], 36: [[0.0124, 0.0133, 0.0117], [0.4075, 0.6813, 0.6772, 1.0]], 37: [[0.0105, 0.0146, 0.0104], [0.4365, 0.6879, 0.5085, 1.0]], 38: [[0.0258, 0.0283, 0.0283], [0.6691, 0.4444, 0.8198, 1.0]], 39: [[0.0142, 0.0146, 0.0129], [0.6146, 0.5034, 0.8204, 1.0]], 40: [[0.0261, 0.0264, 0.0253], [0.5519, 0.4915, 0.7486, 1.0]], 41: [[0.0107, 0.0137, 0.0129], [0.4902, 0.3334, 0.8042, 1.0]], 42: [[0.01, 0.0142, 0.0121], [0.462, 0.5694, 0.4513, 1.0]], 43: [[0.0128, 0.0104, 0.0119], [0.1222, 0.493, 0.5988, 1.0]], 44: [[0.0135, 0.0115, 0.0113], [0.5903, 0.8255, 0.5013, 1.0]], 45: [[0.0262, 0.0257, 0.0293], [0.5172, 0.5639, 0.2873, 1.0]], 46: [[0.0297, 0.0287, 0.0266], [0.1856, 0.4403, 0.585, 1.0]], 47: [[0.026, 0.0252, 0.0287], [0.8557, 0.2358, 0.1663, 1.0]], 48: [[0.0297, 0.0299, 0.0276], [0.7302, 0.6216, 0.8746, 1.0]], 49: [[0.0285, 0.0291, 0.0282], [0.8668, 0.4253, 0.881, 1.0]], 50: [[0.0149, 0.0144, 0.0123], [0.758, 0.7659, 0.6712, 1.0]], 51: [[0.03, 0.0295, 0.0253], [0.5713, 0.1853, 0.7528, 1.0]], 52: [[0.0109, 0.0135, 0.0144], [0.295, 0.7474, 0.4573, 1.0]], 53: [[0.0129, 0.0119, 0.0131], [0.7363, 0.8958, 0.3108, 1.0]], 54: [[0.0133, 0.01, 0.0105], [0.8603, 0.6225, 0.5709, 1.0]], 55: [[0.0257, 0.0269, 0.0267], [0.8191, 0.6726, 0.4744, 1.0]], 56: [[0.0286, 0.027, 0.0253], [0.6977, 0.7065, 0.5718, 1.0]], 57: [[0.0144, 0.0125, 0.0111], [0.6535, 0.7838, 0.6343, 1.0]], 58: [[0.0137, 0.0115, 0.0124], [0.1923, 0.8019, 0.1552, 1.0]], 59: [[0.0258, 0.0277, 0.0284], [0.5151, 0.418, 0.5077, 1.0]], 60: [[0.0293, 0.0298, 0.0253], [0.6818, 0.285, 0.7324, 1.0]], 61: [[0.0108, 0.0137, 0.0135], [0.4882, 0.7181, 0.517, 1.0]], 62: [[0.0295, 0.0277, 0.0253], [0.6206, 0.5617, 0.8806, 1.0]], 63: [[0.0103, 0.0122, 0.012], [0.5088, 0.27, 0.8563, 1.0]], 64: [[0.0258, 0.0288, 0.026], [0.302, 0.4256, 0.7416, 1.0]], 65: [[0.0126, 0.0104, 0.0112], [0.3447, 0.8134, 0.8105, 1.0]], 66: [[0.0262, 0.027, 0.0265], [0.7574, 0.7556, 0.5048, 1.0]], 67: [[0.0138, 0.0138, 0.0129], [0.2114, 0.422, 0.4106, 1.0]], 68: [[0.0141, 0.0135, 0.014], [0.7794, 0.2869, 0.8142, 1.0]], 69: [[0.0277, 0.0254, 0.0289], [0.8961, 0.873, 0.8332, 1.0]], 70: [[0.0261, 0.0286, 0.0274], [0.3278, 0.6657, 0.7636, 1.0]], 71: [[0.0131, 0.0117, 0.0127], [0.2688, 0.8168, 0.7829, 1.0]], 72: [[0.0139, 0.0134, 0.0139], [0.2764, 0.4531, 0.5006, 1.0]], 73: [[0.011, 0.0128, 0.0144], [0.1394, 0.6613, 0.2189, 1.0]], 74: [[0.0136, 0.0132, 0.0114], [0.1037, 0.4059, 0.2122, 1.0]], 75: [[0.0263, 0.0297, 0.0266], [0.3146, 0.3605, 0.1213, 1.0]], 76: [[0.0256, 0.0282, 0.0255], [0.6996, 0.1943, 0.589, 1.0]], 77: [[0.0144, 0.0139, 0.011], [0.3058, 0.8853, 0.5149, 1.0]], 78: [[0.0293, 0.029, 0.0271], [0.7114, 0.7254, 0.4608, 1.0]], 79: [[0.0252, 0.0264, 0.0268], [0.2592, 0.2464, 0.2794, 1.0]], 80: [[0.0261, 0.0263, 0.0299], [0.5121, 0.8025, 0.1194, 1.0]], 81: [[0.0284, 0.0278, 0.0256], [0.7902, 0.7042, 0.6039, 1.0]], 82: [[0.0126, 0.0142, 0.0111], [0.7471, 0.1177, 0.8553, 1.0]], 83: [[0.029, 0.0274, 0.0271], [0.3933, 0.6876, 0.4097, 1.0]], 84: [[0.0268, 0.029, 0.0257], [0.3723, 0.7419, 0.6257, 1.0]], 85: [[0.0107, 0.0128, 0.0138], [0.3318, 0.8069, 0.3866, 1.0]], 86: [[0.0257, 0.0285, 0.0282], [0.5244, 0.1721, 0.4504, 1.0]], 87: [[0.013, 0.0101, 0.0141], [0.5619, 0.6535, 0.1596, 1.0]], 88: [[0.0113, 0.0113, 0.0128], [0.2037, 0.4442, 0.7386, 1.0]], 89: [[0.013, 0.0104, 0.0145], [0.4971, 0.2382, 0.8087, 1.0]], 90: [[0.0299, 0.0276, 0.0255], [0.5392, 0.1515, 0.7472, 1.0]], 91: [[0.0272, 0.0279, 0.0274], [0.6191, 0.5203, 0.3461, 1.0]], 92: [[0.0135, 0.0127, 0.0109], [0.7801, 0.1271, 0.2521, 1.0]], 93: [[0.0102, 0.0132, 0.0124], [0.202, 0.411, 0.4243, 1.0]], 94: [[0.0267, 0.0261, 0.0253], [0.1199, 0.2647, 0.1439, 1.0]], 95: [[0.0291, 0.0288, 0.0281], [0.8512, 0.1254, 0.4538, 1.0]], 96: [[0.0291, 0.026, 0.0286], [0.1908, 0.2686, 0.7173, 1.0]], 97: [[0.0136, 0.0146, 0.0134], [0.7723, 0.3147, 0.3063, 1.0]], 98: [[0.0115, 0.01, 0.0134], [0.8129, 0.6308, 0.5641, 1.0]], 99: [[0.0267, 0.0267, 0.0277], [0.1949, 0.5013, 0.214, 1.0]], 100: [[0.0125, 0.0117, 0.0125], [0.5669, 0.1599, 0.528, 1.0]], 101: [[0.0266, 0.0269, 0.0277], [0.5513, 0.6766, 0.3878, 1.0]], 102: [[0.0102, 0.0104, 0.0142], [0.1553, 0.1524, 0.2355, 1.0]], 103: [[0.0256, 0.0272, 0.028], [0.898, 0.8394, 0.4335, 1.0]], 104: [[0.0112, 0.0145, 0.012], [0.1932, 0.3893, 0.6954, 1.0]], 105: [[0.0133, 0.0147, 0.0119], [0.397, 0.3512, 0.3219, 1.0]], 106: [[0.0142, 0.0109, 0.0115], [0.7574, 0.2018, 0.8852, 1.0]], 107: [[0.011, 0.0106, 0.0135], [0.2242, 0.6484, 0.8707, 1.0]], 108: [[0.0289, 0.0255, 0.0272], [0.8595, 0.8306, 0.5356, 1.0]], 109: [[0.0149, 0.0149, 0.0126], [0.6568, 0.6734, 0.4185, 1.0]], 110: [[0.011, 0.0147, 0.0134], [0.7531, 0.4217, 0.1794, 1.0]], 111: [[0.0105, 0.0117, 0.0139], [0.3429, 0.1957, 0.6306, 1.0]], 112: [[0.0257, 0.0278, 0.0253], [0.8887, 0.3157, 0.2606, 1.0]], 113: [[0.0284, 0.0266, 0.0294], [0.4688, 0.3379, 0.444, 1.0]], 114: [[0.026, 0.0284, 0.0278], [0.5506, 0.8421, 0.5794, 1.0]], 115: [[0.0141, 0.0129, 0.0128], [0.89, 0.3684, 0.6358, 1.0]], 116: [[0.0147, 0.0139, 0.0113], [0.1151, 0.2355, 0.3389, 1.0]], 117: [[0.0256, 0.0258, 0.0279], [0.7308, 0.7349, 0.312, 1.0]], 118: [[0.0283, 0.0268, 0.0275], [0.3733, 0.4928, 0.2027, 1.0]], 119: [[0.0143, 0.0135, 0.0137], [0.8827, 0.535, 0.4144, 1.0]], 120: [[0.0129, 0.0137, 0.0115], [0.8204, 0.8049, 0.6163, 1.0]], 121: [[0.0125, 0.0114, 0.0107], [0.8649, 0.6468, 0.2472, 1.0]], 122: [[0.0114, 0.0138, 0.0101], [0.6901, 0.151, 0.6844, 1.0]], 123: [[0.012, 0.0111, 0.0104], [0.3474, 0.8987, 0.3287, 1.0]], 124: [[0.0122, 0.0122, 0.0136], [0.596, 0.8042, 0.4707, 1.0]], 125: [[0.0272, 0.0276, 0.0267], [0.8041, 0.5445, 0.6634, 1.0]], 126: [[0.0278, 0.0298, 0.0295], [0.2618, 0.537, 0.37, 1.0]], 127: [[0.0107, 0.0113, 0.0135], [0.2634, 0.1988, 0.5307, 1.0]], 128: [[0.0133, 0.0104, 0.0141], [0.8195, 0.5427, 0.7763, 1.0]], 129: [[0.0145, 0.0125, 0.012], [0.255, 0.7705, 0.8919, 1.0]], 130: [[0.0277, 0.0265, 0.0268], [0.6814, 0.4493, 0.1115, 1.0]], 131: [[0.0141, 0.0126, 0.0109], [0.4624, 0.5666, 0.2391, 1.0]], 132: [[0.026, 0.0299, 0.0285], [0.4675, 0.3553, 0.8849, 1.0]], 133: [[0.0144, 0.0116, 0.0146], [0.8404, 0.404, 0.5648, 1.0]], 134: [[0.0265, 0.0296, 0.0257], [0.1486, 0.459, 0.1454, 1.0]], 135: [[0.0143, 0.013, 0.014], [0.3542, 0.8449, 0.1504, 1.0]], 136: [[0.0139, 0.0146, 0.01], [0.4246, 0.3319, 0.8988, 1.0]], 137: [[0.0105, 0.0122, 0.014], [0.4841, 0.1667, 0.5605, 1.0]], 138: [[0.0128, 0.0109, 0.0145], [0.7375, 0.6717, 0.3158, 1.0]], 139: [[0.0282, 0.0292, 0.0291], [0.7052, 0.4345, 0.6344, 1.0]], 140: [[0.0296, 0.0282, 0.0271], [0.3228, 0.5714, 0.3843, 1.0]], 141: [[0.0287, 0.0291, 0.0272], [0.6182, 0.1889, 0.8673, 1.0]], 142: [[0.0127, 0.0145, 0.0136], [0.6075, 0.5134, 0.5949, 1.0]], 143: [[0.0118, 0.0102, 0.0148], [0.2836, 0.4126, 0.1526, 1.0]], 144: [[0.0146, 0.0118, 0.0117], [0.4046, 0.6429, 0.2554, 1.0]], 145: [[0.0274, 0.0261, 0.0252], [0.7512, 0.3923, 0.5647, 1.0]], 146: [[0.012, 0.0139, 0.0124], [0.8737, 0.6232, 0.8241, 1.0]], 147: [[0.0267, 0.0257, 0.0293], [0.4793, 0.7202, 0.5843, 1.0]], 148: [[0.0272, 0.0265, 0.0256], [0.3242, 0.2449, 0.1867, 1.0]], 149: [[0.0125, 0.0109, 0.0101], [0.5279, 0.3426, 0.6351, 1.0]], 150: [[0.0145, 0.0109, 0.012], [0.1932, 0.4394, 0.7827, 1.0]], 151: [[0.0284, 0.0287, 0.0251], [0.8726, 0.3294, 0.8443, 1.0]], 152: [[0.0257, 0.0283, 0.0251], [0.6997, 0.8156, 0.8889, 1.0]], 153: [[0.0137, 0.0124, 0.0147], [0.7246, 0.3513, 0.6352, 1.0]], 154: [[0.0255, 0.0294, 0.0253], [0.135, 0.7219, 0.6353, 1.0]], 155: [[0.0254, 0.0254, 0.0291], [0.2375, 0.3027, 0.7849, 1.0]], 156: [[0.0255, 0.0289, 0.0284], [0.1882, 0.107, 0.4603, 1.0]], 157: [[0.0265, 0.0253, 0.0298], [0.6542, 0.6218, 0.5745, 1.0]], 158: [[0.0295, 0.0299, 0.025], [0.4886, 0.3477, 0.1595, 1.0]], 159: [[0.0289, 0.0254, 0.027], [0.7394, 0.7997, 0.6807, 1.0]], 160: [[0.0258, 0.0272, 0.0272], [0.6905, 0.1545, 0.2134, 1.0]], 161: [[0.0124, 0.0121, 0.0104], [0.4147, 0.8507, 0.6801, 1.0]], 162: [[0.0145, 0.0133, 0.0102], [0.5378, 0.2756, 0.8797, 1.0]], 163: [[0.0107, 0.0136, 0.0128], [0.1329, 0.7218, 0.7998, 1.0]], 164: [[0.0295, 0.025, 0.0272], [0.4372, 0.3952, 0.3467, 1.0]], 165: [[0.0285, 0.0288, 0.026], [0.346, 0.3606, 0.723, 1.0]], 166: [[0.027, 0.0252, 0.0277], [0.1379, 0.4492, 0.2061, 1.0]], 167: [[0.0286, 0.0283, 0.0286], [0.2993, 0.2599, 0.8555, 1.0]], 168: [[0.0147, 0.0113, 0.0107], [0.4044, 0.1088, 0.4545, 1.0]], 169: [[0.0256, 0.0258, 0.0296], [0.2074, 0.8322, 0.789, 1.0]], 170: [[0.026, 0.0254, 0.0297], [0.4855, 0.3445, 0.593, 1.0]], 171: [[0.0138, 0.0127, 0.0127], [0.5552, 0.2835, 0.3668, 1.0]], 172: [[0.0293, 0.0263, 0.026], [0.2796, 0.3643, 0.2477, 1.0]], 173: [[0.0109, 0.011, 0.0118], [0.4079, 0.7715, 0.373, 1.0]], 174: [[0.0294, 0.0274, 0.0253], [0.4954, 0.776, 0.4679, 1.0]], 175: [[0.0295, 0.0266, 0.0266], [0.8103, 0.5373, 0.632, 1.0]], 176: [[0.0113, 0.011, 0.0135], [0.5424, 0.1924, 0.3599, 1.0]], 177: [[0.0276, 0.0299, 0.0288], [0.5711, 0.2667, 0.6786, 1.0]], 178: [[0.0272, 0.0258, 0.0278], [0.4606, 0.2773, 0.117, 1.0]], 179: [[0.014, 0.0136, 0.0106], [0.5328, 0.6054, 0.5389, 1.0]], 180: [[0.012, 0.0122, 0.0111], [0.4801, 0.3989, 0.81, 1.0]], 181: [[0.0105, 0.0114, 0.01], [0.2488, 0.4829, 0.1672, 1.0]], 182: [[0.0131, 0.0148, 0.0117], [0.2827, 0.3151, 0.5697, 1.0]], 183: [[0.0291, 0.03, 0.0298], [0.3401, 0.7065, 0.313, 1.0]], 184: [[0.0268, 0.0264, 0.0255], [0.1838, 0.3457, 0.4252, 1.0]], 185: [[0.026, 0.0276, 0.026], [0.7638, 0.5138, 0.4137, 1.0]], 186: [[0.0135, 0.0134, 0.0108], [0.4095, 0.2612, 0.1134, 1.0]], 187: [[0.0123, 0.0114, 0.0128], [0.8725, 0.5821, 0.6368, 1.0]], 188: [[0.0282, 0.0286, 0.0259], [0.1531, 0.1403, 0.7786, 1.0]], 189: [[0.0268, 0.0284, 0.027], [0.534, 0.5478, 0.1352, 1.0]], 190: [[0.0139, 0.0125, 0.0131], [0.1229, 0.6524, 0.1363, 1.0]], 191: [[0.0256, 0.0273, 0.0292], [0.4489, 0.4424, 0.6202, 1.0]], 192: [[0.0107, 0.014, 0.0139], [0.5907, 0.8419, 0.1529, 1.0]], 193: [[0.0258, 0.029, 0.0289], [0.2053, 0.3712, 0.1438, 1.0]], 194: [[0.0134, 0.0113, 0.0122], [0.7548, 0.7403, 0.5451, 1.0]], 195: [[0.028, 0.0258, 0.0263], [0.5515, 0.6775, 0.1817, 1.0]], 196: [[0.0116, 0.0125, 0.0101], [0.6894, 0.349, 0.7989, 1.0]], 197: [[0.014, 0.0117, 0.0147], [0.6205, 0.3979, 0.8222, 1.0]], 198: [[0.0293, 0.0297, 0.0265], [0.5543, 0.7775, 0.3461, 1.0]], 199: [[0.0116, 0.0142, 0.0111], [0.2199, 0.2513, 0.8895, 1.0]], 200: [[0.0122, 0.015, 0.0101], [0.7127, 0.2039, 0.8561, 1.0]], 201: [[0.0101, 0.0139, 0.0123], [0.7165, 0.2508, 0.5414, 1.0]], 202: [[0.0145, 0.0127, 0.0122], [0.6902, 0.2927, 0.4518, 1.0]], 203: [[0.0122, 0.0139, 0.0145], [0.381, 0.7173, 0.2822, 1.0]], 204: [[0.0273, 0.0272, 0.0258], [0.4754, 0.4775, 0.2444, 1.0]], 205: [[0.0257, 0.0259, 0.0262], [0.1107, 0.8595, 0.4052, 1.0]], 206: [[0.0259, 0.025, 0.0272], [0.7804, 0.7543, 0.6903, 1.0]], 207: [[0.0105, 0.0119, 0.0135], [0.749, 0.8141, 0.6933, 1.0]], 208: [[0.0119, 0.0107, 0.0107], [0.856, 0.3805, 0.5218, 1.0]], 209: [[0.0292, 0.0284, 0.0293], [0.2554, 0.1836, 0.6714, 1.0]], 210: [[0.0114, 0.0141, 0.015], [0.3084, 0.4264, 0.3643, 1.0]], 211: [[0.0119, 0.0135, 0.0117], [0.7887, 0.7195, 0.8287, 1.0]], 212: [[0.0285, 0.0276, 0.0273], [0.7933, 0.3202, 0.5979, 1.0]], 213: [[0.0275, 0.0251, 0.0262], [0.2273, 0.6012, 0.1663, 1.0]], 214: [[0.0294, 0.0275, 0.0282], [0.1559, 0.4034, 0.7372, 1.0]], 215: [[0.0265, 0.027, 0.0262], [0.3759, 0.6901, 0.6485, 1.0]], 216: [[0.0146, 0.0144, 0.0118], [0.4761, 0.7369, 0.77, 1.0]], 217: [[0.0277, 0.0269, 0.0264], [0.7804, 0.3474, 0.7687, 1.0]], 218: [[0.0111, 0.0125, 0.014], [0.7399, 0.7214, 0.7766, 1.0]], 219: [[0.0259, 0.0265, 0.0276], [0.1382, 0.5468, 0.6439, 1.0]], 220: [[0.0258, 0.0272, 0.0275], [0.1019, 0.7202, 0.8051, 1.0]], 221: [[0.0103, 0.0143, 0.0109], [0.8494, 0.6106, 0.689, 1.0]], 222: [[0.0119, 0.0123, 0.0106], [0.6777, 0.3019, 0.1906, 1.0]], 223: [[0.0142, 0.0102, 0.0127], [0.6006, 0.8386, 0.4337, 1.0]], 224: [[0.0261, 0.0292, 0.03], [0.7282, 0.4049, 0.6946, 1.0]], 225: [[0.029, 0.0255, 0.0293], [0.7953, 0.4592, 0.2398, 1.0]], 226: [[0.0141, 0.0117, 0.011], [0.3067, 0.7225, 0.8106, 1.0]], 227: [[0.0253, 0.0269, 0.0277], [0.2385, 0.2223, 0.5098, 1.0]], 228: [[0.0274, 0.0253, 0.0278], [0.3035, 0.2827, 0.3733, 1.0]], 229: [[0.0281, 0.0269, 0.0296], [0.8317, 0.2099, 0.3577, 1.0]], 230: [[0.0147, 0.0119, 0.0134], [0.692, 0.2368, 0.4432, 1.0]], 231: [[0.0276, 0.0262, 0.029], [0.8697, 0.6931, 0.561, 1.0]], 232: [[0.0113, 0.0124, 0.0131], [0.8249, 0.2718, 0.7071, 1.0]], 233: [[0.026, 0.0259, 0.0289], [0.1262, 0.5121, 0.6388, 1.0]], 234: [[0.0256, 0.0283, 0.0283], [0.3092, 0.7984, 0.8786, 1.0]], 235: [[0.0123, 0.0108, 0.0108], [0.5889, 0.3904, 0.2951, 1.0]], 236: [[0.0105, 0.0117, 0.0115], [0.1966, 0.2143, 0.4989, 1.0]], 237: [[0.0274, 0.0264, 0.0262], [0.1634, 0.6346, 0.1642, 1.0]], 238: [[0.03, 0.0255, 0.0285], [0.8952, 0.3815, 0.5734, 1.0]], 239: [[0.0281, 0.0279, 0.0297], [0.6354, 0.4762, 0.3478, 1.0]], 240: [[0.011, 0.0139, 0.0129], [0.1629, 0.6062, 0.2258, 1.0]], 241: [[0.0124, 0.0107, 0.0141], [0.7121, 0.1756, 0.1465, 1.0]], 242: [[0.0124, 0.0126, 0.0142], [0.1671, 0.7849, 0.5919, 1.0]], 243: [[0.0146, 0.014, 0.0146], [0.2937, 0.8674, 0.1784, 1.0]], 244: [[0.0271, 0.0285, 0.0265], [0.1184, 0.7183, 0.854, 1.0]], 245: [[0.0147, 0.0147, 0.012], [0.4884, 0.7611, 0.7558, 1.0]], 246: [[0.0114, 0.0145, 0.0112], [0.5929, 0.5567, 0.1261, 1.0]], 247: [[0.0139, 0.0119, 0.0132], [0.7868, 0.6975, 0.5545, 1.0]], 248: [[0.0294, 0.0293, 0.0254], [0.1756, 0.7987, 0.5523, 1.0]], 249: [[0.0286, 0.026, 0.0265], [0.7197, 0.4706, 0.1552, 1.0]]}
        # randomize target
        self.obj_gid = self.sim.model.geom_name2id('obj')
        self.tar_gid = self.sim.model.geom_name2id('target')

        self.obj_t_gid = self.sim.model.geom_name2id('top')
        self.obj_b_gid = self.sim.model.geom_name2id('bot')
        self.tar_t_gid = self.sim.model.geom_name2id('t_top')
        self.tar_b_gid = self.sim.model.geom_name2id('t_bot')

        geom_type = self.np_random.choice([3,4,5,6])

        if geom_type == 3:
            size = caps[self.np_random.choice(range(len(caps)))][0]
            color = caps[self.np_random.choice(range(len(caps)))][1]
            top_pos = np.array([0, 0, 1.3*size[1]])
            bot_pos = np.array([0, 0, 1.3*-size[1]])
        elif geom_type == 4:
            size = ellips[self.np_random.choice(range(len(ellips)))][0]
            color = ellips[self.np_random.choice(range(len(ellips)))][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])
        elif geom_type == 5:
            size = cyl[self.np_random.choice(range(len(cyl)))][0]
            color = cyl[self.np_random.choice(range(len(cyl)))][1]
            top_pos = np.array([0, 0, size[1]])
            bot_pos = np.array([0, 0, -size[1]])
        else:
            size = box[self.np_random.choice(range(len(box)))][0]
            color = box[self.np_random.choice(range(len(box)))][1]
            top_pos = np.array([0, 0, size[2]])
            bot_pos = np.array([0, 0, -size[2]])

        color = [x/255 for x in [128, 0, 0, 255]]

        self.sim.model.geom_size[self.obj_gid] = size
        self.sim.model.geom_type[self.obj_gid] = geom_type
        self.sim.model.geom_rgba[self.obj_gid] = color
        self.sim.model.geom_pos[self.obj_t_gid] = top_pos
        self.sim.model.geom_pos[self.obj_b_gid] = bot_pos

        self.sim.model.body_mass[self.obj_bid] = 1.2

        self.sim.model.geom_size[self.tar_gid] = size
        self.sim.model.geom_type[self.tar_gid] = geom_type
        self.sim.model.geom_rgba[self.tar_gid] = color
        self.sim.model.geom_pos[self.tar_t_gid] = top_pos
        self.sim.model.geom_pos[self.tar_b_gid] = bot_pos

        self.sim.model.geom_condim[self.obj_gid] = 3
        desired_orien = np.zeros(3)
        desired_orien[0] = self.np_random.uniform(low=-1, high=1)
        desired_orien[1] = self.np_random.uniform(low=-.8, high=1.2)
        self.sim.model.body_quat[self.target_obj_bid] = euler2quat(desired_orien)

        self.sim.model.site_rgba[self.success_indicator_sid, :2] = np.array([2, 0])

        self.robot.sync_sims(self.sim, self.sim_obsd)
        obs = super().reset()
        return obs

